/*
	$Id: PlateEventTypeTransfer.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005, 2006 Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.PlateEventType;
import net.sf.basedb.core.PlateType;
import net.sf.basedb.core.ProtocolType;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

/**
This class provides methods for transferring PlateEventType data from base 1
to BASE 2.

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/

public class PlateEventTypeTransfer extends Transfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.PlateEventTypeTransfer");
	
	private ProtocolTypeTransfer protocolTypeT = null;
	private PlateTypeTransfer plateTypeT = null;
	private Map<Integer,Integer> ordinals = null;

	/**
	Constructor for PlateEventTypeTransfer depends on ProtocolTypeTransfer
	and PlateTypeTransfer.
	
	@param protocolTypeT {@link ProtocolTypeTransfer}
	@param plateTypeT {@link PlateTypeTransfer}
	 */
	PlateEventTypeTransfer(ProtocolTypeTransfer protocolTypeT,
			PlateTypeTransfer plateTypeT)
	{
		log.debug("Checking that necessary transfers have been done before proceeding with PlateEventType");
		if (protocolTypeT.isDone() && plateTypeT.isDone())
		{
			this.protocolTypeT = protocolTypeT;
			this.plateTypeT = plateTypeT;
		}
		else
		{
			if (!protocolTypeT.isDone()) log.debug("Dependency: Protocols haven't been transfered");
			if (!plateTypeT.isDone()) log.debug("Dependency: PlateTypes haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring PlateEventTypes...");
		int count = getInt("countPlateEventTypes");
		initIdMap(count);
		startProgress(count, "PlateEventTypes");
		run(prepareStatementFromFile("selectPlateEventTypes"));
		done();
		log.info("Transering PlateEventTypes...DONE");
	}

	/**
	Creates a {@link PlateEventType} from the current position in
	<code>ResultSet</code>.
	
	@param b2c {@link DbControl} used to transfer this item
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>PlateEventType</code>
	@throws BaseException
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected PlateEventType createItem(DbControl b2c, ResultSet rs)
			throws BaseException, SQLException
	{
		PlateType plateType = plateTypeT.getPlateTypeById(b2c, rs.getInt(3));
		PlateEventType plateEventType = PlateEventType.getNew(b2c, plateType);
		plateEventType.setName(rs.getString(2));
		plateEventType.setDescription(rs.getString(5));
		plateEventType.setRemoved(rs.getBoolean(6));
		plateEventType.setOrdinal(getNextOrdinal(plateType));
		ProtocolType protocolType = protocolTypeT.getProtocolTypeById(
				b2c, rs.getInt(4));
		plateEventType.setProtocolType(protocolType);
		if (log.isDebugEnabled())
		{
			log.debug("Transfered PlateEventType: " + plateEventType.getName() + "\tBASE1 id(" + rs.getInt(1) + ")");			
		}
		return plateEventType;
	}
	
	private int getNextOrdinal(PlateType plateType)
	{
		if (ordinals == null)
		{
			ordinals = new HashMap<Integer, Integer>();
		}
		Integer next = ordinals.get(plateType.getId());
		next = next == null ? 1 : next + 1;
		ordinals.put(plateType.getId(), next);
		return next;
	}
	
}
