/*
	$Id: ProtocolTransfer.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005 Jari Hakkinen, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.CommonItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileType;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.ProtocolType;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.User;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
This class provides methods to transfer Protocol data from base 1 to BASE 2.

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/

public class ProtocolTransfer extends CommonItemTransfer
{
	private FileType fileType = null;
	private UploadFileTransfer uploadFT = null;
	private ProtocolTypeTransfer protocolTypeT = null;
	
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.ProtocolTransfer");

	/**
	Constructs an ExtractTransfer
	
	@param userT A finished {@link UserTransfer}
	@param groupT A finished {@link GroupTransfer}
	@param uploadFT A finished {@link UploadFileTransfer}
	@param protocolTypeT A finished {@link ProtocolTypeTransfer}
	 */
	ProtocolTransfer(UserTransfer userT, GroupTransfer groupT,
			UploadFileTransfer uploadFT, ProtocolTypeTransfer protocolTypeT)
	{
		super(userT, groupT);
		log.debug("Checking that necessary transfers have been done before proceeding with Protocol");
		if (protocolTypeT.isDone())
		{
			this.uploadFT = uploadFT;
			this.protocolTypeT = protocolTypeT;
		}
		else
		{
			if (!uploadFT.isDone()) log.debug("Dependency: UploadFiles haven't been transfered");
			if (!protocolTypeT.isDone()) log.debug("Dependency: ProtocolTypes haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		try
		{
			log.info("Transferring Protocols...");
			DbControl dc = newDbControl();
			fileType = FileType.getById(
					dc, SystemItems.getId(FileType.PROTOCOL));
			dc.close();
			int count = getInt("countProtocols");
			initIdMap(count);
			startProgress(count, "Protocols");
			run(prepareStatementFromFile("selectProtocols"), 0, getBatchSize());
			done();
			log.info("Transferring Protocols...DONE");
		}
		catch (Exception e)
		{
			log.error("Transferring Protocols FAILED", e);
			stop();
		}
	}

	/**
	Creates an {@link Protocol} from the current position in
	<code>ResultSet</code>.
	
	@param dc {@link DbControl} used to transfer this item
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>Protocol</code> or null if protcol doesn't need saving
	@throws BaseException
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	Protocol createItem(DbControl dc, ResultSet rs) throws BaseException,
			SQLException
	{
		User owner = getOwner(dc, rs.getInt(2));
		if (owner != null)
		{
			File file = uploadFT.getProtocolFileById(dc, rs.getInt(1));
			Integer protocolTypeId = protocolTypeT.getBase2Id(rs.getInt(9));
			if (protocolTypeId != null)
			{
				ProtocolType protocolType = ProtocolType.getById(
						dc, protocolTypeId);
				Protocol protocol = protocolType.newProtocol();
				setCommonItemData(dc, (CommonItem) protocol, rs);
				protocol.setName(rs.getString(NAME));				
				if (file != null)
				{
					protocol.setFile(file);
					file.setFileType(fileType);
				}
				if (log.isDebugEnabled())
				{
					log.debug("Transfered Protocol: " + protocol.getName());
					log.debug("--------------------------------------------------------------------------");
				}
				return protocol;
			}
			else
			{
				log.warn("Id: " + rs.getInt(ID) + " : "
						+ rs.getString(NAME)
						+ " points to an unmapped ProtocolType");
			}
		}
		else
		{
			log.warn("User with id " + rs.getInt(OWNER) + " doesn't exist.");
		}
		return null;
	}
	
	/**
	Returns a Protocol by the given base 1 protocol id.
	
	@param dbc <code>DbControl</code> used to get the Protocol.
	@param base1ProtocolId
	@return <code>Protocol</code> or null if the given protocol id does not exsist.
	 */
	public Protocol getProtocolById(DbControl dbc, int base1ProtocolId)
	{
		try
		{
			Integer base2Id = getBase2Id(base1ProtocolId);
			return base2Id == null ? null : Protocol.getById(dbc, base2Id);
		}
		catch (BaseException e)
		{
			log.warn("Couldn't get protocol from BASE1 by id: " + base1ProtocolId);
			if (log.isDebugEnabled())
			{
				log.debug("\t" + e);
			}
		}
		return null;
	}
	
}
