/*
	$Id: QuotaTransfer.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2006 Johan Enell, Nicklas Nordborg
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.Quota;
import net.sf.basedb.core.QuotaType;
import net.sf.basedb.core.SystemItems;

/**
	 @author nicklas, Martin
	 @version 2.0.1
	 @base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/

public class QuotaTransfer 
	extends Transfer 
{
	
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.QuotaTransfer");
	
	private Map<Long, Quota> quotas = null;
	
	QuotaTransfer()
	{}
	
	public void start()
	{
		log.info("Transferring quotas...");
		DbControl dc = newDbControl();		
		try
		{
			QuotaType totalQuotaType = QuotaType.getById(dc, SystemItems.getId(QuotaType.TOTAL));
			ItemQuery<Quota> query = Quota.getQuery();
			ItemResultList<Quota> result = query.list(dc);
			quotas = new HashMap<Long, Quota>();
			for (Quota quota : result)
			{
				long totalQuota = quota.getQuotaValue(totalQuotaType, Location.PRIMARY);
				quotas.put(totalQuota, quota);
				log.debug("Transferring " + quota);
			}
			done();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		log.info("Transferring quotas...DONE");
	}
	
	/**
		Get the quota which has a total of the given total quota. If no such quota exist
		a new quota is created and saved to the database.
	*/
	public Quota getQuotaForTotal(DbControl dc, long totalQuota)
	{
		Quota quota = quotas.get(totalQuota);
		if (quota == null)
		{
			quota = Quota.getNew(dc, totalQuota);
			quota.setName(formatBytes(totalQuota));
			dc.saveItem(quota);
			quotas.put(totalQuota, quota);
		}
		else if (quota.isInDatabase())
		{
			quota = Quota.getById(dc, quota.getId());
		}
		return quota;
	}
	
	private static final long GB = 1073741824;
	private static final long MB = 1048576;
	private static final long kB = 1024;
	
	/**
		Formats a value using units of bytes, kilobytes, megabytes or gigabytes.

		@param bytes The value to format
		@return A string with a decimal value follwed by bytes, Kb, Mb or Gb.
	*/
	public static final String formatBytes(long bytes)
	{
		if (bytes >= GB)
		{
			float gb = (float)bytes / GB;
			return formatNumber(gb, 1, " GB");
		}
		else if (bytes >= MB)
		{
			float mb = (float)bytes / MB;
			return formatNumber(mb, 1, " MB");
		}
		else if (bytes >= kB)
		{
			float kb = (float)bytes / kB;
			return formatNumber(kb, 1, " kB");
		}
		else
		{
			return bytes + " bytes";
		}
	}
	
	/**
		Formats a decimal number with the specified number of decimals
		and optionally adding a unit.
	
		@param number The number to be formatted
		@param decimals The number of decimals to display, use a negative value to
			display all decimals
		@param unit A string that is added to the end of the formatted number, or
			null to not add anything
		@return A string with the formatted number, or an empty string
			if the number is null
	*/
	public static final String formatNumber(Float number, int decimals, String unit)
	{
		if (number == null || number.isNaN()) return "";
		double fNumber = number.doubleValue();
		double exp = Math.pow(10, decimals);
		String result = null;
		if (decimals == 0)
		{
			result = String.valueOf(Math.round(fNumber));
		}
		else if (decimals < 0)
		{
			result = number.toString();
		}
		else
		{
			long lNumber = number.longValue();
			result =  lNumber+"."+Math.round(fNumber*exp-lNumber*exp);
		}
		if (unit != null) result += unit;
		return result;
	}

	
}
