/*
	$Id: UnmappedPlateTransfer.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005, 2006 Jari Hakkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Hakkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.CommonItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Plate;
import net.sf.basedb.core.PlateType;
import net.sf.basedb.core.Well;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

/**
This class provides methods for transferring Unmapped plate data from base 1
to BASE 2.

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2008-09-11 22:05:50 +0200 (Thu, 11 Sep 2008) $
*/

public class UnmappedPlateTransfer extends CommonItemTransfer
{
	PlateTypeTransfer plateTypeT = null;
	ReporterTransfer reporterT = null;
	// Maps Base 1 ID to destroyed status for a plate
	private Map<Integer, Boolean> destroyedPlates = null;
	
	/**
		The logger
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.WellTransfer");

	/**
	Constructs a new UnmappedPlateTransfer.
	
	@param userT A finished {@link UserTransfer}
	@param groupT A finished {@link GroupTransfer}
	@param plateTypeT A finished {@link PlateTypeTransfer}
	@param reporterT A finished {@link ReporterTransfer}
	 */
	UnmappedPlateTransfer(UserTransfer userT, GroupTransfer groupT,
			PlateTypeTransfer plateTypeT, ReporterTransfer reporterT)
	{
		super(userT, groupT);
		log.debug("Checking that necessary transfers have been done before proceeding with UnmappedPlate");
		if (plateTypeT.isDone() && reporterT.isDone())
		{
			this.plateTypeT = plateTypeT;
			this.reporterT = reporterT;
		}
		else
		{
			if (!plateTypeT.isDone()) log.debug("Dependency: PlateTypes haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring UnmappedPlates...");
		int count = getInt("countUnmappedPlates");
		// init id map for all plates, not just unmapped ones
		initIdMap(count);
		destroyedPlates = new HashMap<Integer, Boolean>(count);
		startProgress(count, "UnmappedPlates");
		runUnBatched(prepareStatementFromFile("selectUnmappedPlates"), 0, getBatchSize());
		done();
		log.info("Transferring UnmappedPlates...DONE");
	}

	/**
	Creates a {@link Plate} from the current position in
	<code>ResultSet</code>.
	
	@param rs <code>ResultSet</code> containing the data for this item.
	@throws BaseException
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected void createItem(ResultSet rs) throws BaseException, SQLException
	{
		DbControl dc = newDbControl();
		PlateType plateType = plateTypeT.getPlateTypeById(dc, rs.getInt(9));
		Plate plate = Plate.getNew(dc, plateType);
		setCommonItemData(dc, (CommonItem) plate, rs);
		plate.setBarcode(rs.getString(11));
		connectReportersToWells(dc, plate, rs.getInt(ID));
		dc.saveItem(plate);
		commit(dc);
		if (log.isDebugEnabled())
		{
			log.debug("Transfered Plate: " + plate.getName() + "\tBASE1 id(" + rs.getInt(1));
			
		}
		mapId(rs.getInt(1), plate.getId());
		destroyedPlates.put(rs.getInt(1), rs.getBoolean(10));
		log.debug("--------------------------------------------------------------------------");
	}

	/**
	Returns the destroyed plates index.
	
	@return <code>boolean[]</code> each index corresponds to the base 2 id.
	 */
	public Map<Integer, Boolean> getDestroyedPlates()
	{
		return destroyedPlates;
	}

	/**
	Connects reporters to wells for a given plate.
	
	@param dc <code>DbControl</code>
	@param plate Wells from this plate will be connected
	@param base1PlateId The corresponding base 1 id for this plate
	 */
	private void connectReportersToWells(DbControl dc, Plate plate,
			int base1PlateId)
	{
		ResultSet rs = null;
		try
		{
			PreparedStatement ps = prepareStatementFromFile("selectWellAndProbesByPlateId");
			ps.setInt(1, base1PlateId);
			rs = ps.executeQuery();
			while (rs.next())
			{
				Well well = plate.getWell(rs.getInt(3)-1, rs.getInt(2)-1);
				well.setReporter(reporterT.getReporterById(dc, rs.getInt(4)));
				if (log.isDebugEnabled())
				{
					log.debug("Connected Reporter: " + well.getReporter().getName() + "to Well: " + plate.getName());
				}
			}
		}
		catch (Exception e)
		{
			log.error("FAILED connecting reporter to well", e);
			stop();
		}
	}
}
