/*
	$Id: DynamicUtil.java 4698 2008-12-10 13:59:34Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.taglib.table.TableColumn;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.MultiFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;


/**
	This class is used to collect some useful methods for building and executing
	dynamic queries.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-12-10 14:59:34 +0100 (Wed, 10 Dec 2008) $
*/
public final class DynamicUtil
{

	public static void addSpotColumns(List<TableColumn> columns, DbControl dc, int channels)
	{
		SessionControl sc = dc.getSessionControl();
		Formatter<Number> numberFormatter = FormatterFactory.getNumberFormatter(sc);
		Formatter<Number> intFormatter = FormatterFactory.getIntFormatter(sc);
		columns.add(new TableColumn("POSITION", "POSITION", "pos()", Type.INT, 
			"Position", "Spot position", "always", true, true, true, 
			Formula.AverageMethod.NONE, intFormatter));
		for (int channel = 1; channel <= channels; ++channel)
		{
			TableColumn tc = new TableColumn("ch"+channel, "%"+channel, "ch("+channel+")", 
				Type.FLOAT, "Ch "+channel, "Channel " + channel + " intensity", "auto", true, true, true,
				Formula.AverageMethod.ARITHMETIC_MEAN, numberFormatter);
			columns.add(tc);
		}
	}
	
	public static void addRawDataColumns(List<TableColumn> columns, DbControl dc, RawDataType rawDataType, String idPrefix, String propertyPrefix, String titlePrefix)
	{
		if (rawDataType.isStoredInDb())
		{
			SessionControl sc = dc.getSessionControl();
			Formatter<Number> intFormatter = FormatterFactory.getIntFormatter(sc);
			columns.add(new TableColumn(idPrefix + "position", propertyPrefix+"position", "raw('position')", 
				Type.INT, titlePrefix+"Position", "Raw spot position", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "block", propertyPrefix+"block", "raw('block')", 
				Type.INT, titlePrefix+"Block", "Block number", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "metaGridX", propertyPrefix+"metaGridX", "raw('metaGridX')", 
				Type.INT, titlePrefix+"Meta grid X", "Meta X coordinate", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "metaGridY", propertyPrefix+"metaGridY", "raw('metaGridY')", 
				Type.INT, titlePrefix+"Meta grid Y", "Meta Y coordinate", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "row", propertyPrefix+"row", "raw('row')", Type.INT, 
				titlePrefix+"Row", "Row number", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "column", propertyPrefix+"column", "raw('column')", 
				Type.INT, titlePrefix+"Column", "Column number", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "x", propertyPrefix+"x", "raw('x')", Type.INT, 
				titlePrefix+"X", "X coordinate", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
			columns.add(new TableColumn(idPrefix + "y", propertyPrefix+"y", "raw('y')", Type.INT, 
				titlePrefix+"Y", "Y coordinate", "auto", true, true, true, 
				Formula.AverageMethod.NONE, intFormatter));
	
			List<RawDataProperty> rdpList = rawDataType.getProperties();
			for(RawDataProperty rdp : rdpList)
			{
				String rdpName = rdp.getName();
				TableColumn tc = new TableColumn(idPrefix+rdpName, propertyPrefix+rdpName, "raw('"+rdpName+"')", rdp.getType(), 
					titlePrefix + rdp.getTitle(), rdp.getDescription(), "auto", true, true, true, 
					rdp.getAverageMethod(), FormatterFactory.getExtendedPropertyFormatter(sc, rdp));
				columns.add(tc);
			}
		}
	}

	public static void addReporterColumns(List<TableColumn> columns, DbControl dc, String idPrefix, String propertyPrefix, String titlePrefix)
	{
		SessionControl sc = dc.getSessionControl();
		Formatter<String> stringFormatter = FormatterFactory.getStringFormatter(sc);
		Formatter<Date> dateFormatter = FormatterFactory.getDateFormatter(sc);
		columns.add(new TableColumn(idPrefix + "id", propertyPrefix+"id", "rep('id')", 
			Type.INT, titlePrefix + "Internal ID", "Reporter internal ID", "auto", true, true, true,
			Formula.AverageMethod.NONE, null));
		columns.add(new TableColumn(idPrefix + "name", propertyPrefix+"name", "rep('name')", 
			Type.STRING, titlePrefix + "Name", "Reporter name", "auto", true, true, true, 
			Formula.AverageMethod.NONE, stringFormatter));
		columns.add(new TableColumn(idPrefix + "externalId", propertyPrefix+"externalId", "rep('externalId')", 
			Type.STRING, titlePrefix + "External ID", "Reporter ID", "auto", true, true, true, 
			Formula.AverageMethod.NONE, stringFormatter));
		columns.add(new TableColumn(idPrefix + "symbol", propertyPrefix+"symbol", "rep('symbol')", 
			Type.STRING, titlePrefix + "Gene symbol", "Gene symbol", "auto", true, true, true, 
			Formula.AverageMethod.NONE, stringFormatter));
		columns.add(new TableColumn(idPrefix + "description", propertyPrefix+"description", "rep('description')", 
			Type.STRING, titlePrefix + "Description", "Reporter description", "auto", true, true, true, 
			Formula.AverageMethod.NONE, stringFormatter));
		columns.add(new TableColumn(idPrefix + "entryDate", propertyPrefix+"entryDate", "rep('entryDate')", 
				Type.DATE, titlePrefix + "Registered", "Registered", "auto", true, true, true, 
				Formula.AverageMethod.NONE, dateFormatter));
		columns.add(new TableColumn(idPrefix + "lastUpdate", propertyPrefix+"lastUpdate", "rep('lastUpdate')", 
			Type.DATE, titlePrefix + "Last update", "Last updateded", "auto", true, true, true, 
			Formula.AverageMethod.NONE, dateFormatter));

		List<ExtendedProperty> extended = ExtendedProperties.getProperties("ReporterData");
		if (extended != null)
		{
			for (ExtendedProperty ep : extended)
			{
				String name = ep.getName();
				TableColumn tc = new TableColumn(idPrefix + name, propertyPrefix + name, "rep('"+name+"')", 
					ep.getType(), titlePrefix + ep.getTitle(), ep.getDescription(), "auto", true, true, true,
					ep.getAverageMethod(), FormatterFactory.getExtendedPropertyFormatter(sc, ep));
				columns.add(tc);
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	public static void addExtraColumns(List<TableColumn> columns, DbControl dc, BioAssaySet bas, String idPrefix, String propertyPrefix, String titlePrefix)
	{
		ItemQuery<ExtraValue> query = bas.getExtraValues();
		query.order(Orders.asc(Hql.property("extraValueType.name")));
		ItemResultIterator<ExtraValue> result = query.iterate(dc);
		SessionControl sc = dc.getSessionControl();
		while (result.hasNext())
		{
			ExtraValue ev = result.next();
			int id = ev.getId();
			ExtraValueType evt = null;
			TableColumn tc = null;
			try
			{
				evt = ev.getExtraValueType();
				tc = new TableColumn(idPrefix+id, propertyPrefix + id, "xtra('" + evt.getExternalId() + "')", 
						ev.getValueType(), titlePrefix+evt.getName(), evt.getDescription(), 
						"auto", true, true, true, evt.getAverageMethod(), null);
			}
			catch (Throwable t)
			{
				tc = new TableColumn(idPrefix+id, propertyPrefix + id, "xtra('" + id + "')", 
						ev.getValueType(), titlePrefix+ev.getValueType(), "", 
						"auto", true, true, true, null, null);
			}
			Formatter<?> formatter = FormatterFactory.getTypeFormatter(sc, ev.getValueType());
			if (ev.getValueType().isNumerical() && evt != null)
			{
				// Numerical types *should* return a numeric formatter which is required by the color formatter
				formatter = FormatterFactory.getColorFormatter(sc, evt.getColoring(), 
					(Formatter<Number>)formatter);
			}
			tc.setFormatter(formatter);
			columns.add(tc);
		}
		result.close();
	}
	
	public static void addFormulaColumns(List<TableColumn> columns, DbControl dc, 
		RawDataType rawDataType, Formula.Type type, String idPrefix, String titlePrefix, boolean allowRawFunction)
	{
		ItemQuery<Formula> query = Formula.getQuery(type, rawDataType);
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		query.order(Orders.asc(Hql.property("name")));
		ItemResultIterator<Formula> result = query.iterate(dc);
		SessionControl sc = dc.getSessionControl();
		Formatter<Number> floatFormatter = FormatterFactory.getNumberFormatter(sc);
		Formatter<Number> intFormatter = FormatterFactory.getIntFormatter(sc);
		Formatter<Date> dateFormatter = FormatterFactory.getDateFormatter(sc);
		Formatter<Boolean> booleanFormatter = FormatterFactory.getBooleanFormatter(sc);
		Formatter<String> stringFormatter = FormatterFactory.getStringFormatter(sc);
		Formatter<Object> defaultFormatter = new ToStringFormatter<Object>();
		while (result.hasNext())
		{
			Formula f = result.next();
			String jepExpression = f.getFormula(0);
			if (allowRawFunction || !jepExpression.contains("raw("))
			{
				TableColumn tc = new TableColumn(idPrefix+f.getId(), "="+jepExpression, jepExpression, Type.FLOAT, 
						titlePrefix + f.getName(), f.getDescription(), "auto", true, true, true, 
						f.getAverageMethod(), null);
				Formatter<Number> intColorFormatter = FormatterFactory.getColorFormatter(sc, f.getColoring(), intFormatter);
				Formatter<Number> floatColorFormatter = FormatterFactory.getColorFormatter(sc, f.getColoring(), floatFormatter);
				MultiFormatter multiFormatter = new MultiFormatter(defaultFormatter, true);
				multiFormatter.registerFormatter(Integer.class, intColorFormatter);
				multiFormatter.registerFormatter(Long.class, intColorFormatter);
				multiFormatter.registerFormatter(Number.class, floatColorFormatter);
				multiFormatter.registerFormatter(Date.class, dateFormatter);
				multiFormatter.registerFormatter(Boolean.class, booleanFormatter);
				multiFormatter.registerFormatter(String.class, stringFormatter);
				tc.setFormatter(multiFormatter);
				columns.add(tc);
			}
		}
		result.close();
	}
	
	public static SelectedInfo getSelectedColumns(ItemContext cc, List<TableColumn> columns, boolean isAveraging)
	{
		List<String> selectedIds = new ArrayList<String>(columns.size());
		List<String> selectedProperties = new ArrayList<String>(columns.size());
		List<TableColumn> selectedColumns = new ArrayList<TableColumn>(columns.size());
		
		String visibleColumns = cc.getSetting("columns");
		if (visibleColumns == null) visibleColumns = "all";
		boolean allVisible = "all".equals(visibleColumns);
		Set<String> visible = new HashSet<String>(Arrays.asList(visibleColumns.split(",")));
		for (TableColumn tc : columns)
		{
			if (allVisible || visible.contains(tc.getId()) || tc.getShow().equals("always"))
			{
				if (!isAveraging || tc.isAveragable())
				{
					selectedProperties.add(tc.getProperty());
					selectedIds.add(tc.getId());
					selectedColumns.add(tc);
				}
			}
		}
		return new SelectedInfo(selectedProperties, selectedIds, selectedColumns);
	}	
	
	public static class SelectedInfo
	{
		public final List<String> selectedProperties;
		public final List<String> selectedIds;
		public final List<TableColumn> selectedColumns;
		
		private SelectedInfo(List<String> properties, List<String> ids, List<TableColumn> columns)
		{
			this.selectedIds = ids;
			this.selectedProperties = properties;
			this.selectedColumns = columns;
		}
	}
	
}

