/*
	$Id: PermissionUtil.java 4690 2008-12-08 12:27:52Z nicklas $

	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web;

import net.sf.basedb.core.Identifiable;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;

import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;
import java.util.EnumSet;


/**
	This class is used to collect some useful methods for assigning
	permissions to roles and shared items.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-12-08 13:27:52 +0100 (Mon, 08 Dec 2008) $
*/
public final class PermissionUtil
{

	public static final String getShortPermissions(Identifiable item)
	{
		Set<Permission> itemPermissions = item.getType().getDefinedPermissions();
		StringBuilder sb = new StringBuilder();
		if(itemPermissions != null)
		{
			if (item.hasPermission(Permission.READ) && itemPermissions.contains(Permission.READ)) sb.append("R");
			if (item.hasPermission(Permission.USE) && itemPermissions.contains(Permission.USE)) sb.append("U");
			if (item.hasPermission(Permission.WRITE) && itemPermissions.contains(Permission.WRITE)) sb.append("W");
			if (item.hasPermission(Permission.DELETE) && itemPermissions.contains(Permission.DELETE)) sb.append("D");
			if (item.hasPermission(Permission.SET_OWNER) && itemPermissions.contains(Permission.SET_OWNER)) sb.append("O");
			if (item.hasPermission(Permission.SET_PERMISSION) && itemPermissions.contains(Permission.SET_PERMISSION)) sb.append("P");
		}
		else
		{
			if (item.hasPermission(Permission.READ)) sb.append("R");
			if (item.hasPermission(Permission.USE)) sb.append("U");
			if (item.hasPermission(Permission.WRITE)) sb.append("W");
			if (item.hasPermission(Permission.DELETE)) sb.append("D");
		}
		if(sb.length()==0)sb.append("-Not defined-");
		return sb.toString();
	}
	
	public static final String getFullPermissionNames(Identifiable item)
	{
		Set<Permission> itemPermissions = item.getType().getDefinedPermissions();
		StringBuilder sb = new StringBuilder();
		if(itemPermissions != null)
		{
			if (item.hasPermission(Permission.READ) && itemPermissions.contains(Permission.READ)) sb.append("Read");		
			if (item.hasPermission(Permission.USE) && itemPermissions.contains(Permission.USE)) sb.append(sb.length()==0 ? "Use" : ", Use");
			if (item.hasPermission(Permission.WRITE) && itemPermissions.contains(Permission.WRITE)) sb.append(sb.length()==0 ? "Write" : ", Write");
			if (item.hasPermission(Permission.DELETE) && itemPermissions.contains(Permission.DELETE)) sb.append(sb.length()==0 ? "Delete" : ", Delete");
			if (item.hasPermission(Permission.SET_OWNER) && itemPermissions.contains(Permission.SET_OWNER)) sb.append(sb.length()==0 ? "Take ownership" : ", Take ownership");
			if (item.hasPermission(Permission.SET_PERMISSION) && itemPermissions.contains(Permission.SET_PERMISSION)) sb.append(sb.length()==0 ? "Change permisson" : ", Change permission");
		}
		else
		{			
			if(item.hasPermission(Permission.READ)) sb.append("Read");
			if (item.hasPermission(Permission.USE)) sb.append(sb.length()==0 ? "Use" : ", Use");
			if (item.hasPermission(Permission.WRITE)) sb.append(sb.length()==0 ? "Write" : ", Write");
			if (item.hasPermission(Permission.DELETE)) sb.append(sb.length()==0 ? "Delete" : ", Delete");
		}
		if(sb.length()==0)sb.append("No permissions defined for this item");
		return sb.toString();
	}
	
	/**
		Create a comma-separated string with "readable" permission names.
		@since 2.10
	*/
	public static final String getFullPermissionNames(Set<Permission> permissions)
	{
		StringBuilder sb = new StringBuilder();
		if (permissions != null)
		{
			if (permissions.contains(Permission.READ)) sb.append("Read");
			if (permissions.contains(Permission.USE)) sb.append(sb.length()==0 ? "Use" : ", Use");
			if (permissions.contains(Permission.WRITE)) sb.append(sb.length()==0 ? "Write" : ", Write");
			if (permissions.contains(Permission.DELETE)) sb.append(sb.length()==0 ? "Delete" : ", Delete");
			if (permissions.contains(Permission.SET_OWNER)) sb.append(sb.length()==0 ? "Set owner" : ", Set owner");
			if (permissions.contains(Permission.SET_PERMISSION)) sb.append(sb.length()==0 ? "Change permission" : ", Change permission");
		}
		if (sb.length()==0) sb.append("No permissions defined for this item");
		return sb.toString();
	}

	
	/**
		Translate the given set of permissions to an integer code that
		can be used to later recreate the set again.
		@param permissions A <code>Set</code> with the permissions
		@return An integer
		@see #getPermissions(int)
	*/
	public static final int getPermissionCode(Set<Permission> permissions)
	{
		int result = 0;
		for (Permission p : Permission.values())
		{
			result += permissions.contains(p) ? 1 << p.ordinal() : 0;
		}
		return result;
	}
	
	/**
		Translate a given integer to a set of permissions.
		@param permissionCode The integer representing permissions
		@return A <code>Set</code> containing the <code>Permission</code>:s
		@see #getPermissionCode(Set)
	*/
	public static final Set<Permission> getPermissions(int permissionCode)
	{
		EnumSet<Permission> permissions = EnumSet.noneOf(Permission.class);
		for (Permission p : Permission.values())
		{
			if ((permissionCode & (1 << p.ordinal())) > 0) permissions.add(p);
		}
		return permissions;
	}
	
	private static Map<String, List<Item>> permissionGroups;
	
	/**
		All items which permissions depend on some other item only.
	*/
	private static final Set<Item> ignored = EnumSet.of(
		Item.SYSTEM, Item.HELP, 
		Item.GLOBALDEFAULTSETTING, Item.USERDEFAULTSETTING, Item.CLIENTDEFAULTSETTING, Item.USERCLIENTSETTING,
		Item.ANNOTATIONSET, Item.ANNOTATION, Item.BIOMATERIALEVENT, 
		Item.ITEMKEY, Item.PROJECTKEY, Item.ROLEKEY,
		Item.PLATEEVENT, Item.PLATEEVENTTYPE, Item.WELL,
		Item.ARRAYDESIGNBLOCK, Item.FEATURE, Item.ARRAYDESIGNPLATE,
		Item.IMAGE, Item.SPOTIMAGES, Item.RAWDATA, 
		Item.BIOASSAYSET, Item.BIOASSAY, Item.TRANSFORMATION, Item.EXTRAVALUE,
		Item.VIRTUALDB, Item.DATACUBE, Item.DATACUBELAYER, Item.DATACUBECOLUMN, Item.DATACUBEFILTER,
		Item.DATACUBEEXTRAVALUE, 
		Item.FILESET, Item.FILESETMEMBER, 
		Item.PLATFORMVARIANT, Item.PLATFORMFILETYPE
	);
	
	public static final synchronized Map<String, List<Item>> getPermissionGroups()
	{
		if (permissionGroups == null)
		{
			if (permissionGroups == null)
			{
				Map<String, List<Item>> temp = new LinkedHashMap<String, List<Item>>();
				
				temp.put("Server administration", 
					Arrays.asList(new Item[] 
					{ 
						Item.USER, Item.GROUP, Item.ROLE, Item.QUOTA, Item.QUOTATYPE,
						Item.NEWS, Item.CLIENT, Item.SESSION, Item.SETTING, Item.DISKUSAGE
					})
				);
				
				temp.put("Other administration", 
					Arrays.asList(new Item[]
					{
						Item.PROTOCOL, Item.PROTOCOLTYPE, Item.HARDWARE, Item.HARDWARETYPE, Item.SOFTWARE, Item.SOFTWARETYPE,
						Item.ANNOTATIONTYPE, Item.ANNOTATIONTYPECATEGORY, Item.FILETYPE, Item.MIMETYPE, Item.REPORTER, 
						Item.REPORTERLIST, Item.REPORTERTYPE, Item.EXTRAVALUETYPE,
						Item.PLATFORM, Item.DATAFILETYPE, Item.QUANTITY, Item.UNIT
					})
				);
				
				temp.put("Biomaterials &amp; experiments",
					Arrays.asList(new Item[]
					{
						Item.BIOSOURCE, Item.SAMPLE, Item.EXTRACT, Item.LABELEDEXTRACT, Item.LABEL,
						Item.BIOMATERIALLIST,
						Item.HYBRIDIZATION, Item.SCAN, Item.RAWBIOASSAY, Item.EXPERIMENT, Item.FORMULA
					})
				);
				
				temp.put("Array LIMS",
					Arrays.asList(new Item[]
					{
						Item.PLATEGEOMETRY, Item.PLATETYPE, Item.PLATE, Item.PLATEMAPPING,
						Item.ARRAYDESIGN, Item.ARRAYBATCH, Item.ARRAYSLIDE
					})
				);
				
				temp.put("Other",
					Arrays.asList(new Item[]
					{
						Item.FILE, Item.DIRECTORY, Item.PROJECT, Item.MESSAGE,
						Item.PLUGINTYPE, Item.PLUGINDEFINITION, Item.PLUGINCONFIGURATION, Item.JOB, 
						Item.JOBAGENT
					})
				);

				permissionGroups = Collections.unmodifiableMap(temp);
			}
		}
		return permissionGroups;
	}
	
	public static String translatePermissionsToString(Set<Permission> permissions)
	{
		StringBuilder permissionString = new StringBuilder();
		permissionString.append(permissions.contains(Permission.READ) ? "R" : "");
		permissionString.append(permissions.contains(Permission.USE) ? "U" : "");
		permissionString.append(permissions.contains(Permission.WRITE) ? "W" : "");
		permissionString.append(permissions.contains(Permission.DELETE) ? "D" : "");
		permissionString.append(permissions.contains(Permission.SET_OWNER) ? "O" : "");
		permissionString.append(permissions.contains(Permission.SET_PERMISSION) ? "P" : "");
		return permissionString.toString();
	}
	
}

