/*
	$Id: Icon.java 4512 2008-09-11 20:05:50Z jari $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import net.sf.basedb.util.Values;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;

/**
	This tag is used to make it easier to display icons
	on a BASE web page.
	<p>
	<b>Example:</b>
	<pre>
   &lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
   &lt;base:icon image="required.gif" /&gt;
</pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:icon
      id=...
      clazz=...
      style=...
      image=...
      tooltip=...
      onclick=...
      visible=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>The ID of the icon. The value of this attribute
		goes directly into the standard <code>id</code> attribute.
		This means that if you need a reference to the note object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		icon = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>icon</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different icons.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different icons.
		</td>
	</tr>
	<tr>
		<td>image</td>
		<td></td>
		<td>yes</td>
		<td>
			The name of the image to display. The image must be in the <code>/images</code>
			directory. Only include the filename, not the path.
			This taglib will find the correct path from the current jsp page.
		</td>
	</tr>
	<tr>
		<td>tooltip</td>
		<td></td>
		<td>no</td>
		<td>
			A short explanation that will show up as a tooltip.
		</td>
	</tr>
	<tr>
		<td>onclick</td>
		<td></td>
		<td>no</td>
		<td>
			An optional JavaScript statement that will exceute when the user
			clicks on the icon.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the note should be visible or not.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Icon 
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = 6886281373891822996L;

	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;

	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = "icon";
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;
	
	/**
		The filename of the image.
	*/
	private String image = null;

	/**
		An optional tooltip.
	*/
	private String tooltip = null;
	
	/**
		An optional JavaScript statement.
	*/
	private String onClick = null;

	/**
		If the icon should be visible or not.
	*/
	private boolean visible = true;

	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}

	public void setImage(String image)
	{
		this.image = image;
	}
	public String getImage()
	{
		return image;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}

	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}

	public void setOnclick(String onClick)
	{
		this.onClick = onClick;
	}
	public String getOnclick()
	{
		return onClick;
	}

	public int doStartTag()
		throws JspException
	{
		page = (Page)findAncestorWithClass(this, Page.class);
		if (page == null) throw new JspTagException("Tag <base:icon> must be inside a <base:page> tag");
		if (!isVisible()) return SKIP_BODY;

		String theStyle = getStyle();
		if (getOnclick() != null)
		{
			theStyle = Values.getString(theStyle) + "cursor: pointer;";
		}

		StringBuilder sb = new StringBuilder();
		sb.append("<img border=0");
		if (getId() != null) sb.append(" id=\"").append(getId()).append("\"");
		if (getClazz() != null) sb.append(" class=\"").append(getClazz()).append("\"");
		if (theStyle != null) sb.append(" style=\"").append(theStyle).append("\"");
		if (getTooltip() != null) sb.append(" title=\"").append(getTooltip()).append("\"");
		if (getOnclick() != null) sb.append(" onclick=\"").append(getOnclick()).append("\"");
		sb.append(" src=\"").append(page.getRoot()).append("images/").append(image).append("\"");
		sb.append(">");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}
}

