/*
	$Id: AnnotationSimpleRestriction.java 4555 2008-10-02 11:55:17Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;

import java.util.List;

import org.hibernate.SQLQuery;

/**
	Restricts a query using annotation values with a simple expression:
	<code>annotation operator value</code>.
	
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Annotations
	@base.modified $Date: 2008-10-02 13:55:17 +0200 (Thu, 02 Oct 2008) $
*/
public class AnnotationSimpleRestriction
	extends AnnotationRestriction
{
	private final Operator operator;
	private final Object value;
	private final List<Object> values;

	/**
		Create a new annotation restriction using a simple expression:
		<code>annotation operator value</code>. The restriction is applied to
		the root element of the query.
		
		@deprecated Use {@link #AnnotationSimpleRestriction(String, AnnotationType, Operator, Object, boolean)}
			instead with alias=null
	*/
	public AnnotationSimpleRestriction(AnnotationType annotationType, Operator operator, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		this(null, annotationType, operator, value, includeInheriting);
	}

	/**
		Create a new annotation restriction using a simple expression:
		<code>annotation operator value</code>. 
		
		@param alias The alias of a joined item where the annotations are
			located or null to use the root entity of the query
		@param annotationType The annotation type to use in the query
		@param operator The operator, it must be one of the expression
			operators, EQ, NEQ, etc., not a boolean operator, AND, OR, etc.
		@param value The value to use in the query, it must be of the 
			correct value type for the annotation as defined by the 
			{@link AnnotationType#getValueType()} property
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@throws InvalidDataException If any of the parameters are null
			or not follow the rules above.
		@since 2.4
	*/
	public AnnotationSimpleRestriction(String alias, AnnotationType annotationType, Operator operator, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		super(alias, annotationType, includeInheriting);
		if (operator == null) throw new InvalidUseOfNullException("operator");
		if (value == null) throw new InvalidUseOfNullException("value");
		if (!operator.isExpressionOperator())
		{
			throw new InvalidDataException("Invalid operator for expression: "+operator);
		}
		if (!valueType.isCorrectType(value) && !(valueType.isNumerical() && value instanceof Number))
		{
			throw new InvalidDataException("Value '"+value+"' is a "+
				value.getClass()+", not a "+valueType);
		}
		this.operator = operator;
		this.value = value;
		this.values = null;
	}

	
	/**
	 	Creates a new annotation restriction for a specific <code>valueType</code>.
	 	The restriction is applied to the root element of the query.
		
		@deprecated Use {@link #AnnotationSimpleRestriction(String, int, Type, Operator, Object, boolean)}
			instead with alias=null
	*/
	public AnnotationSimpleRestriction(int annotationTypeId, Type valueType, Operator operator, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		this(null, annotationTypeId, valueType, operator, value, includeInheriting);
	}
	
	/**
	 	Creates a new annotation restriction for a specific <code>valueType</code>.
		
		@param alias The alias of a joined item where the annotations are
			located or null to use the root entity of the query
	  	@param annotationTypeId the id of the annotation type to use in the query
	 	@param valueType the type of values for annotations
	  	@param operator The operator, it must be one of the expression
			operators, EQ, NEQ, etc., not a boolean operator, AND, OR, etc.
	  	@param value The value to use in the query, it must be of the 
			correct value type for the annotation as defined by the 
			{@link AnnotationType#getValueType()} property
	  	@param includeInheriting TRUE if items inheriting the annotation should be returned by the query, FALSE otherwise
	  	@throws InvalidDataException If any of the parameters are null
			or not follow the rules above
		@since 2.4
	*/
	public AnnotationSimpleRestriction(String alias, int annotationTypeId, Type valueType, Operator operator, Object value, boolean includeInheriting)
		throws InvalidDataException
	{
		super(alias, annotationTypeId, valueType, includeInheriting);
		if (valueType == null) throw new InvalidUseOfNullException("valueType");
		if (operator == null) throw new InvalidUseOfNullException("operator");
		if (value == null) throw new InvalidUseOfNullException("value");
		if (!operator.isExpressionOperator())
		{
			throw new InvalidDataException("Invalid operator for expression: "+operator);
		}
		if (!valueType.isCorrectType(value) && !(valueType.isNumerical() && value instanceof Number))
		{
			throw new InvalidDataException("Value '"+value+"' is a "+
				value.getClass()+", not a "+valueType);
		}
		this.operator = operator;
		this.value = value;
		this.values = null;
	}
	
	/**
	 	Creates a new annotation restriction for a specific <code>valueType</code>.
	 	The restriction is with a list of values.
		
		@param alias The alias of a joined item where the annotations are
			located or null to use the root entity of the query
	  	@param annotationTypeId the id of the annotation type to use in the query
	 	@param valueType the type of values for annotations
	  	@param operator The operator, it must be one of the expression
			operators, EQ, NEQ, etc., not a boolean operator, AND, OR, etc.
	  	@param values The list of values that should be used in the query, they must be of the 
			correct value type for the annotation as defined by the 
			{@link AnnotationType#getValueType()} property.
	  	@param includeInheriting TRUE if items inheriting the annotation should be returned by the query, FALSE otherwise
	  	@throws InvalidDataException If any of the parameters are null
			or not follow the rules above
		@since 2.5
	*/
	public AnnotationSimpleRestriction(String alias, int annotationTypeId, Type valueType, Operator operator, List<Object> values, boolean includeInheriting)
		throws InvalidDataException
	{
		super(alias, annotationTypeId, valueType, includeInheriting);
		if (valueType == null) throw new InvalidUseOfNullException("valueType");
		if (operator == null) throw new InvalidUseOfNullException("operator");
		if (values == null) throw new InvalidUseOfNullException("values");
		if (!operator.isExpressionOperator())
		{
			throw new InvalidDataException("Invalid operator for expression: "+operator);
		}
		for (Object obj : values)
		{
			if (!valueType.isCorrectType(obj) && !(valueType.isNumerical() && obj instanceof Number))
			{
				throw new InvalidDataException("Value '"+obj+"' is a "+
					obj.getClass()+", not a "+valueType);
			}
		}
		this.operator = operator;
		this.values = values;
		this.value = null;
	}
	
	/*
		From the AnnotationRestriction class
		-------------------------------------------
	*/
	String getRestrictionSql(Query query, DbControl dc)
		throws BaseException
	{
		org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();		
		String restriction = null;
		if (!operator.isListOperator())
		{
			restriction = "v."+dialect.openQuote()+"value"+dialect.closeQuote()+" "+operator.getSqlSymbol()+ " :value";
		}
		else
		{
			if (operator == Operator.IN || operator == Operator.NOTIN)
			{
				restriction = "v."+dialect.openQuote()+"value"+dialect.closeQuote()+" "+operator.getSqlSymbol()+ " (:value)";
			}
			else if (operator == Operator.NOTLIKE_IN || operator == Operator.LIKE_IN)
			{
				boolean isFirst = true;
				restriction = "(";
				for (Object obj : values)
				{
					restriction += isFirst ? "" : 
						operator == Operator.NOTLIKE_IN ? " AND ": " OR ";
					restriction += "v."+dialect.openQuote()+"value"+dialect.closeQuote()+" "+operator.getSqlSymbol()+ " :value"+System.identityHashCode(obj);
					isFirst = false;
				}
				restriction += ")"; 
			}
		}
		return restriction;
	}
	
	void setRestrictionParameters(SQLQuery query)
	{
		if (!operator.isListOperator())
		{
			query.setParameter("value", value);
		}
		else
		{
			if (operator == Operator.IN || operator == Operator.NOTIN)
			{
				query.setParameterList("value", values);
			}
			else if (operator == Operator.LIKE_IN || operator == Operator.NOTLIKE_IN)
			{
				for (Object obj : values)
				{
					query.setParameter("value"+System.identityHashCode(obj), obj);
				}
			}
		}
		
	}
	// -----------------------------------------

}

