/*
	$Id: BioAssay.java 4517 2008-09-11 20:11:02Z jari $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.BioAssayData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

/**
	Represents the current intensities of raw data after some transformation 
	has been applied to it. Usually a single bioassay represents the data
	from a single {@link RawBioAssay} but it is also possible that
	a transformation has merged several bioassays into a single one.
	<p>
	Bioassays are created by one of the {@link BioAssaySet#newRootBioAssay(Collection)},
	{@link BioAssaySet#newBioAssay(BioAssay)} or 
	{@link BioAssaySet#newBioAssay(Collection)}. Which one to use depends on the
	relationship to it's parents. Bioassays can only be created if the bioassayset
	hasn't been committed to the database.
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2005-10-19 15:57:12 +0200 (on, 19 okt 2005) $
*/
public class BioAssay
	extends ChildItem<BioAssayData>
	implements Nameable, Annotatable
{
	/**
		The type of item represented by this class.
		@see Item#BIOASSAY
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOASSAY;

	/**
		Create a new <code>BioAssay</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param bioAssaySet The bioassayset the bioassay belongs to
		@param column The data cube column where the bioassay should store it's data;
			must be unique among bioassays in a bioassayset
		@return The new <code>BioAssay</code> item
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for the experiment
		@throws InvalidDataException If the bioassayset is null
		@throws BaseException If there is an error
	*/
	static BioAssay getNew(DbControl dc, BioAssaySet bioAssaySet, DataCubeColumn column)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		BioAssay ba = dc.newItem(BioAssay.class);
		ba.setName("New bioassay");
		ba.setBioAssaySet(bioAssaySet);
		ba.setDataCubeColumn(column);
		return ba;
	}

	/**
		Get a <code>BioAssay</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioAssay</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static BioAssay getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioAssay ba = dc.loadItem(BioAssay.class, id);
		if (ba == null) throw new ItemNotFoundException("BioAssay[id=" + id + "]");
		return ba;
	}
	
	/**
		Get a query configured to retrieve bioassays in
		a given bioassayset.
		@param bioAssaySet The bioassayset to retreive bioassays for, null
			is not allowed
		@return An {@link ItemQuery} object
	 	@throws InvalidDataException If the parameter is null.
	*/
	public static ItemQuery<BioAssay> getQuery(BioAssaySet bioAssaySet)
		throws InvalidDataException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		ItemQuery<BioAssay> query = new ItemQuery<BioAssay>(BioAssay.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("bioAssaySet"),
				Hql.entity(bioAssaySet)
			)
		);
		return query;
	}

	/**
	 	Creates a new bioassay item from the given data.
		@param data the data
	*/
	BioAssay(BioAssayData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		@since 2.2
	*/
	public AnnotationSet getAnnotationSet()
		throws PermissionDeniedException, BaseException
	{
		DbControl dc = getDbControl();
		AnnotationSet as = dc.getItem(AnnotationSet.class, getData().getAnnotationSet(), this);
		if (as == null)
		{
			as = AnnotationSet.getNew(dc, this);
			getData().setAnnotationSet(as.getData());
			dc.saveItemIf(this, as, false);
		}
		return as;
	}
	/**
		@since 2.2
	*/
	public boolean isAnnotated()
	{
		return getData().getAnnotationSet() != null;
	}
	/**
		@since 2.2
	*/
	public void removeAnnotations()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		AnnotationSetData as = getData().getAnnotationSet();
		if (as != null)
		{
			HibernateUtil.deleteData(getDbControl().getHibernateSession(), as);
			getData().setAnnotationSet(null);
		}
	}
	/**
		Default implementation returns null. Should be overriden by subclasses
		that has protocols.
		@return Always null
		@since 2.2
	*/
	public Protocol getProtocol()
	{
		return null;
	}
	/**
		Get the parent bioassay set, all parent bioassays and all parent 
		raw bioassays.
		@since 2.2
	*/
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		DbControl dc = getDbControl();
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		annotatable.add(getBioAssaySet());
		ItemQuery<BioAssay> parentQuery = getParents();
		if (parentQuery != null)
		{
			parentQuery.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
			annotatable.addAll(parentQuery.list(dc));
		}
		ItemQuery<RawBioAssay> rawQuery = getRawBioAssays();
		if (rawQuery != null)
		{
			rawQuery.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.IN_PROJECT);
			annotatable.addAll(rawQuery.list(dc));
		}
		return annotatable;
	}
	// -------------------------------------------

	/*
		From BasicItem
		-------------------------------------------
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.CREATE)
		{
			getData().setNumSpots(getBioAssaySet().getSpotCount(getDataCubeColumnNo()));
		}
	}
	/**
		Deny DELETE permission. BioAssays are automatically deleted when
		it's parent bioassayset is deleted.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		super.initPermissions(granted, denied | Permission.deny(Permission.DELETE));
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	SharedData getSharedParent()
	{
		return getData().getBioAssaySet().getExperiment();
	}
	/**
		USE permission is enough to be able to manage bioassays.
	*/
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the experiment this bioassay belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getBioAssaySet().getExperiment());
	}
	
	/**
		Get the bioassayset this bioassay belongs to.
		@return A <code>BioAssaySet</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the bioassayset
		@throws BaseException If there is another error
	*/
	public BioAssaySet getBioAssaySet()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioAssaySet.class, getData().getBioAssaySet());
	}
	
	/**
		Set the experiment. This cannot be changed after the bioassayset
		has been created.
	*/
	private void setBioAssaySet(BioAssaySet bioAssaySet)
		throws PermissionDeniedException, InvalidDataException
	{
		if (bioAssaySet == null) throw new InvalidUseOfNullException("bioAssaySet");
		bioAssaySet.checkPermission(Permission.WRITE);
		getData().setBioAssaySet(bioAssaySet.getData());
	}
	
	/**
		Get the raw data type the experiment uses for the raw data.
		@return A <code>RawDataType</code> object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getSafeRawDataType(getData().getBioAssaySet().getExperiment().getRawDataType());
	}

	/**
		Get the number of spots in this bioassay.
	*/
	public int getNumSpots()
	{
		return getData().getNumSpots();
	}
	
	/**
		Get the number of spots that are stored in files.
		@since 2.8
	*/
	public int getNumFileSpots()
	{
		return getData().getNumFileSpots();
	}
	
	/**
		Set the number of spots that are stored in files.
		@param numSpots The number of spots
		@since 2.8
	*/
	public void setNumFileSpots(int numSpots)
	{
		checkPermission(Permission.WRITE);
		getData().setNumFileSpots(numSpots);
	}

	/**
		Get the number of the data cube this bioassay stores it's data in.
		@see #getDataCubeLayerNo()
		@see #getDataCubeColumnNo()
	*/
	public short getDataCubeNo()
	{
		return getData().getDataCubeColumn().getDataCube().getCubeNo();
	}

	/**
		Get the layer coordinate in the data cube that this bioassay stores it's data in.
		@see #getDataCubeNo()
		@see #getDataCubeColumnNo()
	*/
	public short getDataCubeLayerNo()
	{
		return getData().getBioAssaySet().getDataCubeLayer().getLayerNo();
	}
	
	/**
		Get the column coordinate in the data cube that this bioassay
		stores it's data in.
		@see #getDataCubeNo()
		@see #getDataCubeLayerNo()
	*/
	public short getDataCubeColumnNo()
	{
		return getData().getDataCubeColumn().getColumnNo();
	}

	/**
		Get the <code>DataCubeColumn</code> object where this bioassay
		store it's data.
	*/
	DataCubeColumn getDataCubeColumn()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(DataCubeColumn.class, getData().getDataCubeColumn());
	}
	private void setDataCubeColumn(DataCubeColumn column)
	{
		if (column == null) throw new InvalidUseOfNullException("dataCubeColumn");
		getData().setDataCubeColumn(column.getData());
	}

	/**
		Get a query that returns spot data for this bioassay. The query will
		select column, position and spot intensities by default. This method
		uses the {@link BioAssaySet#getSpotData()} method and adds a restriction
		to the returned query.

		@return A <code>DynamicSpotQuery</code> object
		@see BioAssaySet#getSpotData()
	*/
	public DynamicSpotQuery getSpotData()
	{
		DynamicSpotQuery query = getBioAssaySet().getSpotData();
		query.restrictPermanent(
			new DynamicRestriction(VirtualColumn.COLUMN, getDataCubeColumnNo())
		);
		return query;
	}
	
	
	/**
		Get a query for loading the parents bioassays to this bioassay. If
		this bioassay belong to a root bioassayset, null is returned.
		@return A query returning the parent bioassays of this bioassay, or null
	*/
	public ItemQuery<BioAssay> getParents()
	{		
		BioAssaySet bioAssaySet = getBioAssaySet().getTransformation().getSource();		
		ItemQuery<BioAssay> query = null;
		
		if (bioAssaySet != null)
		{
			query = bioAssaySet.getBioAssays();	
			query.joinPermanent(Hql.innerJoin("children", "c"));
			query.restrictPermanent(
				Restrictions.eq(
					Hql.alias("c"), 
					Hql.entity(this)
				)
			);
		}		
		return query;		
	}
	
	/**
		@return A query returning the raw bioassays this bioassay is created from
	*/
	public ItemQuery<RawBioAssay> getRawBioAssays()
	{
		ItemQuery<RawBioAssay> query = RawBioAssay.getQuery();
		query.joinPermanent(Hql.innerJoin("bioAssays", Item.BIOASSAY.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.BIOASSAY.getAlias()),
				Hql.entity(this)
			)
		);
		return query;
	}
}
