/*
	$Id: DataFileType.java 4594 2008-10-21 07:20:40Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.DataFileTypeData;
import net.sf.basedb.core.filehandler.DataFileMetadataReader;
import net.sf.basedb.core.filehandler.DataFileValidator;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.info.DataFileTypeInfo;
import net.sf.basedb.info.ToTransferable;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.JarClassLoader;

/**
	This class is used to represent types of files that are used
	to hold data instead of importing the data into into the database.
	<p>
	This class can define a {@link DataFileValidator} and a 
	{@link DataFileMetadataReader} that can be used to validate the if a
	file is actually of the correct type and, if it is, extract metadata from it.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2008-10-21 09:20:40 +0200 (Tue, 21 Oct 2008) $
*/
public class DataFileType
	extends BasicItem<DataFileTypeData>
	implements Nameable, Removable, ToTransferable<DataFileTypeInfo>, Validatable
{
	
	/**
		The type of item represented by this class.
		@see Item#DATAFILETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.DATAFILETYPE;

	/**
		The external ID for the file type representing
		Affymetrix CDF files.
	*/
	public static String AFFYMETRIX_CDF = "affymetrix.cdf";

	/**
		The external ID for the file type representing
		Affymetrix CEL files.
	*/
	public static String AFFYMETRIX_CEL = "affymetrix.cel";

	/**
		The external ID for the file type representing
		a generic raw data file of any type.
	*/
	public static String GENERIC_RAW_DATA = "generic.rawdata";

	/**
		The external ID for the file type representing
		a generic reporter map (array design layout) of any type.
	*/
	public static String GENERIC_REPORTER_MAP = "generic.reportermap";

	/**
		The external ID for the file type representing
		a generic print map (array design layout) of any type.
	*/
	public static String GENERIC_PRINT_MAP = "generic.printmap";

	/**
		The external ID for the file type representing
		generic analysed spot data.
		@since 2.8
	*/
	public static String GENERIC_SPOT_DATA = "generic.spotdata";

	/**
		The external ID for the file type representing
		generic analysed extra value data.
		@since 2.8
	*/
	public static String GENERIC_EXTRAVALUE_DATA = "generic.extravaluedata";

	/**
		The maximum length of the file extension that can be stored in the database.
		@see #setExtension(String)
	*/
	public static final int MAX_EXTENSION_LENGTH = DataFileTypeData.MAX_EXTENSION_LENGTH;

	/**
		The maximum length of the class names that can be stored in the database.
		@see #setValidatorClass(String)
		@see #setMetadataReaderClass(String)
	*/
	public static final int MAX_CLASSNAME_LENGTH = DataFileTypeData.MAX_CLASSNAME_LENGTH;

	/**
		The maximum length of the jar paths that can be stored in the database.
		@see #setValidatorJarPath(String)
		@see #setMetadataReaderJarPath(String)
	*/
	public static final int MAX_JARPATH_LENGTH = DataFileTypeData.MAX_JARPATH_LENGTH;
	
	/**
		Create a new data file type.
		
		@param dc The DbControl to use for database access
		@param externalId The external ID of the new type, must be 
			unique for all file types
		@param itemType The type of item that files of this type can
			be attached to. The item must implement the {@link FileStoreEnabled}
			interface
		@return The new file set member type object
	*/
	public static DataFileType getNew(DbControl dc, String externalId, Item itemType)
	{
		DataFileType type = dc.newItem(DataFileType.class);
		type.setExternalId(externalId);
		type.setName("New data file type");
		type.setItemType(itemType);
		return type;
	}

	/**
		Get a <code>DataFileType</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>DataFileType</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is another error
	*/
	public static DataFileType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		DataFileType type = dc.loadItem(DataFileType.class, id);
		if (type == null) throw new ItemNotFoundException("DataFileType[id="+id+"]");
		return type;
	}
	
	/**
		Get a <code>DataFileType</code> item when you know the external ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external ID of the item to load
		@return The <code>DataFileType</code> item.
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the item
		@throws BaseException If there is another error
	*/
	public static DataFileType getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_DATAFILETYPE_FOR_EXTERNAL_ID");
		/*
			SELECT dft
			FROM DataFileTypeData dft
			WHERE dft.externalId = :externalId
		*/
		query.setString("externalId", externalId);
		DataFileType dft = dc.getItem(DataFileType.class, HibernateUtil.loadData(DataFileTypeData.class, query));
		if (dft == null) throw new ItemNotFoundException("DataFileType[externalId="+externalId+"]");
		return dft;
	}
	
	/**
		Get a query configured to retrieve <code>DataFileType</code> items.
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
	*/
	public static ItemQuery<DataFileType> getQuery()
		throws BaseException
	{
		return new ItemQuery<DataFileType>(DataFileType.class);
	}
	
	/**
		Get a query configured to retrieve <code>DataFileType</code>:s
		that can be used on a given item. This method will check the
		{@link FileStoreEnabled#getPlatform()} and {@link FileStoreEnabled#getVariant()}
		values and restrict the query to only return items that can be used for
		that platform. If the item isn't attached to any platform or variant
		all data file types for the {@link Identifiable#getType()} are returned.
		
		@param item The item to get data file types for, or null
			to return all data file types
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<DataFileType> getQuery(FileStoreEnabled item)
	{
		ItemQuery<DataFileType> query = null;
		if (item == null)
		{
			query = getQuery();
		}
		else
		{
			query = getQuery(item.getPlatform(), item.getVariant(), item.getType(), null);
		}
		return query;		
	}

	/**
		Get a query to retrieve <code>DataFileType</code>:s that can be used
		on specific platform/variant for a given item and/or generic file type.
		All parameters are optional. If all are null the result is the same
		as {@link #getQuery()}.
		
		@param platform The platform, or null to not restrict by platform (unless
			a variant is given)
		@param variant The platform variant, or null to not restrict by variant
		@param itemType The type of item it should be possible to
			attach file to, or null to not restrict by item
		@param genericType The generic file type, or null to not restrict
			by file type
		@return A query returning data file types matching the specified
			restrictions
		@since 2.9
	 */
	public static ItemQuery<DataFileType> getQuery(Platform platform, PlatformVariant variant,
		Item itemType, FileType genericType)
	{
		ItemQuery<DataFileType> query = getQuery();
		// Restrict on itemType
		if (itemType != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("itemType"), 
					Expressions.integer(itemType.getValue())
				)
			);
		}
		
		// Restrict on generic file type
		if (genericType != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("genericType"), 
					Hql.entity(genericType))
			);
		}
		
		// Restrict on platform/variant
		if (variant != null && platform == null) platform = variant.getPlatform();
		if (platform != null)
		{
			query.join(Hql.innerJoin("platforms", "pft"));
			query.restrict(Restrictions.eq(Hql.property("pft", "platform"), Hql.entity(platform)));
			if (variant != null)
			{
				query.restrict(Restrictions.or(
					Restrictions.eq(Hql.property("pft", "variant"), null),
					Restrictions.eq(Hql.property("pft", "variant"), Hql.entity(variant))
				));
			}
			else
			{
				query.restrict(Restrictions.eq(Hql.property("pft", "variant"), null));
			}
		}
		return query;
	}
	
	
	/**
		Creates a new filetype item from the given data.
	
		@param data the data.
	*/
	DataFileType(DataFileTypeData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
	From the BasicItem class
	-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A file has been tagged with this type in a file set
		<li>A platform is using this file type
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILESETMEMBERS_FOR_DATAFILETYPE", "count(*)");
		/*
			SELECT {1}
			FROM FileSetMemberData mbr
			WHERE mbr.dataFileType = :dataFileType
		*/
		query.setEntity("dataFileType", this.getData());
		long count = HibernateUtil.loadData(Long.class, query);
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATFORMFILETYPE_FOR_DATAFILETYPE", "count(*)");
			/*
				SELECT {1}
				FROM PlatformFileTypeData pft
				WHERE pft.dataFileType = :dataFileType
			*/
			query.setEntity("dataFileType", this.getData());
			count = HibernateUtil.loadData(Long.class, query);
		}
		return count > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link FileSetMember}:s using this file type
		<li>{@link PlatformFileType}:s using this file type
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_FILESETMEMBERS_FOR_DATAFILETYPE", "mbr.id");
			/*
				SELECT {1}
				FROM FileSetMemberData mbr
				WHERE mbr.dataFileType = :dataFileType
			*/
		query.setEntity("dataFileType", this.getData());
		addUsingItems(using, Item.FILESETMEMBER, query);
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATFORMFILETYPE_FOR_DATAFILETYPE", "pft.id");
			/*
				SELECT {1}
				FROM PlatformFileTypeData pft
				WHERE pft.dataFileType = :dataFileType
			*/
		query.setEntity("dataFileType", this.getData());
		addUsingItems(using, Item.PLATFORMFILETYPE, query);
		return using;
	}
	
	

	/**
		Check that the validator and meta data reader classes implement the
		correct interfaces.
		@since 2.6
	*/
	@Override
	void validate() 
		throws InvalidDataException, BaseException
	{
		super.validate();
		getValidator();
		getMetadataReader();
	}

	/**
	 	@since 2.5
	 */
	public DataFileTypeInfo toTransferable(DataFileTypeInfo info)
	{
		super.toTransferable(info);
		info.setName(getName());
		info.setDescription(getDescription());
		info.setExtension(this.getExtension());
		info.setExternalId(this.getExternalId());
		return info;
	}
	// -------------------------------------------
	
	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = DataFileTypeData.MAX_EXTERNAL_ID_LENGTH;

	/**
		Get the external id for the data file type.
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	private void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		getData().setExternalId(StringUtil.setNullableString(
			externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
		Get the generic file type for this data file type.
		@return A <code>FileType</code> item
	*/
	public FileType getGenericType()
	{
		return getDbControl().getItem(FileType.class, getData().getGenericType());
	}
	
	/**
		Set the generic file type of this item. Null is not allowed.
		@param genericType The generic type
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
	*/
	public void setGenericType(FileType genericType)
	{
		checkPermission(Permission.WRITE);
		getData().setGenericType(genericType == null ? null : genericType.getData());
	}
	
	/**
		Get the common file extension for data files of this type.
		@return A string or null if there is no common file extension
	*/
	public String getExtension()
	{
		return getData().getExtension();
	}

	/**
		Set the most common gile extension for data file of this
		type.
		@param extension The extension or null
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws StringTooLongException If the string is longer that {@link #MAX_EXTENSION_LENGTH}.
	*/
	public void setExtension(String extension)
	{
		checkPermission(Permission.WRITE);
		getData().setExtension(StringUtil.setNullableString(extension, "extension", MAX_EXTENSION_LENGTH));
	}

	private void setItemType(Item itemType)
	{
		if (itemType == null) throw new InvalidUseOfNullException("itemType");
		if (!FileStoreEnabled.class.isAssignableFrom(itemType.getItemClass()))
		{
			throw new InvalidDataException(itemType + " isn't a subclass of FileStoreEnabled");
		}
		getData().setItemType(itemType.getValue());
	}
	
	/**
		Get the type of item files of this type can be used on. The
		item should implement the {@link FileStoreEnabled} interface.
		@return An <code>Item</code> object
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	/**
		Get the class name of the Java class that is used to validate the files of
		this type. The class must implement the {@link DataFileValidator} interface.
		@return A class name or null if no validator exists
	*/
	public String getValidatorClass()
	{
		return getData().getValidatorClass();
	}
	
	/**
		Set the class name of the Java class that is used to validate the files of
		this type. The class must implement the {@link DataFileValidator} interface.
		@param validatorClass The name of the class or null to disable validation
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws StringTooLongException If the string is longer that {@link #MAX_CLASSNAME_LENGTH}.
	*/
	public void setValidatorClass(String validatorClass)
	{
		checkPermission(Permission.WRITE);
		getData().setValidatorClass(StringUtil.setNullableString(validatorClass, 
				"validatorClass", MAX_CLASSNAME_LENGTH));
	}
	
	/**
		Get the path to the JAR file where the validator class is located. If the
		path is null the validator must be in the class path.
		@return The JAR path or null
		@since 2.6
	*/
	public String getValidatorJarPath()
	{
		return PluginDefinition.convertToAbsolute(getData().getValidatorJarPath());
	}
	
	/**
		Set the path to the JAR file where the validator class is located.
		@param validatorJarPath The path to the JAR file or null if the validator is in the class path
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws StringTooLongException If the string is longer that {@link #MAX_JARPATH_LENGTH}.
		@since 2.6
	*/
	public void setValidatorJarPath(String validatorJarPath)
	{
		checkPermission(Permission.WRITE);
		validatorJarPath = PluginDefinition.convertToRelative(validatorJarPath);
		getData().setValidatorJarPath(StringUtil.setNullableString(validatorJarPath, 
			"validatorJarPath", MAX_JARPATH_LENGTH));
	}
	
	
	/**
		Create a new instance of the {@link DataFileValidator} class specified
		by {@link #getValidatorClass()}.
		@return A validator instance or null if no class has been specified
		@throws BaseException If an instance of the class can't be created
	*/
	public DataFileValidator getValidator()
	{
		String validatorClass = getValidatorClass();
		if (validatorClass == null) return null;
		DataFileValidator validator = null;
		try
		{
			String jarPath = getValidatorJarPath();
			ClassLoader loader = jarPath == null ? null :JarClassLoader.getInstance(jarPath);
			validator = (DataFileValidator)ClassUtil.checkAndLoadClass(
					loader, validatorClass, true, DataFileValidator.class).newInstance();
		}
		catch (Throwable t)
		{
			throw new BaseException(t);
		}
		return validator;
	}
	
	/**
		Get the class name of the Java class that is used to extract metadata from files of
		this type. The class must implement the {@link DataFileMetadataReader} interface.
		@return A class name or null if no validator exists
	*/
	public String getMetadataReaderClass()
	{
		return getData().getMetadataReaderClass();
	}
	
	/**
		Set the class name of the Java class that is used to extract metadata from files of
		this type. The class must implement the {@link DataFileMetadataReader} interface.
		@param metadataReaderClass The name of the class or null to disable validation
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws StringTooLongException If the string is longer that {@link #MAX_CLASSNAME_LENGTH}.
	*/
	public void setMetadataReaderClass(String metadataReaderClass)
	{
		checkPermission(Permission.WRITE);
		getData().setMetadataReaderClass(StringUtil.setNullableString(metadataReaderClass, 
			"metadataReaderClass", MAX_CLASSNAME_LENGTH));
	}
	
	/**
		Get the path to the JAR file where the meta data reader class is located. If the
		path is null the reader must be in the class path.
		@return The JAR path or null
		@since 2.6
	*/
	public String getMetadataReaderJarPath()
	{
		return PluginDefinition.convertToAbsolute(getData().getMetadataReaderJarPath());
	}
	
	/**
		Set the path to the JAR file where the meta data reader class is located.
		@param metadataReaderJarPath The path to the JAR file or null if the reader is in the class path
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws StringTooLongException If the string is longer that {@link #MAX_JARPATH_LENGTH}.
		@since 2.6
	*/
	public void setMetadataReaderJarPath(String metadataReaderJarPath)
	{
		checkPermission(Permission.WRITE);
		metadataReaderJarPath = PluginDefinition.convertToRelative(metadataReaderJarPath);
		getData().setMetadataReaderJarPath(StringUtil.setNullableString(metadataReaderJarPath, 
			"metadataReaderJarPath", MAX_JARPATH_LENGTH));
	}	
	
	/**
		Create a new instance of the {@link DataFileMetadataReader} class specified
		by {@link #getMetadataReaderClass()}.
		@return A metadata reader instance or null if no class has been specified
		@throws BaseException If an instance of the class can't be created
	*/
	public DataFileMetadataReader getMetadataReader()
	{
		String readerClass = getMetadataReaderClass();
		if (readerClass == null) return null;
		DataFileMetadataReader reader = null;
		try
		{
			String jarPath = getMetadataReaderJarPath();
			ClassLoader loader = jarPath == null ? null : JarClassLoader.getInstance(jarPath);
			reader = (DataFileMetadataReader)ClassUtil.checkAndLoadClass(
					loader, readerClass, true, DataFileMetadataReader.class).newInstance();
		}
		catch (Throwable t)
		{
			throw new BaseException(t);
		}
		return reader;
	}

	/**
		Get a query that returns all platforms/variants that are using this
		file type.
		@return A query object
	*/
	public ItemQuery<PlatformFileType> getPlatforms()
	{
		ItemQuery<PlatformFileType> query = PlatformFileType.getQuery(null, null, false);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("dataFileType"),
				Hql.entity(this)
			)
		);
		return query;
	}
	
}
