/*
	$Id: DynamicPositionQuery.java 4517 2008-09-11 20:11:02Z jari $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.JoinType;

/**
	Represents a query for position information against the dynamic database. A single
	query object may query against data in one bioassayset at a time. Use factory methods 
	in the {@link Dynamic} class to join other tables and create restrictions based on
	the data in those tables. You may of course
	also use the regular expressions and restrictions from the {@link
	net.sf.basedb.core.query.Expressions}, {@link net.sf.basedb.core.query.Restrictions} 
	and other classes.
	<p>
	This type of query is executed by the {@link #iterate(DbControl)} method
	and returns the result as a {@link DynamicResultIterator}.
	
	@author Nicklas
	@version 2.0
	@see BioAssaySet#getPositionData()
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
public class DynamicPositionQuery
	extends DynamicQuery
{

	private Select[] defaultSelects;

	/**
		Create a new dynamic query.
		@param cube The cube to query against
	*/
	DynamicPositionQuery(DataCube cube)
	{
		super(cube, VirtualTable.POSITION);
	}
	
	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/
	/**
		Default selection: position
	*/
	Select[] getDefaultSelects()
	{
	
		if (defaultSelects == null)
		{
			defaultSelects = new Select[1];
			defaultSelects[0] = Dynamic.select(VirtualColumn.POSITION);
		}
		return defaultSelects;
	}
	// -------------------------------------------

	/**
		Join information about reporters in the query. Now it becomes possible
		to select reporter data or restrict the results based on reporter
		information.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If joinType is null. 
	 	@throws BaseException If anything else goes wrong. 
		@see Dynamic#selectReporter(String)
		@see Dynamic#reporter(String)
	*/
	public void joinReporters(JoinType joinType)
		throws InvalidDataException, BaseException
	{
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		join(new RealJoin(joinType, VirtualTable.POSITION, RealTable.REPORTERS, VirtualColumn.REPORTER_ID));
	}
	
	/**
		Join extra values in the query. Now it becomes possible to select
		those extra values or use them for restrictions.
		@param extraValue The extra values to join. Can not be null.
		@param joinType The type of join. Can not be null.
	 	@throws InvalidDataException If any of the required parameters are null. 
	 	@throws BaseException If something else fails.
		@see Dynamic#select(ExtraValue, String)
		@see Dynamic#extraValue(ExtraValue)
	*/
	public void joinExtraValue(ExtraValue extraValue, JoinType joinType)
		throws InvalidDataException, BaseException
	{
		if (extraValue == null) throw new InvalidUseOfNullException("extraValue");
		if (joinType == null) throw new InvalidUseOfNullException("joinType");
		if (extraValue.getCoordinateType() != ExtraValue.CoordinateType.POSITION)
		{
			throw new InvalidDataException("Extra value "+ extraValue+
				" is not attached to position: "+extraValue.getCoordinateType());
		}
		if (!getDataCube().equals(extraValue.getDataCubeExtraValue().getDataCube()))
		{
			throw new InvalidDataException("Extra value "+extraValue+
				" is not part of the same data cube as this query: " + getDataCube());
		}
		join(new ExtraValueJoin(extraValue, joinType));
	}

}
