/*
	$Id: ItemContext.java 4607 2008-10-28 08:16:02Z nicklas $

	Copyright (C) 2005, 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Order;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.data.ContextData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.ClientData;
import net.sf.basedb.util.jep.ChannelFunction;
import net.sf.basedb.util.jep.Jep;
import net.sf.basedb.util.jep.JepFunction;
import net.sf.basedb.util.jep.RawFunction;
import net.sf.basedb.util.jep.ReporterFunction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.EnumSet;
import java.util.List;

/**
	This item is a helper class for client applications when they need to store
	information about the current context for plugins that needs to find out the
	current context of the running client application. It is optional to use this 
	class and it is recommended that plugins and client applications that expects 
	to find context information also can handle situations without any context 
	information. Some of the settings in this objects are saved in the database,
	some not.
	<p>
	Typical example for using this class are:
	<p>
	
	<b>A client application storing list/table settings</b><br>
	<pre class="code">
SessionControl sc = ...
ItemContext current = sc.getCurrentContext(Item.SAMPLE);
current.setRowsPerPage(50);
current.setPage(0);
current.setSortProperty("name");
current.setSortDirection(SortDirection.DESC);
current.setPropertyFilter(
      new PropertyFilter("name", Operator.LIKE, "Nicklas%", Type.STRING)
   );
ItemQuery query = Sample.getQuery();
current.configureQuery(query, true);
// The query will now return the 50 first samples with a
// a name starting with 'Nicklas' sorted in descending order
</pre>
	
	<p>
	
	<b>A client application using filter information when creating new items</b><br>
	<pre class="code">
SessionControl sc = ...
DbControl dc = sc.newDbControl();
ItemContext current = sc.getCurrentContext(Item.SAMPLE);
Sample sample = Sample.getNew(dc);
if (current.getPropertyValue("name") != null)
{
   sample.setName(current.getPropertyValue("name"));
}
// The name is now 'Nicklas%'
</pre>

	<p>
	<b>Plugin finds out which items in a list has been selected</b><br>
	<pre class="code">
SessionControl sc = ...
ItemContext current = sc.getCurrentContext(Item.SAMPLE);
ItemQuery query = Sample.getQuery();
Set selected = current.getSelected();
if (selected != null && selected.size() > 0)
{
   query.restrict(
         Restrictions.in(Hql.property("id"), 
         Expressions.parameter("selected", selected, Type.INT))
      );
)
// The query will now return the selected items only
</pre>
	
	<p>
	<b>An import plugin asks client application for a file with a specific file type</b><br>
	<pre class="code">
public RequestInformation getRequestInformation(GuiContext context, String command)
   throws BaseException
{
   RequestInformation requestInformation = null;
   if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
   {
      requestInformation = getConfigurePluginParameters(context);
   }
   else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
   {
      requestInformation = getConfigureJobParameters(context);
      ItemContext fileContext = sc.getCurrentContext(Item.FILE);
      fileContext.setPropertyFilter(FileType.getPropertyFilter(FileType.REPORTER));
      fileContext.setPropertyFilter(new PropertyFilter("location", Operator.EQ, 
         Integer.toString(Location.PRIMARY.getValue()), Type.INT));
   }
   return requestInformation;
}
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-10-28 09:16:02 +0100 (Tue, 28 Oct 2008) $
	@see SessionControl#getCurrentContext(Item)
	@see SessionControl#saveCurrentContextAs(Item, String, boolean, boolean)
*/
public class ItemContext
{

	/**
		This is the name of the default context which is automatically saved
		when a user is logging out. A client application shouldn't use this name
		for other named contexts that it wants to reload in the future.
	*/
	public static final String DEFAULT_NAME = "- default -";

	// For parsing JEP expressions
	private static final JepFunction ch = new ChannelFunction();
	private static final JepFunction raw = new RawFunction(true);
	private static final JepFunction rep = new ReporterFunction();
	
	private final String name;
	private final Item itemType;
	private final String subContext;
	private int contextId;
	private int itemId;
	private Set<Integer> selected;
	private int rowsPerPage;
	private int page;
	private String sortProperty = null;
	private String message;
	private SortDirection sortDirection = null;
	private Set<Include> include;
	private Permission permission;
	private Map<String, PropertyFilter> propertyFilters;
	private Map<String, String> settings;
	private Map<String, Object> objects;
	private Query query;
	
	/**
		Create a new context.
		@param sortProperty The property used to sort a list
		@param sortDirection If the items should be sorted in ascending or descending order
		@param rowsPerPage The number of rows to display on a single page
		@param page The page number of the currently displayed page, starting at 0
	*/
	public ItemContext(String sortProperty, SortDirection sortDirection, int rowsPerPage, int page)
	{
		this.name = null;
		this.itemType = null;
		this.subContext = null;
		this.sortProperty = sortProperty;
		this.sortDirection = sortDirection;
		this.rowsPerPage = rowsPerPage;
		this.page = page;
	}
	
	/**
		Create a new empty context.
	*/
	ItemContext(Item itemType, String subContext, String name)
	{
		this.itemType = itemType;
		this.subContext = subContext;
		this.name = name;
		this.rowsPerPage = 30;
		this.page = 0;
		this.include = EnumSet.of(Include.MINE);
	}
	
	/**
		Create a new context from a loaded database object.
	*/
	ItemContext(ContextData context)
	{
		this.itemType = Item.fromValue(context.getItemType());
		this.subContext = context.getSubContext();
		this.name = context.getName();
		this.contextId = context.getId();
		this.itemId = context.getItemId();
		this.rowsPerPage = context.getRowsPerPage();
		this.page = context.getPage();
		this.sortProperty = context.getSortProperty();
		this.sortDirection = SortDirection.fromValue(context.getSortDirection());
		this.include = Include.fromInt(context.getInclude());
		for (Map.Entry<String, PropertyFilterData> entry : context.getPropertyFilters().entrySet())
		{
			setPropertyFilter(new PropertyFilter(entry.getKey(), entry.getValue()));
		}
		if (context.getSettings().size() != 0)
		{
			settings = new HashMap<String, String>(context.getSettings());
		}
	}
	
	/**
		Create a new database context object from this context. 
		@param name The name to use, or null if the current name should be used
		@param user The user that should own the context object
		@param client The client application currently in use
		@param isPublic If the context should be public or not
	*/
	ContextData getData(String name, UserData user, ClientData client, boolean isPublic)
	{
		ContextData context = new ContextData();
		context.setUser(user);
		context.setClient(client);
		context.setPublic(isPublic);
		context.setName(name == null ? this.name : name);
		context.setSubContext(subContext == null ? "" : subContext);
		context.setItemType(itemType.getValue());
		context.setItemId(itemId);
		context.setRowsPerPage(rowsPerPage);
		context.setPage(page);
		context.setSortProperty(sortProperty);
		if (sortDirection != null) context.setSortDirection(sortDirection.getValue());
		if (include != null) context.setInclude(Include.toInt(include));
		if (propertyFilters != null)
		{
			Map<String, PropertyFilterData> filterData = context.getPropertyFilters();
			for (PropertyFilter filter : propertyFilters.values())
			{
				filterData.put(filter.getProperty(), filter.getData());
			}
		}
		if (settings != null)
		{
			context.getSettings().putAll(settings);
		}
		return context;
	}

	/**
		Get the type of item this object is storing context for.
	*/
	public Item getItemType()
	{
		return itemType;
	}
	
	/**
		Get the sub-context name of this context.
	*/
	public String getSubContext()
	{
		return subContext;
	}

	/**
		Get the name of this context.
	*/
	public String getName()
	{
		return name;
	}

	/**
		Get the database ID of this context. Can be used to load contexts
		@return The database ID of the context, or 0 if it hasn't been saved to 
			the database
		@since 2.4
	*/
	public int getContextId()
	{
		return contextId;
	}
	
	/**
		Get the ID of the current item in this context.
		@return The ID of the item, or 0 if no item is current
	*/
	public int getId()
	{
		return itemId;
	}
	
	/**
		Set the ID of the current item in this context. This property is stored in 
		the database.
		@param id The ID of the item, or 0 if no item is current
	*/
	public void setId(int id)
	{
		this.itemId = id;
	}
	
	/**
		Get the number of rows the current table page is displaying.
		@return The number of rows
	*/
	public int getRowsPerPage()
	{
		return rowsPerPage;
	}
	
	/**
		Set the number of rows the current table page should show. 
		This property is stored in the database.
		@param rowsPerPage The number of rows to show on a page
	*/
	public void setRowsPerPage(int rowsPerPage)
	{
		this.rowsPerPage = rowsPerPage;
	}
	
	/**
		Get the current table page.
		@return The number of the page, starting at 0
	*/
	public int getPage()
	{
		return page;
	}
	
	/**
		Set the current table page.
		This property is stored in the database.
		@param page The number of the page, starting at 0
	*/
	public void setPage(int page)
	{
		this.page = page;
	}

	/**
		Get the name of the property the current listing is sorted by.
		If multiple properties are used for sorting they are separated by a comma.
	*/
	public String getSortProperty()
	{
		return sortProperty;
	}
	
	/**
		Set the name of the property that the current listing is sorted by.
		If you want to sort by multiple properties, separate them with a 
		comma.
		This property is stored in the database.
		@param sortProperty The name of the property
	*/
	public void setSortProperty(String sortProperty)
	{
		this.sortProperty = sortProperty;
	}

	/**
		If the items are sorted in ascending or descending order.
	*/
	public SortDirection getSortDirection()
	{
		return sortDirection;
	}
	
	/**
		Set the sort direction (ascending or descending) of the current listing.
		This property is stored in the database.
		@param sortDirection The sort direction
	*/
	public void setSortDirection(SortDirection sortDirection)
	{
		this.sortDirection = sortDirection;
	}

	// TODO - if multiple sort orders are used, this only returns the first one
	public Order getSortOrder()
	{
		Order sortOrder = null;
		String sortProperties = getSortProperty();
		if (sortProperties != null)
		{
			String[] sp = sortProperties.split(",");
			String sortProperty = sp[0];
			String alias = null;
			if (sortProperty.startsWith("$"))
			{
				// sortProperty is $alias.property or $alias only
				int dotIndex = sortProperty.indexOf(".");
				if (dotIndex == -1)
				{
					// alias only
					alias = sortProperty.substring(1);
					sortProperty = null;
				}
				else
				{
					// alias.property
					alias = sortProperty.substring(1, dotIndex);
					sortProperty = sortProperty.substring(dotIndex+1);
				}
			}
			Expression hqlSortby = Hql.property(alias, sortProperty);
			sortOrder = (SortDirection.ASC == getSortDirection() ? Orders.asc(hqlSortby) : Orders.desc(hqlSortby));
		}
		return sortOrder;
	}
	
	/**
		Get a message that should be displayed by a client application as
		soon as it is appropriate to do so. Typically, the {@link #configureQuery(EntityQuery, boolean)}
		sets an error message if it couldn't configure the query for some reason. 
	*/
	public String getMessage()
	{
		return message;
	}
	
	/**
		Set a message to be displayed by a client application as soon as
		it is appropriate to do so. Typically, the {@link #configureQuery(EntityQuery, boolean)}
		sets an error message if it couldn't configure the query for some reason. 
		This property is not stored in the database.
		@param message The message
	*/
	public void setMessage(String message)
	{
		this.message = message;
	}
	
	/**
		Get a <code>Set</code> object to specify {@link Include} options for 
		a query. This property is stored in the database.
		@see EntityQuery
	*/
	public Set<Include> getInclude()
	{
		if (include == null) include = EnumSet.noneOf(Include.class);
		return include;
	}
	
	/**
	 	Specify the permission the logged in user must have for items returned
	 	by a query. This property is not stored in the database.
		@see EntityQuery#setItemPermission(Permission)
	*/
	public void setItemPermission(Permission permission)
	{
		this.permission = permission;
	}
	
	/**
	 	The permission the logged in user must have for items returned
	 	by a query. This property is not stored in the database.
		@see EntityQuery#getItemPermission()
	*/
	public Permission getItemPermission()
	{
		return permission;
	}
	
	/**
		Get a <code>Set</code> object to specify the ID:s of all items that are
		currently selected. This property is not stored in the database.
	*/
	public Set<Integer> getSelected()
	{
		if (selected == null) selected = new HashSet<Integer>();
		return selected;
	}
	
	/**
		Set a property filter. If a filter for the same property already exists it
		is replaced by the new filter.
		This property is stored in the database.
		@param filter The new property filter
	*/
	public void setPropertyFilter(PropertyFilter filter)
	{
		if (propertyFilters == null)
		{
			propertyFilters = new HashMap<String, PropertyFilter>();
		}
		propertyFilters.put(filter.getProperty(), filter);
	}
	
	/**
		Get a filter for a property.
		@param property The name of the property
		@return A <code>PropertyFilter</code> object or
			null if no filter exists
	*/
	public PropertyFilter getPropertyFilter(String property)
	{
		PropertyFilter filter = null;
		if (propertyFilters != null)
		{
			filter = propertyFilters.get(property);
		}
		return filter;
	}
	
	/**
		Get a collection of all property filters stored in this context.
		@return A Collection containsing PropertyFilter objects or null
	*/
	public Collection<PropertyFilter> getPropertyFilters()
	{
		return propertyFilters == null ? null : propertyFilters.values();
	}
	
	/**
		Get the value of a property filter. It is a shortcut
		for <code>getPropertyFilter(property).getValue()</code>
		but also handles the case of a non-existing filter for the
		specfied property.
		@param property The name of the property
		@return The value as a string
	*/
	public String getPropertyValue(String property)
	{
		String value = null;
		PropertyFilter filter = getPropertyFilter(property);
		if (filter != null)
		{
			value = filter.getValue();
		}
		return value;
	}
	
	/**
		Remove a property filter.
		@param property The name of the property
	*/
	public void removePropertyFilter(String property)
	{
		if (propertyFilters != null)
		{
			propertyFilters.remove(property);
		}
	}
	
	public Object getPropertyObject(String property)
	{
		Object value = null;
		PropertyFilter filter = getPropertyFilter(property);
		if (filter != null)
		{
			value = filter.getValueAsObject();
		}
		return value;
	}
	
	/**
		Remove all property filter from this context.
	*/
	public void removeAllPropertyFilters()
	{
		if (propertyFilters != null)
		{
			propertyFilters.clear();
		}
	}
	
	/**
		Add an item to the "recently used items" list. Each type of item
		has their own list which may contain at most <code>maxInList</code>
		items. 
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the item type. For example the recently used
		protocols are found in the setting: 
		<code>getSetting("PROTOCOL.recent")</code>. Client code is advised to
		not change these settings.
		
		@param item The item to add
		@param maxInList The maximum number of items in the list
	 */
	public void setRecent(BasicItem item, int maxInList)
	{
		setRecent(item.getType().name(), Integer.toString(item.getId()), maxInList);
	}
	
	/**
		Add an item to the "recently used items" list. Each type of item
		has their own list and sublists which may contain at 
		most <code>maxInList</code> items. 
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the item type and sublist name. For example the 
		recently used protocols are found in the setting: 
		<code>getSetting("PROTOCOL.subList.recent")</code>. Client code is advised to
		not change these settings.

		@param item The item to add
		@param subList The sublist
		@param maxInList The maximum number of items in the list
		@since 2.5
	*/
	public void setRecent(BasicItem item, String subList, int maxInList)
	{
		setRecent(item.getType().name() + "." + subList, Integer.toString(item.getId()), maxInList);
	}
	
	/**
		Add a value to the list of recently used values for the specified key.
		Each key has it's own list of values which may contain at most 
		<code>maxInList</code> values. If the value is already in the list
		it is moved to the front of the list.
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the key. For example, if key is "RAWDATATYPE",
		the recently used values are found in the setting: 
		<code>getSetting("RAWDATATYPE.recent")</code>. Client code is advised to
		not change these settings.

		@param key The key to the list
		@param value The value to add
		@param maxInList The maximum number of values in the list
	*/
	public void setRecent(String key, String value, int maxInList)
	{
		List<String> recent = getRecent(key);
		int existingIndex = recent.indexOf(value);
		if (existingIndex >= 0) recent.remove(existingIndex);
		recent.add(0, value);
		if (recent.size() > maxInList) recent = recent.subList(0, maxInList);
		setRecent(key, recent);
	}
	
	/**
		Get the number of recently used items in the list.
		@param itemType The type of items
	*/
	public int getNumRecent(Item itemType)
	{
		return getRecent(itemType.name()).size();
	}
	
	/**
		Get the number of recently used items in a list with a sublist.
		@param itemType The type of items
		@param subList The name of the sublist
		@since 2.5
	*/
	public int getNumRecent(Item itemType, String subList)
	{
		return getRecent(itemType.name() + "." + subList).size();
	}
	
	/**
		Get the number of recently used value in the list.
		@param key The key to the list
	*/
	public int getNumRecent(String key)
	{
		return getRecent(key).size();
	}
	
	/**
		Get all recently used items of the specified type. The list
		only contains items that the logged in user has read permission
		to and that hasn't been deleted. Thus, the {@link #getNumRecent(Item)}
		may return a larger number than what is actually returned in this
		list.
		
		@param dc The DbControl used to access the database
		@param itemType The type of items
		@return A list with the items
	*/
	public List<? extends BasicItem> getRecent(DbControl dc, Item itemType)
	{
		return loadRecent(dc, itemType, itemType.name());
	}
	
	/**
		Get all recently used items of the specified type in a given sublist.
		The list only contains items that the logged in user has read permission
		to and that hasn't been deleted. Thus, the {@link #getNumRecent(Item)}
		may return a larger number than what is actually returned in this
		list.
		
		@param dc The DbControl used to access the database
		@param itemType The type of items
		@param subList The name of the sublist
		@return A list with the items
		@since 2.5
	*/
	public List<? extends BasicItem> getRecent(DbControl dc, Item itemType, String subList)
	{
		return loadRecent(dc, itemType, itemType.name() + "." + subList);
	}
	
	private List<? extends BasicItem> loadRecent(DbControl dc, Item itemType, String key)
	{
		List<String> recentIds = getRecent(key);
		List<BasicItem> recent = new ArrayList<BasicItem>(recentIds.size());
		for (String id : recentIds)
		{
			try
			{
				recent.add(itemType.getById(dc, Integer.parseInt(id)));
			}
			catch (Throwable t)
			{
				Application.getLogger().debug("Failed loading recent " + 
					itemType.name()+" [id="+id+"]", t);
			}
		}
		return recent;
	}
	
	
	public BasicItem getRecent(DbControl dc, Item itemType, int index)
	{
		return loadRecent(dc, itemType, itemType.name(), index);
	}
	
	public BasicItem getRecent(DbControl dc, Item itemType, String subList, int index)
	{
		return loadRecent(dc, itemType, itemType.name() + "." + subList, index);
	}
	
	private BasicItem loadRecent(DbControl dc, Item itemType, String key, int index)
	{
		List<String> recentIds = getRecent(key);
		BasicItem item = null;
		if (index >= 0 && index < recentIds.size())
		{
			String id = recentIds.get(index);
			try
			{
				item = itemType.getById(dc, Integer.parseInt(id));
			}
			catch (Throwable t)
			{
				Application.getLogger().debug("Failed loading recent " + 
					itemType.name()+" [id="+id+"]", t);
			}
			
		}
		return item;
	}

	
	/**
		Get all recently used values for the specified key. Changes 
		to the list are not recorded in the database. Use {@link 
		#setRecent(String, String, int)} instead.
		@param key The key to the list
		@return A list containing the values or an empty list if
			no recently used value has been stored
	*/
	public List<String> getRecent(String key)
	{
		String recent = getSetting(getRecentSettingName(key));
		if (recent == null)
		{
			return new ArrayList<String>();
		}
		else
		{
			return new ArrayList<String>(Arrays.asList(recent.split(":")));
		}
	}
	
	public String getRecent(String key, int index)
	{
		List<String> recent = getRecent(key);
		return index >= 0 && index < recent.size() ? recent.get(index) : null;
	}
	
	/**
		Clear the recently used items list from items of the specified type.
		@param itemType The type of items
	*/
	public void clearRecent(Item itemType)
	{
		setRecent(itemType.name(), null);
	}
	
	/**
		Clear the recently used items sublist from items of the specified type.
		@param itemType The type of items
		@param subList The name of the sublist
	*/
	public void clearRecent(Item itemType, String subList)
	{
		setRecent(itemType.name() + "." + subList, null);
	}
	
	/**
		Clear the recently used values list.
		@param key The key to the list
	*/
	public void clearRecent(String key)
	{
		setRecent(key, null);
	}
	
	/**
		Clear all recently used values lists.
	*/
	public void clearAllRecent()
	{
		if (settings != null)
		{
			Iterator<Map.Entry<String, String>> it = settings.entrySet().iterator();
			while (it.hasNext())
			{
				String key = it.next().getKey();
				if (key.endsWith(".recent"))
				{
					it.remove();
				}
			}
		}
	}
	
	private String getRecentSettingName(String key)
	{
		return key+".recent";
	}

	private void setRecent(String key, List<String> recent)
	{
		setSetting(getRecentSettingName(key), 
			recent == null ? null : StringUtil.join(recent, ":", true));
	}
	
	/**
		Get a value.
		@param name The name of the value
		@return The current value or null if no value exists
		@see #setSetting(String, String)
	*/
	public String getSetting(String name)
	{
		return settings == null ? null : settings.get(name);
	}
	
	/**
		Set a value. This property is stored in the database.
		@param name The name of the value
		@param value The value
		@return The old value or null if no old value existed.
		@see #getSetting(String)
	*/
	public String setSetting(String name, String value)
	{
		if (settings == null) settings = new HashMap<String, String>();
		return settings.put(name, value);
	}
	
	/**
		Remove a value.
		@param name The name of the value
		@return The old value or null if no old value existed.
	*/
	public String removeSetting(String name)
	{
		return settings == null ? null : settings.remove(name);
	}
	
	/**
		Get an object.
		@param name The name of the value
		@return The current value or null if no value exists
		@see #setObject(String, Object)
	*/
	public Object getObject(String name)
	{
		return objects == null ? null : objects.get(name);
	}
	
	/**
		Set a value. This property is not stored in the database.
		@param name The name of the value
		@param value The value
		@return The old value or null if no old value existed.
		@see #getObject(String)
	*/
	public Object setObject(String name, Object value)
	{
		if (objects == null) objects = new HashMap<String, Object>();
		return objects.put(name, value);
	}
	
	/**
		Remove a value.
		@param name The name of the value
		@return The old value or null if no old value existed.
	*/
	public Object removeObject(String name)
	{
		return objects == null ? null : objects.remove(name);
	}

	public Query getQuery()
	{
		return query;
	}
	
	public void setQuery(Query query)
	{
		this.query = query;
	}
	
	/**
		This method will fail if used to filter on reporter lists or other items
		that require a DbControl.
		@deprecated Use {@link #configureQuery(DbControl, EntityQuery, boolean)} instead
	*/
	public void configureQuery(EntityQuery query, boolean autoLeftJoin)
		throws BaseException
	{
		configureQuery(null, query, autoLeftJoin);
	}

	/**
		Use the settings in this context to configure a query. 
		<ul>
		<li>{@link EntityQuery#setReturnTotalCount(boolean)} is set to true.
		<li>{@link EntityQuery#setFirstResult(int)} and {@link EntityQuery#setMaxResults(int)} 
			is set using {@link #getPage()} and {@link #getRowsPerPage()} settings.
			
		<li>The {@link #getSortProperty()} and {@link #getSortDirection()} is
			used to sort the the query. If the sort property starts with a dollar ($)
			sign it is interpreted as a <code>alias.property</code> value, ignoring
			the root alias of the query. The sort property may contain multiple properties
			if they are separated by a comma.
			
		<li>{@link #getInclude()} is used to set {@link EntityQuery#include(Include[])}
			or @link EntityQuery#exclude(Include[])} options.
			
		<li>All property filters are added as {@link Restriction}:s to the 
			query.
			
		</ul>
		
		@param dc The DbControl that will be used to execute the query
		@param query The query to configure
		@param autoLeftJoin If true, and the sort property and filters are checked
			for associations to other items, which are automatically left joined to
			the query (instead of the hibernate default of inner join) (this feature
			is experimental)
	 	@throws BaseException If configuring the query fails.
	 	@since 2.8
	*/
	public void configureQuery(DbControl dc, EntityQuery query, boolean autoLeftJoin)
		throws BaseException
	{
		// Add page information
		query.setReturnTotalCount(true);
		query.setFirstResult(getPage() * getRowsPerPage());
		query.setMaxResults(getRowsPerPage());

		// Temporary storage for all associations that must be left joined
		Set<String> leftJoins = autoLeftJoin ? new HashSet<String>() : null;
		Set<String> sortedProperties = autoLeftJoin ? new HashSet<String>() : null;
		
		// Add sort order information
		String sortProperties = getSortProperty();
		if (sortProperties != null)
		{
			String[] sp = sortProperties.split(",");
			for (String sortProperty : sp)
			{
				String alias = null;
				if (sortProperty.startsWith("$"))
				{
					// sortProperty is $alias.property or $alias only
					int dotIndex = sortProperty.indexOf(".");
					if (dotIndex == -1)
					{
						// alias only
						alias = sortProperty.substring(1);
						sortProperty = null;
					}
					else
					{
						// alias.property
						alias = sortProperty.substring(1, dotIndex);
						sortProperty = sortProperty.substring(dotIndex+1);
					}
				}
				// left join if sort property is association 
				if (autoLeftJoin)
				{
					int dotIndex = sortProperty.lastIndexOf(".");
					if (dotIndex >= 0)
					{
						String toJoin = sortProperty.substring(0, dotIndex);
						leftJoins.add(toJoin);
						sortedProperties.add(toJoin);
					}
				}
				Expression hqlSortby = Hql.property(alias, sortProperty);
				query.order(SortDirection.ASC == getSortDirection() ? Orders.asc(hqlSortby) : Orders.desc(hqlSortby));
			}
		}
		
		// Add include/exclude information
		Set<Include> includes = getInclude();
		for (Include i : Include.values())
		{
			if (includes.contains(i)) 
			{
				query.include(i);
			}
			else
			{
				query.exclude(i);
			}
		}
		
		// Permission info
		if (permission != null) query.setItemPermission(permission);
		
		// Add property filters
		if (propertyFilters != null)
		{
			Restriction[] restrictions = new Restriction[propertyFilters.values().size()];
			int i = 0;
			for (PropertyFilter filter : propertyFilters.values())
			{
				try
				{
					Restriction r = filter.getRestriction(dc, query);
					if (r != null)
					{
						restrictions[i] = r;
						i++;
						// left join if filter property is association 
						if (autoLeftJoin)
						{
							String filterProperty = filter.getProperty();
							int dotIndex = filterProperty == null ? -1 : filterProperty.lastIndexOf(".");
							if (dotIndex >= 0 && !filterProperty.startsWith("$") && !filterProperty.startsWith("") && !filterProperty.startsWith("&"))
							{
								leftJoins.add(filterProperty.substring(0, dotIndex));
							}
						}
					}
				}
				catch (Throwable ex)
				{
					Application.getLogger().warn("Could not filter on '" + 
						getItemType() + "." + filter.getProperty() + "'", ex);
					setMessage(ex.getClass().getSimpleName() + ": " + ex.getMessage());
				}
			}
			if (i > 0)
			{
				Restriction[] r2 = new Restriction[i];
				System.arraycopy(restrictions, 0, r2, 0, i);
				query.restrict(Restrictions.and(r2));
			}
		}
		
		// Add left joins
		if (autoLeftJoin)
		{
			int i = 0;
			// Use fetch joins for columns appearing in the ORDER BY clause
			// if the query is distinct and the database requires it
			boolean selectOrderBy = query.isDistinct() && 
				HibernateUtil.getDbEngine().selectOrderByColumnsIfDistinct();
			for (String property : leftJoins)
			{
				boolean fetch = selectOrderBy && sortedProperties.contains(property);
				if (property.startsWith("@") || property.startsWith("") || property.startsWith("&")) 
				{
					property = property.substring(1);
				}
				i++;
				query.join(Hql.leftJoin(null, property, "alj"+i, null, fetch));
			}
		}
	}
	
	/**
		Copy settings from another context to this context. Only settings which doesn't 
		have been set in this context is copied from the other context, unless
		the <code>forceAll</code> parameter is set.
	*/
	void copyFrom(ItemContext other, boolean forceAll)
	{
		if (this.rowsPerPage == -1 || forceAll) 
		{
			this.rowsPerPage = other.rowsPerPage;
		}
		if (this.page == -1 || forceAll) 
		{
			this.page = other.page;
		}
		if (this.sortProperty == null || forceAll) 
		{
			this.sortProperty = other.sortProperty;
		}
		if (this.sortDirection == null || forceAll) 
		{
			this.sortDirection = other.sortDirection;
		}
		if ((this.include == null  || forceAll) && other.include != null) 
		{
			this.getInclude().addAll(other.include);
		}
		if ((this.settings == null || forceAll) && other.settings != null) 
		{
			this.settings = new HashMap<String, String>(other.settings);
		}
		if ((this.objects == null || forceAll) && other.objects != null) 
		{
			this.objects = new HashMap<String, Object>(other.objects);
		}
	}
	
	/**
		Use the settings in this context to configure a dynamic query. 
		<ul>
		<li>{@link DynamicQuery#setReturnTotalCount(boolean)} is set to true.
		<li>{@link DynamicQuery#setFirstResult(int)} and {@link DynamicQuery#setMaxResults(int)} 
			is set using {@link #getPage()} and {@link #getRowsPerPage()} settings.
		</ul>
		
		@param dc A DbControl to use if the database needs to be accessed
		@param query The query to configure
		@param selectionList A list of expression that we want to select 
			from the query
	 	@throws BaseException If the configuring the query fails. 
	*/
	public void configureQuery(DbControl dc, DynamicQuery query, List<String> selectionList)
		throws BaseException
	{
	
		// Add page information
		query.setReturnTotalCount(true);
		query.setFirstResult(getPage() * getRowsPerPage());
		query.setMaxResults(getRowsPerPage());
		
		// Add selection elements
		if (selectionList != null)
		{
			for (String selectProperty : selectionList)
			{
				Select select = getDynamicSelect(dc, selectProperty); //, joins);
				if (select != null)
				{
					query.select(select);
				}
			}
		}
		
		// Add sort order information
		String sortProperties = getSortProperty();
		if (sortProperties != null)
		{
			boolean selectSortProperties = 
				query.isDistinct() && HibernateUtil.getDbEngine().selectOrderByColumnsIfDistinct();
			int sortIndex = 0;
			String[] sp = sortProperties.split(",");
			for (String sortProperty : sp)
			{
				Expression sortby = getDynamicExpression(dc, sortProperty);
				if (sortby != null)
				{
					query.order(SortDirection.ASC == getSortDirection() ? Orders.asc(sortby) : Orders.desc(sortby));
					if (selectionList == null || !selectionList.contains(sortProperty))
					{
						// We must add the sort property to the selection list
						query.select(Selects.expression(sortby, "sorted" + sortIndex));
					}
				}
				sortIndex++;
			}
		}

		// Add property filters
		if (propertyFilters != null)
		{
			Restriction[] restrictions = new Restriction[propertyFilters.values().size()];
			int i = 0;
			for (PropertyFilter filter : propertyFilters.values())
			{
				try
				{
					Restriction r = filter.getDynamicRestriction(dc);
					if (r != null)
					{
						restrictions[i] = r;
						i++;
					}
				}
				catch (Throwable ex)
				{
					setMessage(ex.getClass().getSimpleName() + ": " + ex.getMessage());
				}
			}
			if (i > 0)
			{
				Restriction[] r2 = new Restriction[i];
				System.arraycopy(restrictions, 0, r2, 0, i);
				query.restrict(Restrictions.and(r2));
			}
		}
	}
	
	/**
		Create an {@link Expression} from a string. If the string starts with:
		
		<ul>
		<li>%: The rest of the string is the channel number. We use the 
			{@link Dynamic#column(VirtualColumn)} to create the expression
			for {@link VirtualColumn#channel(int)}

		<li>$: The rest of the string is a raw data property name. We use
			the {@link Dynamic#rawData(String)} to create the expression.
			
		<li>@: The rest of the string is a reporter property name. We use
			the {@link Dynamic#reporter(String)} to create the expression.
			
		<li>#: The rest of the string is the id of an extra value. 
			We use the {@link Dynamic#extraValue(ExtraValue)} to create the expression and
			{@link DynamicSpotQuery#joinExtraValue(ExtraValue, JoinType)} to automatically
			join the extra values.
			
		<li>=: The rest of the string is JEP expression. It is converted
			to an Expression by the {@link Jep#formulaToExpression(String, JepFunction[])}
			method. 
		</ul>
		
		It also recognises the following strings for columns to use by
		the {@link Dynamic#column(VirtualColumn)} method:
		<ul>
		<li>POSITION: {@link VirtualColumn#POSITION}
		<li>COLUMN: {@link VirtualColumn#COLUMN}
		</ul>

		@param dc A DbControl to use if the database needs to be accessed
		@param propertyDef The string to use for creating the expression.
		@return An expression object
	*/
	public static Expression getDynamicExpression(DbControl dc, String propertyDef)
	{
		Expression e = null;
		if (propertyDef.startsWith("%"))
		{
			// propertyDef is %channel_number
			int channel = Integer.parseInt(propertyDef.substring(1));
			e = Dynamic.column(VirtualColumn.channel(channel));
		}
		else if (propertyDef.startsWith("$"))
		{
			// propertyDef is $rawDataPropertyName
			e = Dynamic.rawData(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("@"))
		{
			// propertyDef is @reporterPropertyName
			e = Dynamic.reporter(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("#"))
		{
			// propertyDef id of extra value: #id
			int extraId = net.sf.basedb.util.Values.getInt(propertyDef.substring(1));
			try
			{
				ExtraValue ev = ExtraValue.getById(dc, extraId);
				e = Dynamic.extraValue(ev);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		else if (propertyDef.startsWith("="))
		{
			// propertyDef is JEP expression
			e = Jep.formulaToExpression(propertyDef.substring(1), ch, raw, rep);
		}
		else
		{
			// propertyDef is a virtual column
			VirtualColumn column = null;
			if ("POSITION".equals(propertyDef))
			{
				column = VirtualColumn.POSITION;
			}
			else if ("COLUMN".equals(propertyDef))
			{
				column = VirtualColumn.COLUMN;
			}
			else
			{
				// TODO - more virtual columns
			}
			if (column != null) e = Dynamic.column(column);
		}
		return e;
	}
	
	/**
		Same as {@link #getDynamicExpression(DbControl, String)} but generates
		a select object instead.
	 	@param dc DbControl used to access the database 
	 	@param propertyDef The string to use for creating the expression.
	 	@return An Expression object.
	*/
	public static Select getDynamicSelect(DbControl dc, String propertyDef)
	{
		Select s = null;
		if (propertyDef.startsWith("%"))
		{
			// propertyDef is $channel_number
			int channel = Integer.parseInt(propertyDef.substring(1));
			s = Dynamic.select(VirtualColumn.channel(channel));
		}
		else if (propertyDef.startsWith("$"))
		{
			// propertyDef is $rawDataPropertyName
			s = Dynamic.selectRawData(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("@"))
		{
			// propertyDef is @reporterPropertyName
			s = Dynamic.selectReporter(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("#"))
		{
			// propertyDef id of extra value: #id
			int extraId = net.sf.basedb.util.Values.getInt(propertyDef.substring(1));
			try
			{
				ExtraValue ev = ExtraValue.getById(dc, extraId);
				s = Dynamic.select(ev, "xtra" + extraId);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		else if (propertyDef.startsWith("="))
		{
			// propertyDef is JEP expression
			Expression e = Jep.formulaToExpression(propertyDef.substring(1), ch, raw, rep);
			s = Selects.expression(e, "jep" + System.identityHashCode(e));
		}
		else
		{
			// propertyDef is a virtual column
			VirtualColumn column = null;
			if ("POSITION".equals(propertyDef))
			{
				column = VirtualColumn.POSITION;
			}
			else if ("COLUMN".equals(propertyDef))
			{
				column = VirtualColumn.COLUMN;
			}
			else
			{
				// TODO - more virtual columns
			}
			if (column != null) s = Dynamic.select(column);
		}
		return s;
	}
	
	/**
		Enumeration used to specify the sort direction.
	*/
	public enum SortDirection
	{
		/**
			Items are sorted in ascending order.
		*/
		ASC(0, "ascending"),
		/**
			Items are sorted in descending order.
		*/
		DESC(1, "descending");
	
		/**
			Get the <code>SortDirection</code> object when you know the integer code.
			
			@param value The integer value.
			@return The location for the integer value.
		*/
		public static SortDirection fromValue(int value)
		{
			SortDirection sd = valueMapping.get(value);
			assert sd != null : "sortDirection == null for value "+value;
			return sd;
		}
		
		/**
			Maps an integer to a sort direction.
		*/
		private static final Map<Integer, SortDirection> valueMapping = new HashMap<Integer, SortDirection>();
		
		static
		{
			for (SortDirection sortDirection : SortDirection.values())
			{
				SortDirection sd = valueMapping.put(sortDirection.getValue(), sortDirection);
				assert sd == null : "Another direction with the value "+sortDirection.getValue()+" already exists";
			}
		}
		
		/**
			The integer value of this direction. 
		*/
		private final int value;
		/**
			The string value of this direction.
		*/
		private final String displayValue;
	
		/**
			Creates a new direction. Used internal only.
			
			@param value The integer value of the direction
			@param displayValue Returned by the toString method
		*/
		private SortDirection(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing an location to the database.
			
			@return The integer value for this location.
		*/
		public int getValue()
		{
			return value;
		}
	}

}
