/*
	$Id: Ownable.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2005, 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

/**
	An <code>Ownable</code> item is an item which has a {@link User} as its owner.
	The {@link OwnableUtil} class provides methods that includes data validation 
	to make it easy to implement this interface.
	<p>

	<b>Reference implementation</b><br>
	<pre class="code">
public User getOwner()
   throws PermissionDeniedException, BaseException
{
   return getDbControl().getItem(User.class, getData().getOwner());
}
public void setOwner(User owner)
   throws PermissionDeniedException, InvalidDataException
{
   checkPermission(Permission.SET_OWNER);
   OwnableUtil.setOwner(getData(), owner);
}
</pre>

	@author Nicklas
	@version 2.0
	@see OwnedItem
	@see OwnableUtil
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
public interface Ownable
	extends Identifiable
{
	/**
		Get the {@link User} that is the owner of the item.
		@return The owner of the item
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the owner
		@throws BaseException If there is another error
	*/
	public User getOwner()
		throws PermissionDeniedException, BaseException;

	/**
		Change the owner of the item.
		@param owner The new owner of the item
		@throws PermissionDeniedException If the logged in user doesn't 
			have permission to change the owner on the item
		@throws InvalidDataException If the owner is null
		@see #takeOwnership()
	*/
	public void setOwner(User owner)
		throws PermissionDeniedException, InvalidDataException;

	/**
		Take ownership of the item.
		@throws PermissionDeniedException If the logged in user doesn't have
			permission to take ownership
		@since 2.2
	*/
	public void takeOwnership()
		throws PermissionDeniedException;
	
	/**
		Check if the logged in user is the owner of this item or not.
		@return TRUE if the logged in user is the owner, FALSE otherwise
		@since 2.2
	*/
	public boolean isOwner();
	
}
