/*
	$Id: ReporterType.java 4696 2008-12-09 13:55:06Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.ReporterTypeData;

/**
	This class holds reporter type information. A reporter may have,
	but is not required to have a type. The main reason for a
	reporter to have a type is to make it easier for client
	application to filter the reporter list.

	@author Nicklas, Samuel
	@version 2.0
	@see ReporterData
*/
public class ReporterType
	extends BasicItem<ReporterTypeData>
	implements Nameable, Removable, Registered
{
	/**
		The type of item represented by this class.
		@see Item#REPORTERTYPE
		@see #getType()
	*/
	private static final Item TYPE = Item.REPORTERTYPE;
	
	/**
		Create a new <code>ReporterType</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@return The new <code>ReporterType</code> item.
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#CREATE CREATE} permission for this type
			of item.
		@throws BaseException If there is another error.
	*/
	public static ReporterType getNew(DbControl dc)
		throws PermissionDeniedException, BaseException
	{
		ReporterType rt = dc.newItem(ReporterType.class);
		rt.setName("New reporter type");
		rt.getData().setEntryDate(new Date());
		return rt;
	}

	/**
		Get a <code>ReporterType</code> object when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load.
		@return The <code>ReporterType</code> item.
		@throws ItemNotFoundException This exception is thrown if an
			item with the specified ID is not found.
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#READ READ}
			permission to the items.
		@throws BaseException This exception is thrown if there is another
			error.
	*/
	public static ReporterType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ReporterType rt = dc.loadItem(ReporterType.class, id);
		if (rt == null) throw new ItemNotFoundException("ReporterType[ID="+id+"]");
		return rt;
	}
	
	/**
		Get a query configured to retrieve reporter types.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<ReporterType> getQuery()
	{
		return new ItemQuery<ReporterType>(ReporterType.class);
	}

	/**
		Constructor that sets reporter type data.
		@param data A {@link ReporterTypeData} object.
	*/
	ReporterType(ReporterTypeData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	/**
		Get the type of item represented by the object. The returned
		value is one of the values defined in the {@link Item} enumeration.
		@return A value indicating the type of item.
	*/
	public Item getType()
	{
		return TYPE;
	}
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if any reporters of this type exists.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_REPORTERS_FOR_REPORTERTYPE", "count(*)");
		/*
			SELECT {1}
			FROM ReporterData rpt
			WHERE rpt.reporterType = :reporterType
		*/
		query.setEntity("reporterType", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Reporter}:s of this type
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_REPORTERS_FOR_REPORTERTYPE", "rpt.id");
			/*
				SELECT {1}
				FROM ReporterData rpt
				WHERE rpt.reporterType = :reporterType
			*/
		query.setEntity("reporterType", this.getData());
		addUsingItems(using, Item.REPORTER, query);
		return using;
	}
	// -------------------------------------------

	/**
		Create a new {@link ReporterData} object of this type.
		@return The new <code>ReporterData</code> object
		@param externalId The external id of the new reporter. Null is not allowed
		@throws PermissionDeniedException This exception is thrown if
			user doesn't have use permission
			for this reporter type.
	 	@throws InvalidDataException If required argument is null.
	 	@throws BaseException If there is another error.
	*/
	public ReporterData newReporter(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		ReporterData rd = Reporter.getNew(externalId);
		Reporter.setReporterType(rd, this);
		return rd;
	}
}
