/*
	$Id: Role.java 4696 2008-12-09 13:55:06Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.RoleData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.EntityQuery;

import java.util.Date;
import java.util.Set;
import java.util.Collections;

/**
	This class is used to represent roles. A role is used in the
	permission system to give users access to various parts
	of BASE. A permission given to a role is global, ie. it applies to
	all items of a spcific type on the entire server, and not only within 
	a group or project. For example it is possible to give <code>READ</code> 
	access to all <code>SAMPLE</code>:s, no matter if the owner has shared them
	to other users or not.
	<p>
	BASE comes with a predefined set of roles, for example
	{@link #ADMINISTRATOR} and {@link #GUEST}, which have been
	configured with what we think is an appropriate combination of
	privileges. If you wish, you may create more roles. Use roles
	only for functional grouping of the users, and not
	for organisational grouping. If you wish to do that you should
	use a {@link Group} or {@link Project} instead.

	@author Nicklas
	@version 2.0
	@see Group
	@see Project
	@see <a href="../../../../../../specifications/core/defaultpermissions.html">Default permissions</a>
	@base.modified $Date: 2008-12-09 14:55:06 +0100 (Tue, 09 Dec 2008) $
*/
public class Role
	extends BasicItem<RoleData>
	implements Nameable, Removable, SystemItem, Registered
{
	/**
		The type of item represented by this class.
		@see Item#ROLE
		@see #getType()
	*/
	public static final Item TYPE = Item.ROLE;
	
	/**
		The id for the <code>Role</code> item representing adminstrators.
		By default administrators have full privileges on
		the server.
	*/
	public static final String ADMINISTRATOR = "net.sf.basedb.core.Role.ADMINISTRATOR";

	/**
		The id for the <code>Role</code> item representing supervisors.
		A supervisor have READ permission to everything in BASE.
	*/
	public static final String SUPERVISOR = "net.sf.basedb.core.Role.SUPERVISOR";

	/**
		The id for the <code>Role</code> item representing power users.
		A power user have less permissions than an administrator
		but may do some things that an ordinary user may not.
	*/
	public static final String POWER_USER = "net.sf.basedb.core.Role.POWER_USER";

	/**
		The id for the <code>Role</code> item representing regular users.
		This role should be sufficient for most regular users
		of BASE.
	*/
	public static final String USER = "net.sf.basedb.core.Role.USER";

	/**
		The id for the <code>Role</code> item representing guests.
		Guests have very limited access to the server.
	*/
	public static final String GUEST = "net.sf.basedb.core.Role.GUEST";

	/**
		The id for the <code>Role</code> item representing job agents.
		The job agents have permission to read jobs and execute them.
	*/
	public static final String JOBAGENT = "net.sf.basedb.core.Role.JOBAGENT";

	/**
		This filter will limit a query to only return roles
		where the logged in user is a member unless the logged in user
		has generic read permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new <code>Role</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@return The new <code>Role</code> item
		@throws BaseException If there is an error
	*/
	public static Role getNew(DbControl dc)
		throws BaseException
	{
		Role r = dc.newItem(Role.class);
		r.setName("New role");
		r.getData().setEntryDate(new Date());
		return r;
	}

	/**
		Get a <code>Role</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Role</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Role getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Role r = dc.loadItem(Role.class, id);
		if (r == null) throw new ItemNotFoundException("Role[id="+id+"]");
		return r;
	}

	/**
		Get a {@link ItemQuery} object configured
		to retrieve <code>Role</code> items. If the logged in user
		doesn't have generic permission to all roles, only roles
		where that user is a member are included in the list.
		@return A {@link ItemQuery} object
	*/
	public static ItemQuery<Role> getQuery()
	{
		return new ItemQuery<Role>(Role.class, RUNTIME_FILTER);
	}

	Role(RoleData roleData)
	{
		super(roleData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}
	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		If the logged in user is a member of this role, read permission
		is granted. If this is a system role, delete and create permissions are
		revoked.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		if (getSessionControl().isMemberOf(this))
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		If this role should be assigned to new users by default or not.
		@since 2.4
	*/
	public boolean isDefault()
	{
		return getData().isDefault();
	}
	
	/**
		If this role should be assigned to new users by default or not.
		Note! The client must call {@link User#addToDefaultRolesAndGroups()}.
		@param isDefault The new setting
		@throws PermissionDeniedException If the logged in user doesn't
			have {@link Permission#WRITE} permission for the role
		@since 2.4
	*/
	public void setDefault(boolean isDefault)
	{
		checkPermission(Permission.WRITE);
		getData().setDefault(isDefault);
	}	
	
	/**
		Assign this <code>Role</code> to a user.
		@param user The user to be assigned this role
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#WRITE} permission for the role and 
			{@link Permission#USE} permission for the user
		@throws InvalidDataException If the user is null
	*/
	public void addUser(User user)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		user.checkPermission(Permission.USE);
		getData().getUsers().add(user.getData());
	}

	/**
		Revoke this <code>Role</code> from a user.
		@param user The user that should be removed from this role
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#WRITE} permission for the role and 
			{@link Permission#USE} permission for the user
		@throws InvalidDataException If the user is null
	*/
	public void removeUser(User user)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		user.checkPermission(Permission.USE);
		getData().getUsers().remove(user.getData());
	}

	/**
		Check if the given user is member of this role or not.
		@param user The user to check
		@return TRUE if the user is member, FALSE otherwise
	*/
	public boolean isMember(User user)
	{
		return getData().getUsers().contains(user.getData());
	}
	
	/**
		Get a query that returns the users that
		are members of this role. This query excludes users that the logged 
		in user doesn't have permission to read.

		@see User#getQuery()
	*/
	public ItemQuery<User> getUsers()
	{
		ItemQuery<User> query = User.getQuery();
		query.joinPermanent(Hql.innerJoin("roles", Item.ROLE.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.ROLE.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		A runtime filter implementation that limits a query to only
		return roles where the logged in user is a member unless the logged in user
		has generic read permission.
	*/
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.isDenied(query.getItemPermission(), Item.ROLE))
			{
				manager.enableFilter("denyAll");
			}
			else if (!sc.hasPermission(query.getItemPermission(), Item.ROLE))
			{
				if (sc.isLoggedIn() && query.getItemPermission() == Permission.READ)
				{
					// Load roles where the logged in user is a member
					org.hibernate.Filter filter = manager.enableFilter("memberOf");
					if (filter != null)
					{
						Set<Integer> roles = sc.getRoles();
						if (roles == null || roles.size() == 0) roles = Collections.singleton(0);
						filter.setParameterList("items", roles);
						filter.setParameter("owner", sc.getLoggedInUserId());
					}
				}
				else
				{
					// Do not load any roles if not logged in
					manager.enableFilter("denyAll");
				}
			}
		}
	}
}
