/*
	$Id: Scan.java 4696 2008-12-09 13:55:06Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ArrayBatchData;
import net.sf.basedb.core.data.ArraySlideData;
import net.sf.basedb.core.data.HardwareData;
import net.sf.basedb.core.data.HybridizationData;
import net.sf.basedb.core.data.ProtocolData;
import net.sf.basedb.core.data.ScanData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Date;
import java.util.Set;
import java.util.HashSet;

/**
	A scan represents the scanning of a {@link Hybridization} by
	some scanner. A single hybridization may be scanned multiple
	times by the same or different scanner.
	<p>
	A single scan normally produces one or more {@link Image}:s
	which are analysed by some software to produce a {@link RawBioAssay}.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2008-12-09 14:55:06 +0100 (Tue, 09 Dec 2008) $
*/
public class Scan 
	extends AnnotatedItem<ScanData>
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#SCAN
		@see #getType()
	*/
	public static final Item TYPE = Item.SCAN;

	/**
		Create a new <code>Scan</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param hybridization The hybridization the scan originates from
		@return The new <code>Scan</code> item
		@throws PermissionDeniedException If the logged in user doesn't
			have use permission for the hybridization
		@throws BaseException If there is an error
	*/
	public static Scan getNew(DbControl dc, Hybridization hybridization)
		throws PermissionDeniedException, BaseException
	{
		Scan scn = dc.newItem(Scan.class);
		scn.setHybridization(hybridization);
		scn.setName("New scan");
		scn.getData().setEntryDate(new Date());
		return scn;
	}

	/**
		Get a <code>Scan</code> object when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Scan</code> item
		@throws ItemNotFoundException If an item with the specified id is
			not found
		@throws PermissionDeniedException If the logged in user doesn't
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Scan getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Scan scn = dc.loadItem(Scan.class, id);
		if (scn == null) throw new ItemNotFoundException("Scan[id=" + id + "]");
		return scn;
	}
	
	/**
		Get a query that returns scans.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Scan> getQuery() 
	{
		return new ItemQuery<Scan>(Scan.class);
	}
	
	Scan(ScanData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
 	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/**
		Get the hybridization.
	*/
	public Set<Annotatable> getAnnotatableParents()
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		try
		{
			annotatable.add(getHybridization());
		}
		catch (PermissionDeniedException ex)
		{}
		return annotatable;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A {@link RawBioAssay} has been created from this scan
		</ul>
	*/
	public boolean isUsed() 
		throws BaseException
	{
		return countRawBioAssays() > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link RawBioAssay}:s created from this scan
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_SCAN", "rba.id");
			/*
				SELECT {1}
				FROM RawBioAssayData rba
				WHERE rba.scan = :scan
			*/
		query.setEntity("scan", this.getData());
		addUsingItems(using, Item.RAWBIOASSAY, query);
		return using;
	}
	/**
		Set scanner and protocol from project default settings.
		@since 2.9
	*/
	@Override
	void setProjectDefaults(Project activeProject)
		throws BaseException
	{
		super.setProjectDefaults(activeProject);
		if (!scannerHasBeenSet)
		{
			HardwareData scanner = 
				(HardwareData)activeProject.getDefaultData(getDbControl(), 
				Project.Default.SCANNER_HARDWARE);
			if (scanner != null) getData().setScanner(scanner);
			scannerHasBeenSet = true;
		}
		if (!protocolHasBeenSet)
		{
			ProtocolData protocol = 
				(ProtocolData)activeProject.getDefaultData(getDbControl(), 
				Project.Default.SCANNING_PROTOCOL);
			if (protocol != null) getData().setProtocol(protocol);
			protocolHasBeenSet = true;
		}
	}
	// -------------------------------------------

	/**
		Get the {@link Hybridization} this scan is created from. A scan
		must always have a hybridization.

		@return The <code>Hybridization</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission for the hybridization
		@throws BaseException If there is another error.
	*/
	public Hybridization getHybridization()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Hybridization.class, getData().getHybridization());
	}
	
	/**
		Set the {@link ArraySlide} this hybridization is using.

		@param hybridization The hybridization, null is not allowed
		@throws InvalidDataException If the hybridization is null
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission for the scan use permission for the hybridization
		@throws BaseException If there is another error
	*/
	public void setHybridization(Hybridization hybridization)
		throws InvalidDataException, PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (hybridization == null) throw new InvalidUseOfNullException("hybridization");
		hybridization.checkPermission(Permission.USE);
		getData().setHybridization(hybridization.getData());
	}

	/**
		Get the {@link Hardware} (scanner) used to create this scan. 

		@return The <code>Hardware</code> item or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission for the scanner
		@throws BaseException If there is another error.
	*/
	public Hardware getScanner()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Hardware.class, getData().getScanner());
	}
	
	private boolean scannerHasBeenSet = false;
	/**
		Set the {@link Hardware} (scanner) used to create this scan. 

		@param scanner The scanner or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission for the scan use permission for the scanner
		@throws BaseException If there is another error
	*/
	public void setScanner(Hardware scanner)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (scanner != null) scanner.checkPermission(Permission.USE);
		scannerHasBeenSet = true;
		getData().setScanner(scanner == null ? null : scanner.getData());
	}

	/**
		Get the {@link Protocol} used to create this scan. 

		@return The <code>Protocol</code> item or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission for the protocol
		@throws BaseException If there is another error.
	*/
	public Protocol getProtocol()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Protocol.class, getData().getProtocol());
	}
	
	private boolean protocolHasBeenSet = false;
	/**
		Set the {@link Protocol} used to create this scan. 

		@param protocol The protocol or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission for the scan use permission for the protocol
		@throws BaseException If there is another error
	*/
	public void setProtocol(Protocol protocol)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (protocol != null) protocol.checkPermission(Permission.USE);
		protocolHasBeenSet = true;
		getData().setProtocol(protocol == null ? null : protocol.getData());
	}

	/**
		Get the array design by connected to this scan following 
		the hybridization to the array slide to the array batch 
		to the array design.
		@return The array design, or null if no array design is connected
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission for the array design
		@throws BaseException If there is another error.
	*/
	public ArrayDesign getArrayDesign()
		throws PermissionDeniedException, BaseException
	{
		HybridizationData hyb = getData().getHybridization();
		if (hyb == null) return null;
		ArraySlideData slide = hyb.getArraySlide();
		if (slide == null) return null;
		ArrayBatchData batch = slide.getArrayBatch();
		if (batch == null) return null;
		return getDbControl().getItem(ArrayDesign.class, batch.getArrayDesign());
	}
	
	/**
		Create a new {@link RawBioAssay} from this scan.
	 
		@param rawDataType The raw data type of the new bioassay
		@return The new <code>RawBioAssay</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission for the scan
		@throws BaseException If there is another error
		@deprecated Use {@link RawBioAssay#getNew(DbControl, Platform, RawDataType)}
			or {@link RawBioAssay#getNew(DbControl, PlatformVariant, RawDataType)} instead
	*/
	public RawBioAssay newRawBioAssay(RawDataType rawDataType) 
		throws PermissionDeniedException, BaseException
	{
		RawBioAssay rba = RawBioAssay.getNew(getDbControl(), rawDataType);
		rba.setScan(this);
		return rba;
	}

	/**
		Get a query that returns all raw bioassays created from this 
		scan.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<RawBioAssay> getRawBioAssays()
	{
		ItemQuery<RawBioAssay> query = RawBioAssay.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("scan"), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Counts all raw bioassays that have been created from this scan. The count
		includes raw bioassays that the logged in user doesn't have access to.
		@return The number of raw bioassays
	 	@throws BaseException If the raw bioassays couldn't be counted. 
	 */
	public long countRawBioAssays()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_SCAN", "count(*)");
		/*
			SELECT {1}
			FROM RawBioAssayData rba
			WHERE rba.scan = :scan
		*/
		query.setEntity("scan", this.getData());
		return HibernateUtil.loadData(Long.class, query);
	}

	/**
		Create a new {@link Image} for this scan.
		@return The new <code>Image</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws BaseException If there is another error
	*/
	public Image newImage()
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		return Image.getNew(getDbControl(), this);
	}
	
	/**
		Get a query that returns all images for this scan.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Image> getImages()
	{
		return Image.getQuery(this);
	}
	
	/**
		Counts all images that have been added to this scan. The count
		includes images that the logged in user doesn't have access to.
		@return The number of images
	 	@throws BaseException If the images couldn't be created. 
	 */
	public long countImages()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_IMAGES_FOR_SCAN", "count(*)");
		/*
			SELECT {1}
			FROM ImageData img
			WHERE img.scan = :scan
		*/
		query.setEntity("scan", this.getData());
		return HibernateUtil.loadData(Long.class, query);
	}

}
