/*
	$Id: PluginDefinitionData.java 4696 2008-12-09 13:55:06Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import net.sf.basedb.core.plugin.About;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.HashSet;

/**
	This class represent an installed plugin.

	@author Nicklas, Samuel
	@version 2.0
	@see net.sf.basedb.core.PluginDefinition
	@see <a href="../../../../../../html/developerdoc/api/data_api.plugins.html">Plugins</a>
	@hibernate.class table="`PluginDefinitions`" lazy="false"
*/
public class PluginDefinitionData
	extends SharedData
	implements RemovableData, AnnotatableData, RegisteredData
{

	public PluginDefinitionData()
	{}

	/*
		From the RemovableData interface
		-------------------------------------------
	*/
	private boolean removed;
	public boolean isRemoved()
	{
		return removed;
	}
	public void setRemoved(boolean removed)
	{
		this.removed = removed;
	}
	// -------------------------------------------
	/*
		From the AnnotatableData interface
		-------------------------------------------
	*/
	private AnnotationSetData annotationSet;
	public AnnotationSetData getAnnotationSet()
	{
		return annotationSet;
	}
	public void setAnnotationSet(AnnotationSetData annotationSet)
	{
		this.annotationSet = annotationSet;
	}
	// -------------------------------------------
	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------
	
	/**
		The maximum length of the class name that can be stored in the
		database.
	*/
	public static final int MAX_CLASSNAME_LENGTH = 255;
	private String className;
	/**
		Get the class name of the Java class that implements this plugin. The
		specified class must implement the {@link net.sf.basedb.core.plugin.Plugin}
		interface.
		@return The class name for this plugin
		@hibernate.property column="`class_name`" type="string" length="255" not-null="true" unique="true"
	*/
	public String getClassName()
	{
		return className;
	}
	public void setClassName(String className)
	{
		this.className = className;
	}


	public static final int MAX_JARPATH_LENGTH = 65535;
	private String jarPath;
	/**
		Get the path to the JAR file where the plugin class is located. If the
		value is null the plugin must be in the classpath.
		@return The class name for this plugin
		@hibernate.property column="`jar_path`" type="text" not-null="false"
	*/
	public String getJarPath()
	{
		return jarPath;
	}
	public void setJarPath(String jarPath)
	{
		this.jarPath = jarPath;
	}
	
	/**
		The maximum length of the name that can be stored in the
		database.
	*/
	public static final int MAX_NAME_LENGTH = About.MAX_INFORMATION_LENGTH;
	private String name;
	/**
		Get the name of the plugin.
		@return A string with the name of the plugin
		@hibernate.property column="`name`" type="string" length="255" not-null="true"
	*/
	public String getName()
	{
		return name;
	}
	public void setName(String name)
	{
		this.name = name;
	}

	/**
		The maximum length of the url that can be stored in the
		database.
	*/
	public static final int MAX_DESCRIPTION_LENGTH = 65535;
	private String description;
	/**
		Get a description of the plugin.
		@return A string with a description, or null if not known
		@hibernate.property column="`description`" type="text" not-null="false"
	*/
	public String getDescription()
	{
		return description;
	}
	public void setDescription(String description)
	{
		this.description = description;
	}

	/**
		The maximum length of the version that can be stored in the
		database.
	*/
	public static final int MAX_VERSION_LENGTH = About.MAX_INFORMATION_LENGTH;
	private String version;
	/**
		Get the version of the plugin.
		@return A string with the version of the plugin, or null if not known
		@hibernate.property column="`version_string`" type="string" length="255" not-null="false"
	*/
	public String getVersionString()
	{
		return version;
	}
	public void setVersionString(String version)
	{
		this.version = version;
	}
	
	/**
		The maximum length of the copyright notice that can be stored in the
		database.
	*/
	public static final int MAX_COPYRIGHT_LENGTH = About.MAX_INFORMATION_LENGTH;
	private String copyright;
	/**
		Get a copyright notice for the plugin.
		@return A string with a copyright notice, or null if not known
		@hibernate.property column="`copyright`" type="string" length="255" not-null="false"
	*/
	public String getCopyright()
	{
		return copyright;
	}
	public void setCopyright(String copyright)
	{
		this.copyright = copyright;
	}
	
	/**
		The maximum length of the contact information that can be stored in the
		database.
	*/
	public static final int MAX_CONTACT_LENGTH = About.MAX_INFORMATION_LENGTH;
	private String contact;
	/**
		Get contact information for the plugin.
		@return A string with contact information, or null if not known
		@hibernate.property column="`contact`" type="string" length="255" not-null="false"
	*/
	public String getContact()
	{
		return contact;
	}
	public void setContact(String contact)
	{
		this.contact = contact;
	}
	
	/**
		The maximum length of the email address that can be stored in the
		database.
	*/
	public static final int MAX_EMAIL_LENGTH = About.MAX_INFORMATION_LENGTH;
	private String email;
	/**
		Get an email address that can be used to get more information about the plugin.
		@return A string with the email address, or null if not known
		@hibernate.property column="`email`" type="string" length="255" not-null="false"
	*/
	public String getEmail()
	{
		return email;
	}
	public void setEmail(String email)
	{
		this.email = email;
	}
	
	/**
		The maximum length of the url that can be stored in the
		database.
	*/
	public static final int MAX_URL_LENGTH = About.MAX_INFORMATION_LENGTH;
	private String url;
	/**
		Get a URL with more information about the plugin.
		@return A string containing the URL, or null if not known
		@hibernate.property column="`url`" type="string" length="255" not-null="false"
	*/
	public String getUrl()
	{
		return url;
	}
	public void setUrl(String url)
	{
		this.url = url;
	}

	private int mainType;
	/**
		Get the main type of plugin.
		@return A type code that is defined in the enum {@link net.sf.basedb.core.plugin.Plugin.MainType}.
		@hibernate.property column="`type`" type="int" not-null="true"
	*/
	public int getMainType()
	{
		return mainType;
	}
	public void setMainType(int mainType)
	{
		this.mainType = mainType;
	}
	
	private boolean interactive;
	/**
		If the plugin is interactive or not.
		@hibernate.property column="`interactive`" type="boolean" not-null="true"
	*/
	public boolean isInteractive()
	{
		return interactive;
	}
	public void setInteractive(boolean interactive)
	{
		this.interactive = interactive;
	}
	
	private boolean supportsConfigurations;
	/**
		If the plugin supports configurations or not. 
		@hibernate.property column="`supports_config`" type="boolean" not-null="true"
	*/
	public boolean getSupportsConfigurations()
	{
		return supportsConfigurations;
	}
	public void setSupportsConfigurations(boolean supportsConfigurations)
	{
		this.supportsConfigurations = supportsConfigurations;
	}

	private boolean requiresConfiguration;
	/**
		If the plugin requires a configurations or not. Requires that the plugin 
		supports configurations.
		@hibernate.property column="`requires_config`" type="boolean" not-null="true"
	*/
	public boolean getRequiresConfiguration()
	{
		return requiresConfiguration;
	}
	public void setRequiresConfiguration(boolean requiresConfiguration)
	{
		this.requiresConfiguration = requiresConfiguration;
	}
	
	private boolean usePermissions;
	/**
	  	If the permissions for this plugin should be used or not.
	  	@hibernate.property column="`use_permissions`" type="boolean" not-null="true"
	*/
	public boolean getUsePermissions()
	{
		return usePermissions;
	}
	public void setUsePermissions(boolean usePermissions)
	{
		this.usePermissions = usePermissions;
	}
	
	private Set<GuiContextData> guiContexts;
	/**
		The item code for all items where it makes sense to use the 
		plugin in a client application. Ie. a plugin that imports reporters
		should return a set containing the code for Item.REPORTER.
		@hibernate.set table="`PluginDefinitionGuiContexts`" lazy="true"
		@hibernate.collection-key column="`plugindefinition_id`"
		@hibernate.collection-composite-element class="net.sf.basedb.core.data.GuiContextData"
	*/
	public Set<GuiContextData> getGuiContexts()
	{
		if (guiContexts == null) guiContexts = new HashSet<GuiContextData>();
		return guiContexts;
	}
	void setGuiContexts(Set<GuiContextData> guiContexts)
	{
		this.guiContexts = guiContexts;
	}
	
	private Set<PluginTypeData> pluginTypes;
	/**
		The plugin types of this plugin, ie. all interfaces that it implements.
		@hibernate.set table="`PluginDefinitionTypes`" lazy="true"
		@hibernate.collection-key column="`plugindefinition_id`"
		@hibernate.collection-many-to-many column="`plugintype_id`" class="net.sf.basedb.core.data.PluginTypeData"
	*/
	public Set<PluginTypeData> getPluginTypes()
	{
		if (pluginTypes == null) pluginTypes = new HashSet<PluginTypeData>();
		return pluginTypes;
	}
	void setPluginTypes(Set<PluginTypeData> pluginTypes)
	{
		this.pluginTypes = pluginTypes;
	}
	
	private Map<RoleKeyData, PluginPermission> permissions;
	/**
		Get a map containing the rolekey and the permissions for this plugin.
		@hibernate.map table="`PluginKeys`" lazy="true" cascade="delete"
		@hibernate.collection-key column="`plugindefinition_id`"
		@hibernate.index-many-to-many column="`key_id`" class="net.sf.basedb.core.data.RoleKeyData"
		@hibernate.collection-composite-element class="net.sf.basedb.core.data.PluginPermission"
	*/
	public Map<RoleKeyData, PluginPermission> getPermissions()
	{
		if (permissions == null) permissions = new HashMap<RoleKeyData, PluginPermission>();
		return permissions;
	}
	void setPermissions(Map<RoleKeyData, PluginPermission> permissions)
	{
		this.permissions = permissions;
	}
	
	private boolean trusted;
	/**
		If the plugin should be trusted or not.
		@hibernate.property column="`trusted`" type="boolean" not-null="true"
	*/
	public boolean isTrusted()
	{
		return trusted;
	}
	public void setTrusted(boolean trusted)
	{
		this.trusted = trusted;
	}
	
	private Long maxMemory;
	/**
		The max amount of memory the plugin can use. If the value is null 
		the job agent is free to choose an appropriate value.
		@hibernate.property column="`max_memory`" type="long" not-null="false"
	*/
	public Long getMaxMemory()
	{
		return maxMemory;
	}
	public void setMaxMemory(Long maxMemory)
	{
		this.maxMemory = maxMemory;
	}
	
	private Set<JobAgentSettingsData> jobAgentSettings;
	/**
	 	This is the inverse end.
		@see JobAgentSettingsData#getPluginDefinition()
	 	@hibernate.set inverse="true" lazy="true" cascade="delete"
	 	@hibernate.collection-key column="`plugindefinition_id`"
	 	@hibernate.collection-one-to-many class="net.sf.basedb.core.data.JobAgentSettingsData"
	*/
	Set<JobAgentSettingsData> getJobAgentSettings()
	{
		return jobAgentSettings;
	}
	void setJobAgentSettings(Set<JobAgentSettingsData> jobAgentSettings)
	{
		this.jobAgentSettings = jobAgentSettings;
	}
	
	private Set<PluginConfigurationData> configurations;
	/**
	 	This is the inverse end.
		@see PluginConfigurationData#getPluginDefinition()
	 	@hibernate.set inverse="true" lazy="true"
	 	@hibernate.collection-key column="`plugindefinition_id`"
	 	@hibernate.collection-one-to-many class="net.sf.basedb.core.data.PluginConfigurationData"
	*/
	Set<PluginConfigurationData> getConfigurations()
	{
		return configurations;
	}
	void setConfigurations(Set<PluginConfigurationData> configurations)
	{
		this.configurations = configurations;
	}
	
	private boolean allowImmediateExecution;
	/**
		If the plugin is allowed to be immediately executed or if it
		must be added to the job queue.
		@since 2.2
		@hibernate.property column="`allow_immediate`" type="boolean" not-null="true"
	*/
	public boolean getAllowImmediateExecution()
	{
		return allowImmediateExecution;
	}
	public void setAllowImmediateExecution(boolean allowImmediateExecution)
	{
		this.allowImmediateExecution = allowImmediateExecution;
	}
	
	private boolean useInternalJobQueue = true;
	/**
		If the plug-in can be executed with the internal job queue or
		not. If FALSE is used, the plug-in can only be executed by a
		job agent or other external tool.
		@since 2.5
		@hibernate.property column="`use_internal_queue`" type="boolean" not-null="true"
	*/
	public boolean getUseInternalJobQueue()
	{
		return useInternalJobQueue;
	}
	public void setUseInternalJobQueue(boolean useInternalJobQueue)
	{
		this.useInternalJobQueue = useInternalJobQueue;
	}
	
}
