/*
  $Id $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.HashSet;
import java.util.Set;

/**
	This class holds information units. It is important to
	get the {@link #getReferenceFactor()} and {@link #getReferenceOffset()}
	correct. This is the formula to use:
	
	<pre class="code">
1 reference unit * factor + offset = 1 of this unit
</pre>
	
	Examples:
	<pre class="code">
1 meter * 0.001 + 0 = 1 millimeter
1 kelvin * 1 + 273.15 = 1 Celsius
</pre>
	
	@author Nicklas
	@version 2.9
	@see net.sf.basedb.core.Unit
	@base.modified $Date $
	@hibernate.class table="`Units`" lazy="true"
*/
public class UnitData
	extends BasicData
	implements NameableData, RemovableData
{

	/*
		From the Nameable interface
		-------------------------------------------
	*/
	private String name;
	public String getName()
	{
		return name;
	}
	public void setName(String name)
	{
		this.name = name;
	}

	private String description;
	public String getDescription()
	{
		return description;
	}

	public void setDescription(String description)
	{
		this.description = description;
	}
	// -------------------------------------------
	/*
		From the RemovableData interface
		-------------------------------------------
	*/
	private boolean removed;
	public boolean isRemoved()
	{
		return removed;
	}
	public void setRemoved(boolean removed)
	{
		this.removed = removed;
	}
	// -------------------------------------------

	private QuantityData quantity;
	/**
		Get the quantity this unit belongs to. Can't be
		changed after creation.
		@hibernate.many-to-one column="`quantity_id`" not-null="true" 
			outer-join="false" update="false"
	*/
	public QuantityData getQuantity()
	{
		return quantity;
	}
	public void setQuantity(QuantityData quantity)
	{
		this.quantity = quantity;
	}
	
	public static final int MAX_SYMBOL_LENGTH = 255;
	private String displaySymbol;
	/**
		The symbol to display together with values of this unit.
		@hibernate.property column="`display_symbol`" type="string" length="255" not-null="true"
	*/
	public String getDisplaySymbol()
	{
		return displaySymbol;
	}
	public void setDisplaySymbol(String displaySymbol)
	{
		this.displaySymbol = displaySymbol;
	}

	private double referenceFactor;
	/**
		Factor that controls how this unit relates to the reference
		unit of the quantity. The factor tells us how many of the reference unit
		is needed to get this unit. For example, if the reference
		unit is 'kilogram' and this unit is 'gram', the factor is 0.001.
		@see #getReferenceOffset()
		@hibernate.property column="`reference_factor`" type="double" not-null="true"
	*/
	public double getReferenceFactor()
	{
		return referenceFactor;
	}
	public void setReferenceFactor(double referenceFactor)
	{
		this.referenceFactor = referenceFactor;
	}

	private double referenceOffset;
	/**
		An offset that controls how this unit relates to the reference
		unit of the quantity. The offset tells us how to translate 
		a between units. For example, if the reference unit is
		'Kelvin' and this unit is 'Celsius', the offset is
		273.15.
		@see #getReferenceFactor()
		@hibernate.property column="`reference_offset`" type="double" not-null="true"
	*/
	public double getReferenceOffset()
	{
		return referenceOffset;
	}
	public void setReferenceOffset(double referenceOffset)
	{
		this.referenceOffset = referenceOffset;
	}
	
	private Set<UnitSymbolData> symbols;
	/**
		This is the inverse end.
		@see UnitSymbolData#getUnit()
		@hibernate.set lazy="true" inverse="true" cascade="all-delete-orphan"
		@hibernate.collection-key column="`unit_id`"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.UnitSymbolData"
	*/
	public Set<UnitSymbolData> getSymbols()
	{
		if (symbols == null) symbols = new HashSet<UnitSymbolData>();
		return symbols;
	}
	void setSymbols(Set<UnitSymbolData> symbols)
	{
		this.symbols = symbols;
	}

	private Set<AnnotationTypeData> usedUnits;
	/**
		This is the inverse end. It has lazy="false" because we need
		automatic deletion from the AnnotationTypeUnits table
		@see AnnotationTypeData#getUsableUnits()
		@hibernate.set table="`AnnotationTypeUnits`" lazy="true" 
		@hibernate.collection-key column="`unit_id`"
		@hibernate.collection-many-to-many column="`annotationtype_id`" 
			class="net.sf.basedb.core.data.AnnotationTypeData"
	*/
	Set<AnnotationTypeData> getUsedUnits()
	{
		return usedUnits;
	}
	void setUsedUnits(Set<AnnotationTypeData> usedUnits)
	{
		this.usedUnits = usedUnits;
	}

}
