/**
	$Id:XmlLoader.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.xml;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XMLUtil;
import net.sf.basedb.util.extensions.AboutBean;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionBean;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.ExtensionPointBean;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.RendererFactory;

/**
	Loads extension points and extensions from extension definition 
	XML files. This class may load extensions from more than one file
	and register or unregister all loaded extensions and extension points 
	with a registry.
	<p>
	For a description of the XML file format this class can load
	see xxxx.
	
	<p>
	<b>Factory initialisation</b><br>
	
	This loader uses reflection to create and initialise factory
	instances. The XML file contains the class name of the factory
	to create. This class must declare a public no-argument
	constructor and must also implement one of the specific
	factory interfaces (for example {@link ActionFactory}
	or {@link RendererFactory}). If the tag <code>&lt;parameters&gt;</code>
	is present inside the factory declaration, the loader will
	use reflection to find a setter method for each sub-tag inside
	the <code>&lt;parameters&gt;</code> tag. Here is an example:

	<pre class="code">
// XML contents
&lt;parameters&gt;
   &lt;image&gt;button.png&lt;/image&gt;
   &lt;title&gt;Click button&lt;/title&gt;
&lt;/parameters&gt;

// Factory initialization
factory.setImage("button.png");
factory.setTitle("Click button");
</pre>

	The setter methods must be public and accept a single <code>String</code> 
	parameter.
	<p>
	
	The loader also looks for the <code>setParameter(String, String)</code>
	method signature. If the method exists it will be called for each 
	parameter found with the tagname as the first parameter and the
	value as the second parameter. Ccontinuing the above example:
	
	<pre class="code">
factory.setParameter("image", "button.png");
factory.setParameter("title", "Click button");
</pre>
	
	<p>
	Tags that doesn't have a corresponding setter method are simply ignored.
	The XML file format allows any tags as parameters, but only the first level
	will be parsed.
	
	<p>
	The parameter values may be subject to conversion by one or more 
	{@link ValueConverter}:s. Converters are added to the loader
	by {@link #addValueConverter(ValueConverter)} and are usually
	implemented to react on method annotations. For example, the web 
	client uses this for path and variable substitution.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class XmlLoader 
{

	/**
		The URL pointing to the extensions.xsd schema.
	*/
	private static final String schemaFileURL = 
		XmlLoader.class.getResource("/net/sf/basedb/core/xsd/extensions.xsd").toString();

	/**
		The name of the extensions namespace.
	*/
	private static final String namespace = "http://base.thep.lu.se/extensions.xsd";
	
	// To convert XML elements to a string representation
	private XMLOutputter xmlOut;
	
	// All loaded extension points
	private List<ExtensionPoint<Action>> extensionPoints;
	
	// All loaded extensions
	private List<Extension<Action>> extensions;
	
	// List of registered converters
	private Set<ValueConverter> converters;
	
	// Map with XML representation of factory parameters
	private Map<Object, String> factoryParameters;
	
	// The last document that was validated
	private Document validatedDom;
	
	/**
		Create a new XML loader instance.
	*/
	public XmlLoader()
	{
		extensionPoints = new LinkedList<ExtensionPoint<Action>>();
		extensions = new LinkedList<Extension<Action>>();
		factoryParameters = new HashMap<Object, String>();
		xmlOut = new XMLOutputter(Format.getPrettyFormat());
	}
	
	/**
		Add a value converter to this loader
		@param converter The converter to add
		@throws NullPointerException If the converter is null
	*/
	public void addValueConverter(ValueConverter converter)
	{
		if (converter == null) throw new InvalidUseOfNullException("converter");
		if (converters == null) converters = new LinkedHashSet<ValueConverter>();
		converters.add(converter);
	}

	/**
		Remove a converter.
		@param converter The converter to remove, null values are ignored
	*/
	public void removeValueConverter(ValueConverter converter)
	{
		if (converter == null || converters == null) return;
		converters.remove(converter);
	}
	
	/**
		Remove all value converters.
	*/
	public void clearValueConverters()
	{
		if (converters != null) converters.clear();
	}
	
	/**
		Load an extensions definition XML file. This method may be called 
		multiple times with different XML files. The loading may also be
		done by a two-step process where the first step validates the file
		and the second step loads the extensions. See {@link 
		#validateXmlFile(InputStream, String)} and {@link 
		#loadLastValidatedFile(ClassLoader, boolean)}.

		@param xmlFile An input stream to read the XML data from
		@param filename The original filename the stream is coming from, or null
			if not known. This value is only used when generating error messages
		@param classLoader The classloader to use when loading classes that are
			named in the XML file, or null to use the default classloader (=the same
			class loader that loaded the BASE core classes)
		@param clear TRUE to clear all already loaded extensions before loading 
			the extensions in this file
		@return Information about the extension, or null if no such information
			is available
		@throws JDOMException If validation of the XML file fails
		@throws IOException If there is an error reading the XML file
		@throws ClassNotFoundException If a class named in the XML file can't
			be found
		@throws NoSuchMethodException If a factory class doesn't implement a
			public, no-argument constructor
		@throws IllegalAccessException If a factory class constructor isn't 
			public
		@throws InstantiationException If a factory class can't be instantiated,
			for example, because it is an abstract class or an interface
		@see #validateXmlFile(InputStream, String)
		@see #loadLastValidatedFile(ClassLoader, boolean)
	*/
	public About loadXmlFile(InputStream xmlFile, String filename,
			ClassLoader classLoader, boolean clear)
		throws JDOMException, IOException, ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException
	{
		validateXmlFile(xmlFile, filename);
		return loadLastValidatedFile(classLoader, clear);
	}
	
	/**
		Validate an XML file against the extensions definition schema. If the
		file is valid you may continue to load the extensions. See
		{@link #loadLastValidatedFile(ClassLoader, boolean)}.
		
		@param xmlFile An input stream to read the XML data from
		@param filename The original filename the stream is coming from, or null
			if not known. This value is only used when generating error messages
		@return Information about the extension, or null if no such information
			is available
		@throws JDOMException If validation of the XML file fails
		@throws IOException If there is an error reading the XML file
	*/
	public About validateXmlFile(InputStream xmlFile, String filename)
		throws IOException, JDOMException
	{
		validatedDom = null;
		validatedDom = loadDocument(xmlFile, filename);
		About globalAbout = loadGlobalAbout(validatedDom);
		return globalAbout;
	}
	
	/**
		Continue loading extensions from the last validated XML file. This is
		the second step in a two-step process and requires that 
		{@link #validateXmlFile(InputStream, String)} has been successfully called 
		first.

		@param classLoader The classloader to use when loading classes that are
			named in the XML file, or null to use the default classloader (=the same
			class loader that loaded the BASE core classes)
		@param clear TRUE to clear all already loaded extensions before loading 
			the extensions in this file
		@return Information about the extension, or null if no such information
			is available
		@throws ClassNotFoundException If a class named in the XML file can't
			be found
		@throws NoSuchMethodException If a factory class doesn't implement a
			public, no-argument constructor
		@throws IllegalAccessException If a factory class constructor isn't 
			public
		@throws InstantiationException If a factory class can't be instantiated,
			for example, because it is an abstract class or an interface
	*/
	public About loadLastValidatedFile(ClassLoader classLoader, boolean clear)
		throws ClassNotFoundException, IllegalAccessException, 
		InstantiationException, NoSuchMethodException
	{
		if (validatedDom == null) 
		{
			throw new NullPointerException("No file has been validated");
		}
		if (clear)
		{
			extensionPoints.clear();
			extensions.clear();
		}
		About globalAbout = loadGlobalAbout(validatedDom);
		loadExtensionPoints(validatedDom, classLoader, globalAbout);
		loadExtensions(validatedDom, classLoader, globalAbout);
		validatedDom = null;
		return globalAbout;
	}
	
	/**
		Checks if an XML file has passed validation in the
		{@link #validateXmlFile(InputStream, String)} method. If so,
		the {@link #loadLastValidatedFile(ClassLoader, boolean)} can be called
		to continue loading the extensions.
		<p>
		Note that once the file has been loaded this flag is reset to
		FALSE.
		
		@return TRUE if a file has been validated, FALSE otherwise
	*/
	public boolean hasValidFile()
	{
		return validatedDom != null;
	}
	
	/**
		Get a list with all loaded extension points.
	*/
	public List<ExtensionPoint<Action>> getExtensionPoints()
	{
		return extensionPoints;
	}
	
	/**
		Get a list with all loaded extensions.
	*/
	public List<Extension<Action>> getExtensions()
	{
		return extensions;
	}
	
	/**
		Get a XML-like string representation of the parameters that was
		used to initialise a factory.
		@param factory The factory object
		@return The parameters as found in the XML file
	*/
	public String getFactoryParameters(Object factory)
	{
		return factoryParameters.get(factory);
	}
	
	/**
		Register all loaded extension points with a registry.
		@param registry The registry
		@param update TRUE to update already registered extension
			points, FALSE to ignore
		@return The number of extension points registered or updated
		@see Registry#registerExtensionPoint(ExtensionPoint)
	*/
	public int registerExtensionPoints(Registry registry, boolean update)
	{
		int numRegistered = 0;
		for (ExtensionPoint<Action> ep : extensionPoints)
		{
			if (update || !registry.extensionPointIsRegistered(ep.getId()))
			{
				registry.registerExtensionPoint(ep);
				numRegistered++;
			}
		}
		return numRegistered;
	}
	
	/**
		Unregister all loaded extension points from a registry.
		All extensions will also be unregistered.
		@param registry The registry
		@return The number of extension points that was unregistered
		@see Registry#unregisterExtensionPoint(String)
	*/
	public int unregisterExtensionPoints(Registry registry)
	{
		for (ExtensionPoint<Action> ep : extensionPoints)
		{
			registry.unregisterExtensionPoint(ep.getId());
		}
		return extensionPoints.size();
	}
	
	/**
		Register all loaded extensions with a registry.
		@param registry The registry
		@param update TRUE to update already registered extensions
			FALSE to ignore
		@return The number of registered or updated extensions
		@see Registry#registerExtension(Extension)
	*/
	public int registerExtensions(Registry registry, boolean update)
	{
		int numRegistered = 0;
		for (Extension<Action> ext : extensions)
		{
			if (update || !registry.extensionIsRegistered(ext.getId()))
			{
				registry.registerExtension(ext);
				numRegistered++;
			}
		}
		return numRegistered;
	}
	
	/**
		Unregister all loaded extensions from a registry.
		@param registry The registry
		@return The number of extensions that was unregistered
		@see Registry#unregisterExtension(String)
	*/
	public int unregisterExtensions(Registry registry)
	{
		for (Extension<Action> ext : extensions)
		{
			registry.unregisterExtension(ext.getId());
		}
		return extensions.size();
	}
	
	/**
		Load and validate the XML file.
		@return A JDOM document object
	*/
	protected Document loadDocument(InputStream xmlFile, String filename)
		throws IOException, JDOMException
	{
		return XMLUtil.getSchemaValidatedXML(xmlFile, filename, namespace, schemaFileURL);
	}
	
	/**
		Load the global about tag from the document.
		@return An about bean or null if the about tag is not found.
	*/
	protected AboutBean loadGlobalAbout(Document dom)
	{
		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		return loadAbout(root.getChild("about", ns));
	}
	
	/**
		Load information in an <code>&lt;about&gt;</code> tag.
		@param aboutTag The about tag
	*/
	protected AboutBean loadAbout(Element aboutTag)
	{
		AboutBean about = null;
		if (aboutTag != null)
		{
			Namespace ns = aboutTag.getNamespace();
			about = new AboutBean();
			about.setName(Values.getStringOrNull(aboutTag.getChildText("name", ns)));
			about.setDescription(Values.getStringOrNull(aboutTag.getChildText("description", ns)));
			about.setVersion(Values.getStringOrNull(aboutTag.getChildText("version", ns)));
			about.setContact(Values.getStringOrNull(aboutTag.getChildText("contact", ns)));
			about.setCopyright(Values.getStringOrNull(aboutTag.getChildText("copyright", ns)));
			about.setEmail(Values.getStringOrNull(aboutTag.getChildText("email", ns)));
			about.setUrl(Values.getStringOrNull(aboutTag.getChildText("url", ns)));
		}
		return about;
	}
	
	
	/**
		Create extension points from an XML document.
	*/
	@SuppressWarnings("unchecked")
	protected int loadExtensionPoints(Document dom, ClassLoader classLoader, About globalAbout)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException
	{
		int numLoaded = 0;
		List<ExtensionPoint<Action>> temp = new LinkedList<ExtensionPoint<Action>>();
		/*
			Format of an extension point is:
			<extension-point
				id="..."
				>
				<name>...</name>
				<description>...</description>
				<action-class>...</action-class>
				<renderer-factory override="true|false">
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</renderer-factory>
			</extension-point>
		 */
		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		List<Element> epTags = root.getChildren("extension-point", ns);
		for (Element epTag : epTags)
		{
			ExtensionPointBean<Action> extensionPoint = new ExtensionPointBean<Action>();
			temp.add(extensionPoint);
			numLoaded++;
	
			// Set ID (required)
			String id = Values.getStringOrNull(epTag.getAttributeValue("id"));
			extensionPoint.setId(id);
			
			// Set Name and Description (optional)
			extensionPoint.setName(Values.getStringOrNull(epTag.getChildText("name", ns)));
			extensionPoint.setDescription(Values.getStringOrNull(epTag.getChildText("description", ns)));
			
			// Load and set Action class (required)
			String actionClassName = Values.getStringOrNull(epTag.getChildText("action-class", ns));
			Class actionClass = 
				ClassUtil.checkAndLoadClass(classLoader, actionClassName, false, Action.class);
			extensionPoint.setActionClass(actionClass);
			
			// Load and set RendererFactory (optional)
			Element rfTag = epTag.getChild("renderer-factory", ns);
			if (rfTag != null)
			{
				RendererFactory rf = createFactory(rfTag, classLoader, RendererFactory.class);
				extensionPoint.setRendererFactory(rf);
				extensionPoint.setAllowRendererOverrider(
					Values.getBoolean(rfTag.getAttributeValue("override")));
			}
		}
		extensionPoints.addAll(temp);
		return numLoaded;
	}
	

	/**
		Create extensions from an XML document.
	*/
	@SuppressWarnings("unchecked")
	protected int loadExtensions(Document dom, ClassLoader classLoader, About globalAbout)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException
	{
		int numLoaded = 0;
		List<Extension<Action>> temp = new LinkedList<Extension<Action>>();
		/*
			Format of an extension is:
			<extension
				id="..."
				extends="..."
				>
				<position>...</position>
				<about>
					<name>...</name>
					<description>...</name>
					<version>...</version>
					<contact>...</contact>
					<copyright>...</copyright>
					<email>...</email>
					<url>...</url>
				</about>
				<action-factory>
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</action-factory>
				<action-factory>
					<factory-class>...</factory-class>
					<parameters>
						any-tags
					</parameters>
				</action-factory>
			</extension>
		*/

		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		List<Element> epTags = root.getChildren("extension", ns);
		for (Element epTag : epTags)
		{
			ExtensionBean<Action> ext = new ExtensionBean<Action>();
			temp.add(ext);
			numLoaded++;
			
			// Set ID (required)
			String id = Values.getStringOrNull(epTag.getAttributeValue("id"));
			ext.setId(id);
			
			// Set extension point (required)
			String extensionPoint = Values.getStringOrNull(epTag.getAttributeValue("extends"));
			ext.setExtends(extensionPoint);
			
			// Set index position (optional, default value = 999)
			ext.setIndex(Values.getFloat(epTag.getChildText("index", ns), 999.0f));
			
			// Set About (optional)
			AboutBean about = loadAbout(epTag.getChild("about", ns));
			if (globalAbout != null)
			{
				if (about == null) about = new AboutBean();
				about.copy(globalAbout, false);
			}
			if (about != null) ext.setAbout(about);
			
			// Load and set ActionFactory (required)
			Element afTag = epTag.getChild("action-factory", ns);
			if (afTag != null)
			{
				ActionFactory af = createFactory(afTag, classLoader, ActionFactory.class);
				ext.setActionFactory(af);
			}
			
			// Load and set RendererFactory (optional)
			Element rfTag = epTag.getChild("renderer-factory", ns);
			if (rfTag != null)
			{
				RendererFactory rf = createFactory(rfTag, classLoader, RendererFactory.class);
				ext.setRendererFactory(rf);
			}
		}
		extensions.addAll(temp);
		return numLoaded;
	}
	
	/**
		Create a new factory instance. The factory must be a class
		with a public, no-argument constructor. The factory class name
		is read from the &lt;factory-class&gt; tag and initialisation
		parameters from the &lt;parameters&gt; tag.
		
		@param factoryTag The root tag of the factory definition
		@param classLoader The classloader to use, or null to use
			the BASE core classloader
		@param factoryType Class type of the factory. The named class
			must implement or be a subclass of this class
		@return An initialised factory
	*/
	protected <F> F createFactory(Element factoryTag, ClassLoader classLoader, Class<F> factoryType)
		throws ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException, InstantiationException
	{
		Namespace ns = factoryTag.getNamespace();
		String factoryClassName = Values.getStringOrNull(factoryTag.getChildText("factory-class", ns));
		Class factoryClass = 
			ClassUtil.checkAndLoadClass(classLoader, factoryClassName, true, factoryType);
		F factory = factoryType.cast(factoryClass.newInstance());
		initBeanWithReflection(factory, factoryTag.getChild("parameters", ns));
		return factory;
	}
	
	/**
		Initialise a bean using reflection. For each child tag
		of the 'root' element, this method will check if the
		bean implements a setter method that is compatible with the
		child tag name. The method name should start with 'set' and
		then the tag name with the first letter capitalized. The method
		must take a single String argument. For example, if there is
		a child tag <code>&lt;image&gt;button.png&lt;/image&gt;</code> this will
		be converted to the method call: <code>setImage("button.png")</code>.
		<p>
		Tags that has no matching public setter method are ignored.
		
		@param bean The bean to initialize
		@param root The root element, if null nothing is done
	*/
	@SuppressWarnings("unchecked")
	protected void initBeanWithReflection(Object bean, Element root)
	{
		if (root == null) return;
		Class<?> beanClass = bean.getClass();
		
		List<Element> parameters = (List<Element>)root.getChildren();
		factoryParameters.put(bean, xmlOut.outputString(root));
		
		// Check for generic setParameter(String, String) method
		Method setParameter = getSetParameterMethod(beanClass);
		
		// List all child tags
		for (Element child : parameters)
		{
			// childName is the name of the setter method --> set<ChildName>
			String childName = child.getName();

			// The value passed as an argument to the setter method
			String value = Values.getStringOrNull(child.getText());
			if (value == null) continue;

			// Call generic setParameter(String, String) method
			if (setParameter != null) 
			{
				String converted = getConvertedValue(value, setParameter);
				try
				{
					setParameter.invoke(bean, childName, converted);
				}
				catch (IllegalAccessException ex)
				{}
				catch (InvocationTargetException ex)
				{}
			}
			
			Method setter = getSetterMethod(beanClass, childName);
			if (setter == null) continue;

			// Convert the input value
			String converted = getConvertedValue(value, setter);

			// Invoke the setter method -- ignore all errors
			try
			{
				setter.invoke(bean, converted);
			}
			catch (IllegalAccessException ex)
			{}
			catch (InvocationTargetException ex)
			{}
		}
	}
	
	/**
		Get the setter method for a given tag name.
		@param beanClass The class to look for the setter method in
		@param tagName The tag name in the XML file
		@return A Method object, or null if no method is found
	*/
	protected Method getSetterMethod(Class<?> beanClass, String tagName)
	{
		String methodName = getSetterMethodNameFromTag(tagName);
		Method setter = null;
		try
		{
			setter = beanClass.getMethod(methodName, String.class);
		}
		catch (NoSuchMethodException ex)
		{}
		return setter;
	}
	
	/**
		Check if the bean class has a <code>setParameter(String, String)</code>
		method and return it's reference if it has.
		@param beanClass The class to look for the method in
		@return A Method object, or null if no method is found
	*/
	protected Method getSetParameterMethod(Class<?> beanClass)
	{
		Method setParameter = null;
		try
		{
			setParameter = beanClass.getMethod("setParameter", String.class, String.class);
		}
		catch (NoSuchMethodException ex)
		{}
		return setParameter;
		
	}
	
	/**
		Convert the tag name to a setter method name.
		@param tagName The tag name
		@return The tag name prefixed with 'set' and the first letter captialized
	*/
	protected String getSetterMethodNameFromTag(String tagName)
	{
		String firstLetter = tagName.substring(0, 1).toUpperCase();
		String setterName = "set" + firstLetter + tagName.substring(1);
		return setterName;
	}

	protected String getConvertedValue(String original, Method method)
	{
		String converted = original;
		if (converters != null)
		{
			for (ValueConverter converter : converters)
			{
				converted = converter.convert(converted, method);
			}
		}
		return converted;
	}
	
}
