/**
	$Id: OverviewUtil.java 4767 2009-02-17 14:01:28Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Presets;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Scan;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.overview.filter.BasicItemFilter;
import net.sf.basedb.util.overview.Node;

/**
	Utility class for the overview generation functionality.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-17 15:01:28 +0100 (Tue, 17 Feb 2009) $
*/
public final class OverviewUtil
{

	private static Map<String, List<Validator>> validators;

	/**
		Get all validation rules grouped by relattionship. The key of the map
		is a title that groups related validation rules, for example:
		"Project default", "Missing items", etc.
		@return A map with a string key and list of validation rules 
			as value
	*/
	public static Map<String, List<Validator>> getValidators()
	{
		if (validators == null)
		{
			Map<String, List<Validator>> temp = new LinkedHashMap<String, List<Validator>>();
			temp.put("Project defaults", 
					Arrays.asList(new Validator[] 
					{
						Validator.NONDEFAULT_PLATFORM,
						Validator.NONDEFAULT_RAWDATATYPE, Validator.NONDEFAULT_ARRAYDESIGN, 
						Validator.NONDEFAULT_PROTOCOL, Validator.NONDEFAULT_HARDWARE, 
						Validator.NONDEFAULT_SOFTWARE
					})
				);
			temp.put("Missing items", 
					Arrays.asList(new Validator[] 
					{
						Validator.MISSING_BIOSOURCE, Validator.MISSING_SAMPLE, Validator.MISSING_EXTRACT,
						Validator.MISSING_LABELED_EXTRACT, Validator.MISSING_HYBRIDIZATION, 
						Validator.MISSING_SCAN, Validator.MISSING_ARRAYDESIGN, Validator.MISSING_PROTOCOL, 
						Validator.MISSING_HARDWARE, Validator.MISSING_SOFTWARE,
						Validator.MISSING_ARRAYSLIDE, Validator.MISSING_ARRAYBATCH
					})
				);
			temp.put("Annotations",
					Arrays.asList(new Validator[]
					{
						Validator.MISSING_MIAME_ANNOTATION, Validator.MISSING_FACTORVALUE, 
						Validator.MISSING_PARAMETER, Validator.ANNOTATION_IS_PARAMETER, 
						Validator.ANNOTATION_INVALID_VALUE, Validator.ANNOTATION_INVALID_ITEM,
						Validator.ANNOTATION_INHERIT_FROM_NONPARENT, 
						Validator.ANNOTATION_INHERIT_FROM_DENIEDPARENT,
						Validator.ANNOTATION_INHERIT_MULTIPLE
					})
				);
			
			temp.put("Files", 
					Arrays.asList(new Validator[]
					{
						Validator.DATAFILE_INVALID, Validator.DATAFILE_NOTVALIDATED,
						Validator.MISSING_DATAFILE, Validator.DATAFILE_NOTPLATFORM,
						Validator.DATAFILE_INVALID_ITEM, Validator.DENIED_DATAFILETYPE,
						Validator.DENIED_FILE
					})
				);
			
			temp.put("Denied access",
					Arrays.asList(new Validator[]
				   {
						Validator.DENIED_BIOSOURCE, Validator.DENIED_SAMPLE, Validator.DENIED_EXTRACT,
						Validator.DENIED_LABEL, Validator.DENIED_HYBRIDIZATION, Validator.DENIED_SCAN,
						Validator.DENIED_ARRAYDESIGN, Validator.DENIED_PLATFORM, Validator.DENIED_PROTOCOL, Validator.DENIED_HARDWARE,
						Validator.DENIED_SOFTWARE, Validator.DENIED_ARRAYSLIDE, Validator.DENIED_ARRAYBATCH
				   })
				);
			
			temp.put("Other",
					Arrays.asList(new Validator[]
					{
						Validator.NONMATHCING_ARRAYDESIGN, Validator.MULTIPLE_ARRAYDESIGNS, 
						Validator.INCORRECT_PROTOCOLTYPE, Validator.INCORRECT_SOFTWARETYPE,
						Validator.INCORRECT_HARDWARETYPE, Validator.INCORRECT_ARRAY_NUM,
						Validator.DUPLICATE_ARRAY_NUM, Validator.INCORRECT_NUM_LABELEDEXTRACTS,
						Validator.NONUNIQUE_LABEL, Validator.DIFFERENT_LABEL, 
						Validator.NONUNIQUE_NAME_FOR_TYPE, Validator.NONUNIQUE_NAME_GLOBAL,
						Validator.CIRCULAR_REFERENCE
					})
				);
			
			validators = Collections.unmodifiableMap(temp);
		}
		return validators;
	}

	/**
		Name of setting to store saved presets. DO NOT CHANGE the name since
		since it is used in the database to store the presets.
	*/
	private static final String presetsSetting = "experiment.overview.presets";
	
	/**
		Get the validation options presets for the currently logged in user.
		If changes are made to the presets call {@link #saveValidationPresets(SessionControl)}
		to make sure the changes are saved to the database.
		
		@param sc The session control for the logged in user
		@return The validation option presets
	*/
	public static Presets getValidationPresets(SessionControl sc)
	{
		Presets presets = (Presets)sc.getSessionSetting(presetsSetting);
		if (presets == null)
		{
			String xml = sc.getUserDefaultSetting(presetsSetting);
			presets = new Presets();
			if (xml != null)
			{
				presets.loadFrom(xml);
			}
			sc.setSessionSetting(presetsSetting, presets);
		}
		return presets;
	}
	
	/**
		Save the validation option presets for the logged in user. The presets
		are stored as user default settings 
		({@link SessionControl#setUserDefaultSetting(String, String)})
		with the name <code>experiment.overview.presets</code>
		@param sc The session control for the logged in user 
	*/
	public static void saveValidationPresets(SessionControl sc)
	{
		Presets presets = getValidationPresets(sc);
		if (presets != null)
		{
			sc.setUserDefaultSetting(presetsSetting, presets.asXml());
		}
	}
	
	/**
		Create a new overview for the given item and project. The created
		overview is cached and can be retrieved with {@link 
		#getCurrentOverview(SessionControl)} or {@link 
		#getNewOrCurrentOverview(DbControl, BasicItem, Project)}.
		
		@param dc A DbControl to use for database access
		@param item The item to generate the overview for
		@param project The project that should be used for validation
			of default items
		@return An overview
		@see #getNewOrCurrentOverview(DbControl, BasicItem, Project)
		@see #getCurrentOverview(SessionControl)
	*/
	public static GenericOverview getNewOverview(DbControl dc, BasicItem item, Project project)
	{
		SessionControl sc = dc.getSessionControl();
		GenericOverview overview = new GenericOverview(dc, item, project);
		sc.setSessionSetting("generic.overview", overview);
		return overview;
	}
	
	/**
		Get the current overview or create a new overview. This method will first
		check if a cached overview for the same item/project exists. If it does,
		the existing overview is returned, otherwise a new overview is created.
		
		@param dc A DbControl to use for database access
		@param item The item to generate the overview for
		@param project The project that should be used for validation
			of default items
		@return An overview
		@see #getNewOverview(DbControl, BasicItem, Project)
		@see #getCurrentOverview(SessionControl)
	*/
	public static GenericOverview getNewOrCurrentOverview(DbControl dc, BasicItem item, Project project)
	{
		GenericOverview overview = getCurrentOverview(dc.getSessionControl());
		if (overview != null)
		{
			boolean sameItem = item.equals(overview.getRootItem());
			boolean sameProject = (project==null && overview.getProject() == null) || 
				(project != null && project.equals(overview.getProject()));
			if (!sameItem || !sameProject) overview = null;
		}
		if (overview == null)
		{
			overview = getNewOverview(dc, item, project);
		}
		return overview;
	}
	
	/**
		Get the currently cached overview.
		@param sc The session control that contains the cached overview
		@return An overview, or null if no current overview exists
	*/
	public static GenericOverview getCurrentOverview(SessionControl sc)
	{
		return (GenericOverview)sc.getSessionSetting("generic.overview");
	}
	
	/**
		Get the project default item for a given project and type of item. 
		This method does the same thing as {@link Project#getDefaultItem(DbControl, net.sf.basedb.core.Project.Default)}
		but returns null instead of throwing an exception.
		@param dc The DbControl to use for database access
		@param project The current project, or null if not project is current
		@param defaultType The default type of item to get
		@return The default item, or null if no default item exists or if
			there was an error loading it
	*/
	public static BasicItem getProjectDefaultItem(DbControl dc, Project project, Project.Default defaultType)
	{
		BasicItem defaultItem = null;
		if (project != null)
		{
			try
			{
				defaultItem = project.getDefaultItem(dc, defaultType);
			}
			catch (BaseException ex)
			{}
		}
		return defaultItem;
	}
	
	/**
		Find the first parent node that has a reference to the given item.
		@param node The node to start looking at
		@param item The item to look for
		@return The first parent node (including the given) that has a reference to 
			the given item, or null if no such node exists
	*/
	public static Node getCircularRef(Node node, BasicItem item)
	{
		if (item.equals(node.getItem())) return node;
		return node.getFirstParent(new BasicItemFilter(item));
	}
	
	/**
		Get all annotation types that has been marked as "Required for MIAME".
		The annotation types found by this method will be cached in the overview
		context. Subsequent calls will return the cached annotation types.
		
		@param dc A DbControl to use for database access
		@param context The current overview context (will be used to prepare
			the query for annotation types; see {@link OverviewContext#initQuery(ItemQuery, String)}
		@param itemType An optional item type that is used to limit what
			annotation types to look for
		@return A set with the annotation types
	*/
	@SuppressWarnings("unchecked")
	public static Set<AnnotationType> getMiameAnnotationTypes(DbControl dc, OverviewContext context, Item itemType)
	{
		String cacheKey = "miame.annotationtypes";
		if (itemType != null) cacheKey += "." + itemType.name();
		Set<AnnotationType> miameAnnotationTypes = 
			(Set<AnnotationType>)context.getCachedObject(cacheKey);
		
		if (miameAnnotationTypes == null)
		{
			miameAnnotationTypes = new HashSet<AnnotationType>();
			context.setCachedObject(cacheKey, miameAnnotationTypes);
			
			// Load 'Required for MIAME' annotation types (parameter='false' must also be set)
			ItemQuery<AnnotationType> query = context.initQuery(AnnotationType.getQuery(itemType), "name");
			query.restrict(
				Restrictions.eq(
					Hql.property("requiredForMiame"), 
					Expressions.parameter("miameFlag", true, Type.BOOLEAN)
				)
			);
			query.restrict(
				Restrictions.eq(
					Hql.property("protocolParameter"), 
					Expressions.parameter("parameterFlag", false, Type.BOOLEAN)
				)
			);
			ItemResultIterator<AnnotationType> it = query.iterate(dc);
			while (it.hasNext())
			{
				miameAnnotationTypes.add(it.next());
			}
		}
		return miameAnnotationTypes;
	}
	
	/**
		Get all annotatable parent items for the given item. The parents that are found by
		this method will be cached in the current overview context. Subsequent calls will
		return the cached information.
		@param dc A DbControl to use for database access
		@param context The current overview context
		@param item The annotatable item that we should find parents for
		@return A set with the parents
	*/
	public static Set<Annotatable> getAnnotatableParents(DbControl dc, OverviewContext context, Annotatable item)
	{
		return getAnnotatableParents(dc, context, item, item);
	}
	
	/**
		The internal implementation needs to keep track of the special case when we
		start with raw bioassay as the root item. This is because when we reach
		the associated hybridization we should only load the labeled extracts
		that are on the same array index as the raw bioassay.
	*/
	@SuppressWarnings("unchecked")
	private static Set<Annotatable> getAnnotatableParents(DbControl dc, OverviewContext context, Annotatable item, Annotatable root)
	{
		String cacheKey = "annotatable.parent." + item.getType().name() + "." + item.getId();

		Hybridization hyb = null;
		int arrayNum = 1;
		// Special case when starting with a raw bioassay since we only
		// want to get parents from the sub-array with the same 
		// index = rba.getArrayNum()
		if (root instanceof RawBioAssay)
		{
			RawBioAssay rba = (RawBioAssay)root;
			arrayNum = rba.getArrayNum();
			if (item instanceof Scan)
			{
				cacheKey += "#" + arrayNum;
			}
			else if (item instanceof Hybridization)
			{
				hyb = (Hybridization)item;
				cacheKey += "#" + arrayNum;
			}
		}
	
		Set<Annotatable> parents = (Set<Annotatable>)context.getCachedObject(cacheKey);
		if (parents == null)
		{
			parents = new HashSet<Annotatable>();
			context.setCachedObject(cacheKey, parents);
			Set<Annotatable> temp = hyb != null ?
				hyb.getAnnotatableParents(arrayNum) : item.getAnnotatableParents();
			if (temp != null)
			{
				for (Annotatable a : temp)
				{
					parents.add(a);
					parents.addAll(getAnnotatableParents(dc, context, a, root));
				}
			}
		}
		return parents;
	}
	
}
