/**
	$Id: ArrayDesignLoader.java 4767 2009-02-17 14:01:28Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for array designs. Array designs are
	an endpoint node with the forward-loading direction going to 
	{@link ArrayDesign}:s -&gt; {@link ArrayBatch}:s. An array design can 
	also be loaded as a property node from a raw bioassay.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-17 15:01:28 +0100 (Tue, 17 Feb 2009) $
*/
public class ArrayDesignLoader
	extends BasicItemNodeLoader<ArrayDesign>
{

	public ArrayDesignLoader()
	{
		super(Item.ARRAYDESIGN, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<ArrayDesign>("arraydesign", "Array design"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create a reverse-loading array design node from an array batch node.
		@return An array design node, or null if the batch doesn't have a design
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node batchNode)
	{
		NodeFactory<ArrayDesign> nf = getNodeFactory(dc, context);
		ArrayDesign design = null;
		boolean denied = false;
		try
		{
			design = ((ArrayBatch)batchNode.getItem(dc)).getArrayDesign();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node designNode = createItemNode(nf, design, design, denied, 
				batchNode, ChildNodeDirection.REVERSE);
		return designNode;
	}
	/**
		Create an array design property node from a raw bioassay node.
		@return An array design node, or null if the batch doesn't have a design
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		NodeFactory<ArrayDesign> nf = getNodeFactory(dc, context);
		ArrayDesign design = null;
		boolean denied = false;
		try
		{
			design = ((RawBioAssay)rawBioAssayNode.getItem(dc)).getArrayDesign();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node designNode = createItemNode(nf, design, design, denied, 
				rawBioAssayNode, ChildNodeDirection.PROPERTY);
		return designNode;
	}	
	// --------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of an array design. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Data files: {@link DataFileLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Platform: {@link PlatformLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node designNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, designNode);
		getNodeLoader(context, Item.PLATFORM).createPropertyNode(dc, context, designNode);
		getNodeLoader(context, Item.FILESETMEMBER).createPropertyNode(dc, context, designNode);
	}

	/**
		Loads the array batch nodes that has been created with the given design node.
		@see ArrayBatchLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node designNode)
	{
		getNodeLoader(context, Item.ARRAYBATCH).createForwardNode(dc, context, designNode);
	}
	// ---------------------------------------

}
