/**
	$Id: AnnotationValidator.java 4761 2009-02-16 08:27:07Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;

/**
	Validator implementation for annotations. The node containing
	the annotations is expected to live in a folder-type node that
	has an {@link Annotatable} item as it's parent. 
	Validation rules:
	<ul>
	<li>Annotation is a protocol parameter: {@link Validator#ANNOTATION_IS_PARAMETER}
	<li>Annotation is not valid for item type: {@link Validator#ANNOTATION_INVALID_ITEM}
	<li>Annotation has invalid value: {@link Validator#ANNOTATION_INVALID_VALUE}
	<li>Access denied to annotation type: {@link Validator#DENIED_ANNOTATIONTYPE}
	<li>Annotation is inherited from non-parent item: {@link Validator#ANNOTATION_INHERIT_FROM_NONPARENT}
	<li>Multiple annotations of same annotation type: {@link Validator#ANNOTATION_INHERIT_MULTIPLE}
	<li>Missing a 'Required by MIAME' annotation: {@link Validator#MISSING_MIAME_ANNOTATION}
	<li>Missing experimental factor value: {@link Validator#MISSING_FACTORVALUE}
	</ul>
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-16 09:27:07 +0100 (Mon, 16 Feb 2009) $
*/
public class AnnotationValidator
	extends BasicNodeValidator<Annotation>
{
	
	public AnnotationValidator()
	{
		super(null, null);
	}
	
	/*
		From BasicNodeValidator class
		-----------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		if (parentNode.getNodeType() == Node.Type.FOLDER) parentNode = parentNode.getParent();
		Annotatable parent = (Annotatable)parentNode.getItem();
		Item parentItemType = parentNode.getItemType();
		AnnotationSet annotationSet = parent.getAnnotationSet();
		
		Annotation a = (Annotation)node.getItem();
		boolean isPrimary = a.belongsTo(annotationSet);

		AnnotationType at = null;
		try
		{
			at = a.getAnnotationType();
			
			// Check that the (primary) annotation is not a protocol parameter
			if (isPrimary && at.isProtocolParameter())
			{
				context.createFailure(Validator.ANNOTATION_IS_PARAMETER, node, 
					"Annotation '" + at.getName() + "' is a protocol parameter",
					new Fix("Remove annotation value", (BasicItem)parent, at, false),
					new Fix("Change 'parameter' flag of annotation type", at)
				);
			}

			// Check if the (primary) annotation can be used on this item type
			if (isPrimary && !at.isEnabledForItem(parentItemType))
			{
				context.createFailure(Validator.ANNOTATION_INVALID_ITEM, node, 
					"Annotation '" + at.getName() + "' is invalid for item type: " + parentItemType, 
					new Fix("Remove annotation value", (BasicItem)parent, at, false),
					new Fix("Add '" + parentItemType.name() + "' to annotation type", at)
				);
			}

			// Check that the (primary) annotation values are valid
			if (isPrimary)
			{
				try
				{
					List<?> values = a.getValues();
					if (values != null)
					{
						for (Object value : values)
						{
							at.validateAnnotationValue(value);
						}
					}
				}
				catch (InvalidDataException ex)
				{
					context.createFailure(Validator.ANNOTATION_INVALID_VALUE, node, ex.getMessage(),
						new Fix("Change annotation value", (BasicItem)parent, at, false),
						new Fix("Change annotation type restrictions", at)
					);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{
			// When the user doesn't have access to the annotation type 
			// of the annotation we can't validate or check other things
			context.createFailure(Validator.DENIED_ANNOTATIONTYPE, node, null);
		}

		// If the annotation is inherited from a 'parent item', verify that 
		// it is a reachable parent
		if (!isPrimary)
		{
			Set<Annotatable> reachableParents = OverviewUtil.getAnnotatableParents(dc, context, parent);
			try
			{
				Annotatable inheritedFrom = a.getAnnotationSet().getItem();
				if (!reachableParents.contains(inheritedFrom))
				{
					context.createFailure(Validator.ANNOTATION_INHERIT_FROM_NONPARENT, node, 
							"Inheriting annotation '" + (at != null ? 
								at.getName() : a.getValueType().toString()) + "' from non-parent",
							new Fix("Inherit from another item", (BasicItem)parent, at, true)
						);
				}
			}
			catch (PermissionDeniedException ex)
			{
				context.createFailure(Validator.ANNOTATION_INHERIT_FROM_DENIEDPARENT, node, 
						"Inheriting annotation '" + (at != null ? 
							at.getName() : a.getValueType().toString()) + "' from denied parent",
						new Fix("Inherit from another item", (BasicItem)parent, at, true)
					);
			}
		}
	}
	
	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, Node parentNode) 
	{
		super.postValidateFolder(dc, context, folderNode, parentNode);
		Annotatable parent = (Annotatable)parentNode.getItem();
		
		// Check for 'Required by MIAME' annotations and duplicates
		Set<AnnotationType> miame = OverviewUtil.getMiameAnnotationTypes(dc, context, parentNode.getItemType());
		Set<AnnotationType> all = new HashSet<AnnotationType>();
		Set<AnnotationType> duplicates = new HashSet<AnnotationType>();
		if (folderNode != null && folderNode.numChildren() > 0)
		{
			for (Node child : folderNode.getChildren())
			{
				Annotation a = (Annotation)child.getItem();
				try
				{
					AnnotationType at = a.getAnnotationType();
					if (!all.add(at)) duplicates.add(at);
				}
				catch (PermissionDeniedException ex)
				{}
			}
		}
		
		// Report duplicate annotation types
		if (duplicates.size() > 0)
		{
			for (AnnotationType dup : duplicates)
			{
				context.createFailure(Validator.ANNOTATION_INHERIT_MULTIPLE, parentNode, 
					"Duplicate annotations of type '" + dup.getName() + "'",
					new Fix("Modify inherited annotations", parentNode.getItem(), dup, true)
				);
			}
		}

		// Report missing 'Required by MIAME' annotations
		if (miame != null && miame.size() > 0)
		{
			for (AnnotationType at : miame)
			{
				if (!all.contains(at))
				{
					context.createFailure(Validator.MISSING_MIAME_ANNOTATION, parentNode, 
							"No value for annotation '" + at.getName() + "' (required fo MIAME)", 
							new Fix("Add value for annotation '" + at.getName() + "'", (BasicItem)parent, at, false),
							new Fix("Remove 'Required for MIAME' flag on '" + at.getName() + "'", at)
						);
				}
			}
		}
		
		// Check for missing experimental factor values (if we have an experiment in the parent path)
		if (parentNode.getItemType() == Item.RAWBIOASSAY)
		{
			Node experimentNode = parentNode.getFirstParent(new ItemTypeFilter(Item.EXPERIMENT));
			if (experimentNode != null)
			{
				Experiment exp = (Experiment)experimentNode.getItem(dc);
				ItemQuery<AnnotationType> query = context.initQuery(exp.getExperimentalFactors());
				ItemResultIterator<AnnotationType> it = query.iterate(dc);
				while (it.hasNext())
				{
					AnnotationType factor = it.next();
					if (!all.contains(factor))
					{
						Fix fix = factor.isEnabledForItem(Item.RAWBIOASSAY) ?
							new Fix("Add value to experimental factor", (BasicItem)parent, factor, false) :
							new Fix("Inherit annotation from a parent item", (BasicItem)parent, factor, true);
						context.createFailure(Validator.MISSING_FACTORVALUE, parentNode, 
							"Missing value for experimental factor '" + factor.getName() + "'", fix);
					}
				}
			}
		}
	}

	/**
		@return Always "null"
	*/
	@Override
	protected Fix getMissingItemFix(Node parentNode)
	{
		return null;
	}
	// ----------------------------
	
}
