/**
	$Id: DataFileValidator.java 4779 2009-02-20 09:01:06Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreEnabled;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformFileType;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for data files. The node containing
	the data files is expected to live in a folder-type node that
	has a {@link FileStoreEnabled} item as it's parent (if there 
	are more than one file) or as an immediate child node to
	it's parent item (if there is only one file).
	Validation rules:
	<ul>
	<li>Denied access to file: {@link Validator#DENIED_FILE}
	<li>Denied access to file type: {@link Validator#DENIED_DATAFILETYPE}
	<li>File is not a valid file: {@link Validator#DATAFILE_INVALID}
	<li>File has not been validated: {@link Validator#DATAFILE_NOTVALIDATED}
	<li>File type doesn't match parent item type: {@link Validator#DATAFILE_INVALID_ITEM}
	<li>File type is not part of used platform: {@link Validator#DATAFILE_NOTPLATFORM}
	<li>Missing a required file: {@link Validator#MISSING_DATAFILE}
	</ul>
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-20 10:01:06 +0100 (Fri, 20 Feb 2009) $
*/
public class DataFileValidator
	extends BasicNodeValidator<FileSetMember>
{
	
	public DataFileValidator()
	{
		super(null, null);
	}
	
	/*
		From BasicNodeValidator class
		-----------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		FileSetMember member = (FileSetMember)node.getItem(dc);
		// The parent node is either a folder (if we have more than one file)
		// or the actual parent (for a single file)
		FileStoreEnabled parentItem = (FileStoreEnabled)(parentNode.getNodeType() == Node.Type.FOLDER ? 
				parentNode.getParent().getItem() : parentNode.getItem());
		
		// Check that we have access to the file and it's file type
		DataFileType fileType = null;
		File file = null;
		try
		{
			fileType = member.getDataFileType();
		}
		catch (PermissionDeniedException ex)
		{}
		try
		{
			file = member.getFile();
		}
		catch (PermissionDeniedException ex)
		{}
		if (fileType == null)
		{
			context.createFailure(Validator.DENIED_DATAFILETYPE, node, null);
		}
		if (file == null)
		{
			context.createFailure(Validator.DENIED_FILE, node, null);
		}
		
		// Check if a validator has determined that the file IS invalid
		if (Boolean.FALSE.equals(member.isValid()))
		{
			String errorMsg = fileType == null ? "Data file " : fileType.getName();
			errorMsg += " is not valid: " + member.getErrorMessage();
			context.createFailure(Validator.DATAFILE_INVALID, node, errorMsg,
				new Fix("Select another file", (BasicItem)parentItem, member)
			);
		}
		
		if (fileType != null)
		{
			// Check if the file hasn't been validated even if a validator exists
			if (fileType.getValidatorClass() != null && member.isValid() == null)
			{
				context.createFailure(Validator.DATAFILE_NOTVALIDATED, node, 
					fileType.getName() + " has not been validated", 
					new Fix("Validate the file / Select another file", (BasicItem)parentItem, member));
			}

			// Check that the file type is matching the parent item's type
			if (fileType.getItemType() != parentItem.getType())
			{
				context.createFailure(Validator.DATAFILE_INVALID_ITEM, node, 
					"Data file '" + fileType.getName() + "' can't be used on item type: " + parentItem.getType(), 
					new Fix("Remove file", (BasicItem)parentItem, member)
				);
			}

			// Check that the file type is part of the used platform/variant
			try
			{
				Platform platform = parentItem.getPlatform();
				PlatformVariant variant = parentItem.getVariant();
				if (platform != null)
				{
					// The item to apply fixes on, variant or platform
					String pvName = null;
					BasicItem pvItem = null;
					if (variant != null)
					{
						pvItem = variant;
						pvName = variant.getName();
					}
					else
					{
						pvItem = platform;
						pvName = platform.getName();
					}
					if (platform.getFileType(fileType, variant) == null)
					{
						context.createFailure(Validator.DATAFILE_NOTPLATFORM, node, 
								fileType.getName() + " can't be used with '" + pvName + "' platform", 
								new Fix("Remove '" + fileType.getName() + "' from " + parentItem.getType().toString().toLowerCase(), 
									(BasicItem)parentItem, fileType),
								new Fix("Add '" + fileType.getName() +"' to '" + pvName + "' platform", pvItem, fileType)
							);
					}
				}
			}
			catch (PermissionDeniedException ex)
			{}		
		}
	}
	
	/**
		@return Always "null"
	*/
	@Override
	protected Fix getMissingItemFix(Node parentNode)
	{
		return null;
	}

	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, 
			Node folderNode, Node parentNode)
	{
		super.postValidateFolder(dc, context, folderNode, parentNode);
		FileStoreEnabled parentItem = (FileStoreEnabled)parentNode.getItem();
		
		// Check that we have linked a file for all file types that
		// are required by the platform
		try
		{
			Platform platform = parentItem.getPlatform();
			PlatformVariant variant = parentItem.getVariant();
			BasicItem pvItem = variant != null ? variant : platform;
			if (platform == null) return;
			
			// Create a query that returns all required platform file types
			ItemQuery<PlatformFileType> query = platform.getFileTypes(variant, variant == null);
			query.join(Hql.innerJoin("dataFileType", "dft"));
			query.restrict(
				Restrictions.eq(
					Hql.property("required"), 
					Expressions.parameter("isRequired", true, Type.BOOLEAN)
				)
			);
			query.restrict(
				Restrictions.eq(
					Hql.property("dft", "itemType"), 
					Expressions.integer(parentNode.getItemType().getValue())
				)
			);
			if (parentItem.hasFileSet())
			{
				// Add a restriction that filters out the files that we already have
				Restriction myFileSet = Restrictions.eq(Hql.property("mbr", "fileSet.id"),
						Hql.entity(parentItem.getFileSet()));
				query.join(Hql.leftJoin("dft", "fileSetMembers", "mbr", myFileSet, false));
				query.restrict(Restrictions.eq(Hql.alias("mbr"), null));
			}
			ItemResultIterator<PlatformFileType> it = query.iterate(dc);
			while (it.hasNext())
			{
				PlatformFileType pft = it.next();
				DataFileType dft = pft.getDataFileType();
				context.createFailure(Validator.MISSING_DATAFILE, parentNode, 
						"Missing '" + dft.getName() + "'", 
						new Fix("Add '" + dft.getName() + "' to " + parentItem.getType().toString().toLowerCase(), 
								(BasicItem)parentItem, dft),
						new Fix("Remove 'Required' flag from '" + dft.getName() + "'", pvItem, dft)
					);
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}
	// ----------------------------
		
}
