/*
	$Id: PrintMapFlatFileImporter.java 4513 2008-09-11 20:06:36Z jari $

	Copyright (C) 2005 Johan Enell
	Copyright (C) 2006 Johan Enell, Jari Hakkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArrayDesignBlock;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BlockInfo;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.FeatureBatcher;
import net.sf.basedb.core.FeatureIdentificationMethod;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.FileType;
import net.sf.basedb.core.IntegerUtil;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Plate;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.Well;
import net.sf.basedb.core.data.FeatureData;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.parser.FlatFileParser.Line;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

/**
	A plugin that imports an {@link ArrayDesign} from a printmap. The
	formats supported are TAM and MwBr
	
	@base.modified $Date: 2008-09-11 22:06:36 +0200 (Thu, 11 Sep 2008) $
	@author Enell
	@version 2.0
*/
public class PrintMapFlatFileImporter 
	extends AbstractFlatFileImporter
	implements InteractivePlugin
{
	private static final About about = 
		new AboutImpl
		(
			"Print map importer",
			"This plugin is used to import arraydesign from a print map.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);

	private static final Set<GuiContext> guiContexts = 
		Collections.singleton(new GuiContext(Item.ARRAYDESIGN, GuiContext.Type.ITEM));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	protected static final PluginParameter<String> missingWellErrorParameter = new PluginParameter<String>(
			"missingWellError",
			"Well not found",
			"How to handle errors that are caused by the data not matching the well coordinates on the plates. " +
				"If not specified the default error handling is used.\n\n"+
				"skip = Skip the current data line and continue\n"+
				"fail = Stop with an error message",
				new StringParameterType(255, null, false, 1, 0, 0, 
					Arrays.asList( new String[] { "skip", "fail"} ))
			);

	
	private RequestInformation configureJob;
	
	private DbControl dc;
	private FeatureBatcher batcher;
	private Map<BlockInfo, ArrayDesignBlock> blocks;
	private Map<Integer, MetaBlock> metaBlocks;

	private ArrayDesign arrayDesign;
	private FileSetMember printMapMember;
	private ItemParameterType<ArrayDesign> arrayDesignType;
	private PluginParameter<ArrayDesign> arrayDesignParameter;
	
	private int numFeatures;
	private int numBlocks;
	private Section section;

	private FormatType formatType;

	private int expectedNumBlocks;
	private int spotSize;
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/	
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request write access to ArrayDesign:s, read access to Plate:s, Reporter:s,
	 	DataFileType:s, Platform:s and File:s.
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.ARRAYDESIGN, null, EnumSet.of(Permission.WRITE)));
			permissions.add(new Permissions(Item.PLATE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.PLATFORM, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DATAFILETYPE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/	

	/**
		Returns null if the item is a {@link ArrayDesign} that doesn't already have 
		any features and isn't an affy design.
	*/
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof ArrayDesign))
		{
			message = "The object is not an ArrayDesign: " + item;
		}
		else
		{
			ArrayDesign ad = (ArrayDesign)item;
			if (ad.isFileOnlyPlatform())
			{
				message = "Array design is using a file-only platform";
			}
			else if (ad.getNumDbFeatures() > 0)
			{
				throw new PermissionDeniedException("The array design already has features");
			}
			else if (ad.getNumPlates() == 0)
			{
				message = "The array design hasn't been connected to any plates.";
			}
			else 
			{
				ad.checkPermission(Permission.WRITE);
			}
		}
		return message;
	}

	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}

	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will ask for
		parser regular expressions and column mappings.
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		a file and the array design that features should be added to.
		
		@param command the command for the request
	*/
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.setPropertyFilter(FileType.getPropertyFilter(FileType.PRINT_MAP));
			fileContext.setPropertyFilter(getPrimaryLocationFilter());
		}
		return requestInformation;
	}

	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN}
		 and {@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				storeValue(job, request, fileParameter);
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(job, request, arrayDesignParameter);

				// Error handling parameters
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, invalidUseOfNullErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				storeValue(job, request, missingWellErrorParameter);
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				// TODO - maybe check file size to make a better estimate
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------
	/*
		From the AbstractFlatFileImporter class
		-------------------------------------------
	*/
	/**
		Check that the file is a TAM or MwBr file.
		@return TRUE if a FormatName header is present and contains "TAM" or "MwBr", FALSE
			otherwise
	*/
	@Override
	protected boolean isImportable(FlatFileParser ffp)
	{
		String formatName = ffp.getHeader("FormatName");
		return formatName != null && (formatName.contains("TAM") || formatName.contains("MwBr"));
	}

	/**
		Create a {@link DbControl} and a {@link FeatureBatcher}. 
		Check that the {@link ArrayDesign} doesn't already have features.
	*/
	@Override
	protected void begin(FlatFileParser ffp) 
		throws BaseException
	{
		super.begin(ffp);
		dc = sc.newDbControl();
		arrayDesign = (ArrayDesign)job.getValue("arrayDesign");
		arrayDesign = ArrayDesign.getById(dc, arrayDesign.getId());
		File reporterMapFile = (File)job.getValue("file");
		printMapMember = FileStoreUtil.setGenericDataFile(dc, arrayDesign, 
				FileType.PRINT_MAP, DataFileType.GENERIC_PRINT_MAP, reporterMapFile);
		if (arrayDesign.getNumDbFeatures() > 0)
		{
			throw new PermissionDeniedException("The array design already has features: " + arrayDesign.getName()+" ["+arrayDesign.getId()+"]");
		}
		if (arrayDesign.getNumPlates() == 0)
		{
			throw new InvalidDataException("The array design hasn't been connected to any plates: " + arrayDesign.getName()+" ["+arrayDesign.getId()+"]");
		}
		String method = (String)job.getValue("missingWellError");
		if (method != null)
		{
			addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
		}
		
		batcher = arrayDesign.getFeatureBatcher(FeatureIdentificationMethod.COORDINATES, job.getJob());
		numFeatures = 0;
		numBlocks = 0;
		metaBlocks = new HashMap<Integer, MetaBlock>();
		blocks = new HashMap<BlockInfo, ArrayDesignBlock>();
	}

	@Override
	protected void handleHeader(Line line) 
		throws BaseException
	{
		if (section == Section.BLOCK)
		{
			MetaBlock b = metaBlocks.get(section.getBlockNumber());
			setBlockValue(b, line);
		}
		else if (section == Section.FILE_INFORMATION)
		{
			String name = line.name();
			String value = line.value();
			if (name.equalsIgnoreCase("FormatName"))
			{
				if (value.equalsIgnoreCase("TAM"))
				{
					formatType = FormatType.TAM;
				}
				else if (value.equalsIgnoreCase("MwBr"))
				{
					formatType = FormatType.MWBR;
				}
				else
				{
					throw new BaseException("FormatName must be TAM or MwBr");
				}
			}
			else if (name.equalsIgnoreCase("BlockCount"))
			{
				expectedNumBlocks = Values.getInt(value);
			}
			else if (name.equalsIgnoreCase("SpotSize"))
			{
				spotSize = Values.getInt(value);
			}
		}
	}

	@Override
	protected void handleSection(Line line) 
		throws BaseException
	{
		section = Section.fromString(line.name().toLowerCase());
		if (section == Section.BLOCK)
		{
			int blockNumber = section.getBlockNumber();
			MetaBlock b = metaBlocks.get(blockNumber);
			if (b == null && blockNumber > Integer.MIN_VALUE)
			{
				b = new MetaBlock(blockNumber);
				metaBlocks.put(blockNumber, b);
				numBlocks++;
			}
		}
		else if(section== Section.MAPPING)
		{
			// create all [BlockInfo, ArrayDesignBlocks] and clear MetaBlock
			for (MetaBlock b : metaBlocks.values())
			{
				BlockInfo bi = new BlockInfo(b.getBlock(), b.getMetaGridX(), b.getMetaGridY());
				ArrayDesignBlock adb = arrayDesign.addArrayDesignBlock(bi);
				adb.setFeatureDiameter(spotSize);
				adb.setBlockSizeX(b.getBlockSizeX());
				adb.setBlockSizeY(b.getBlockSizeY());
				adb.setOriginX(b.getOriginX());
				adb.setOriginY(b.getOriginY());
				adb.setSpacingX(b.getSpacingX());
				adb.setSpacingY(b.getSpacingY());
				blocks.put(bi, adb);
			}
			metaBlocks.clear();
		}
	}
	
	@Override
	protected void handleData(Data data) 
		throws BaseException
	{
		if (section == Section.MAPPING && data.columns() >= 11)
		{
			String metaX = data.get(0);
			String metaY = data.get(1);
			int plateNumber = Integer.parseInt(data.get(5)) - 1;
			int plateRow = Integer.parseInt(data.get(6)) - 1;
			int plateColumn = Integer.parseInt(data.get(7)) - 1;
			String block = data.get(10);
			
			if (metaX == null) throw new InvalidUseOfNullException("metaX");
			if (metaY == null) throw new InvalidUseOfNullException("metaY");
			
			BlockInfo bi;
			if (block != null)
			{
				bi = new BlockInfo(Integer.parseInt(block), Integer.parseInt(metaX), Integer.parseInt(metaY));
			}
			else
			{
				bi = new BlockInfo(Integer.parseInt(metaX), Integer.parseInt(metaY));
			}
			ArrayDesignBlock adb = blocks.get(bi);
			if (adb == null)
			{
				throw new ItemNotFoundException("ArrayDesignBlock [coordinates=" + bi + "]");
			}
			Well w = findWell(plateNumber, plateRow, plateColumn);
			if (w == null) throw new ItemNotFoundException("Well [plate="+plateNumber+"; row="+plateRow+"; column="+plateColumn+"]");

			FeatureData f = batcher.newFeature(adb, w);
			int col;
			int row;
			if (this.formatType == FormatType.MWBR)
			{
				row = Integer.parseInt(data.get(2));
				col = Integer.parseInt(data.get(3));
			}
			else
			{
				col = Integer.parseInt(data.get(2));
				row = Integer.parseInt(data.get(3));
			}
			f.setColumn(col);
			f.setRow(row);
			batcher.insert(f);
			numFeatures++;
		}
	}

	@Override
	protected void end(boolean success) 
		throws BaseException
	{
		if (printMapMember != null) printMapMember.setValid(success, null);
		if (blocks != null) blocks.clear();
		try
		{
			if (batcher != null) batcher.close();
			if (success && dc != null)
			{
				dc.commit();
			}
		}
		finally
		{
			if (dc != null) dc.close();
			super.end(success);
		}
	}

	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String skipped = skippedLines > 0 ? "; " + skippedLines + " features skipped due to errors" : "";
		return numFeatures + " feature(s) inserted; " +
			numBlocks + " block(s) inserted" +
			(numBlocks == expectedNumBlocks ? "" : "; expected "+ expectedNumBlocks) +
			skipped;
	}

	@Override
	protected FlatFileParser getInitializedFlatFileParser()
		throws BaseException
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setSectionRegexp(Pattern.compile("\\[(.+)\\]"));
		ffp.setHeaderRegexp(Pattern.compile("(.+)=,(.*)"));
		ffp.setDataSplitterRegexp(Pattern.compile(","));
		ffp.setDataFooterRegexp(Pattern.compile(""));
		ffp.setMinDataColumns(12);
		return ffp;
	}
	// -------------------------------------------

	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			arrayDesignType = new ItemParameterType<ArrayDesign>(ArrayDesign.class, null, true, 1, null);
			arrayDesignParameter = new PluginParameter<ArrayDesign>
			(
				"arrayDesign",
				"Array Design",
				"The Array Design assigned to the imported features",
				arrayDesignType
			);
			
			// Parameters for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(arrayDesignParameter);

			ArrayDesign design = null;
			int currentArrayDesignId = sc.getCurrentContext(Item.ARRAYDESIGN).getId();
			List<File> printMapFiles = null;
			dc = sc.newDbControl();
			try
			{
				design = ArrayDesign.getById(dc, currentArrayDesignId);
				printMapFiles = FileStoreUtil.getGenericDataFiles(dc, design, FileType.PRINT_MAP);
			}
			catch (Throwable t)
			{}
			finally
			{
				if (dc != null) dc.close();
			}
			
			// The print map file to import from - if a file already hase
			// been attached to the array design use it as a default choice
			PluginParameter<File> fileParameter = new PluginParameter<File>(
				"file",
				"Print map file",
				"The file with printing information to import features from",
				new FileParameterType(printMapFiles == null || printMapFiles.isEmpty() ? 
					null : printMapFiles.get(0), true, 1)
			);
			parameters.add(fileParameter);
			parameters.add(Parameters.charsetParameter(null, null, null));
			
			// Error handling parameters
			parameters.add(errorSection);
			parameters.add(defaultErrorParameter);
			parameters.add(missingWellErrorParameter);
			parameters.add(invalidUseOfNullErrorParameter);
			parameters.add(numberFormatErrorParameter);
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Select file to import features from",
				"Here you select which file to import the features from.",
				parameters
			);
		}
		return configureJob;
	}
	
	private Well findWell(int platePosition, int row, int column)
	{
		Plate p = arrayDesign.getPlate(platePosition);
		if (p != null)
		{
			return p.getWell(row, column);
		}
		return null;
	}
	
	private void setBlockValue(MetaBlock block, Line line)
		throws BaseException
	{
		if (block == null)
		{
			return;
		}
		String name = line.name();
		Integer value;
		try
		{
			value = new Integer(line.value());
		}
		catch (NumberFormatException e)
		{
			throw new BaseException(e);
		}
		
		if (name.equalsIgnoreCase("MetaGridX"))
		{
			block.setMetaGridX(value);
		}
		else if (name.equalsIgnoreCase("MetaGridY"))
		{
			block.setMetaGridY(value);
		}
		else if (name.equalsIgnoreCase("OriginX"))
		{
			block.setOriginX(value);
		}
		else if (name.equalsIgnoreCase("OriginY"))
		{
			block.setOriginY(value);
		}
		else if (name.equalsIgnoreCase("BlockSizeX"))
		{
			block.setBlockSizeX(value);
		}
		else if (name.equalsIgnoreCase("BlockSizeY"))
		{
			block.setBlockSizeY(value);
		}
		else if (name.equalsIgnoreCase("SpacingX"))
		{
			block.setSpacingX(value);
		}
		else if (name.equalsIgnoreCase("SpacingY"))
		{
			block.setSpacingY(value);
		}
	}
	
	private class MetaBlock
	{
		private int block;

		private Integer metaGridX = 0;
		
		private Integer metaGridY = 0;
		
		private Integer originX = 0;
		
		private Integer originY = 0;
		
		private Integer blockSizeX = 1;
		
		private Integer blockSizeY = 1;
		
		private Integer spacingX = 0;
		
		private Integer spacingY = 0;

		public MetaBlock(int block)
		{
			this.block = block;
		}

		public Integer getBlockSizeX()
		{
			return blockSizeX;
		}

		public void setBlockSizeX(Integer blockSizeX)
		{
			this.blockSizeX = blockSizeX;
		}

		public Integer getBlockSizeY()
		{
			return blockSizeY;
		}

		public void setBlockSizeY(Integer blockSizeY)
		{
			this.blockSizeY = blockSizeY;
		}

		public Integer getMetaGridX()
		{
			return metaGridX;
		}

		public void setMetaGridX(Integer metaGridX)
		{
			this.metaGridX = metaGridX;
		}

		public Integer getMetaGridY()
		{
			return metaGridY;
		}

		public void setMetaGridY(Integer metaGridY)
		{
			this.metaGridY = metaGridY;
		}

		public Integer getOriginX()
		{
			return originX;
		}

		public void setOriginX(Integer originX)
		{
			this.originX = originX;
		}

		public Integer getOriginY()
		{
			return originY;
		}

		public void setOriginY(Integer originY)
		{
			this.originY = originY;
		}

		public Integer getSpacingX()
		{
			return spacingX;
		}

		public void setSpacingX(Integer spacingX)
		{
			this.spacingX = spacingX;
		}

		public Integer getSpacingY()
		{
			return spacingY;
		}

		public void setSpacingY(Integer spacingY)
		{
			this.spacingY = spacingY;
		}

		public int getBlock()
		{
			return block;
		}
		
		@Override
		public String toString()
		{
			return "MetaBlock["+block+";"+metaGridX+";"+metaGridY+";"+originX+";"+originY+";"+spacingX+";"+spacingY+"]";
		}
	}

	private enum Section
	{
		FILE_INFORMATION("fileinformation"),
		SOURCE("source"),
		TOOL("tool"),
		TARGET("target"),
		SLIDES("slides"),
		BLOCK("block"),
		MAPPING("mapping");
		
		private int blockNumber = Integer.MIN_VALUE;
		
		private String name;
		
		private static final Map<String, Section> stringMapping = new HashMap<String, Section>();
		/**
			Maps an <code>String</code> to a section.
		*/
		static
		{
			for(Section section : Section.values())
			{
				stringMapping.put(section.getName(), section);
			}
		}
		
		public static Section fromString(String s)
		{
			if (s.startsWith("block"))
			{
				Section block = stringMapping.get("block");
				block.setBlockNumber(IntegerUtil.getInt(s.substring(5), Integer.MIN_VALUE));
				return block;
			}
			return stringMapping.get(s);
		}
		
		private Section(String s)
		{
			this.name = s;
		}
		
		public String getName()
		{
			return name;
		}
		
		private void setBlockNumber(int blockNumber)
		{
			this.blockNumber = blockNumber;
		}
		
		public int getBlockNumber()
		{
			return this.blockNumber;
		}
	}

	private enum FormatType
	{
		TAM, MWBR;
	}
}
