/**
	$Id: AbstractItemImporter.java 4736 2009-01-23 15:00:00Z martin $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.io.InputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.HardwareType;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerUtil;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Label;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.NumberOutOfRangeException;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.ProtocolType;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.Scan;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.SoftwareType;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.StringTooLongException;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.ParameterValuesWrapper;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.plugins.AbstractFlatFileImporter;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.ClassMapErrorHandler;
import net.sf.basedb.util.error.ErrorHandler;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	An abstract base class for plug-ins that import items in batch. This class implements
	job configuration, file parsing, logging, error handling, etc. Actual implementations
	have to provide item-specific parts, such as:
	<ul>
	<li>The context the plug-in can be used = the type of item it can import.
		Eg, implement the {@link InteractivePlugin#getGuiContexts()} method.
	<li>The methods that can be used for item identification: 
		{@link #getIdMethods()}
	<li>Job configuration parameters for column mappings to item-specific
		properties: {@link #addMoreColumnMappingParameters(List)}
	<li>Query used to find items of the specific type: {@link #createItemQuery()}
	<li>Create actual column mappers used to extract data from the file
		when running the plug-in: {@link #createColumnMappers(FlatFileParser, boolean)}
	<li>Create a new item and populate it with data from the file: 
		{@link #createItem(DbControl, FlatFileParser.Data)}
	<li>Update an item with data from the file: {@link #updateItem(DbControl, BasicItem, FlatFileParser.Data)}
		and {@link #updateMultiLineItem(DbControl, BasicItem, FlatFileParser.Data, int)}
	</ul>

	@author nicklas
	@version 2.8
	@base.modified $Date: 2009-01-23 16:00:00 +0100 (Fri, 23 Jan 2009) $
*/
public abstract class AbstractItemImporter<I extends BasicItem>
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample
{
	
	// second step in the wizard
	private static final String CONFIGURE_OPTIONS = "configure_options";
	
	protected static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	protected static final StringParameterType requiredColumnMapping = new StringParameterType(255, null, true);

	protected static final PluginParameter<String> identificationSection = new PluginParameter<String>(
			"idSection",
			"Item identification",
			"Options related to how find existing items.", 
			null
		);
	
	protected static final PluginParameter<String> internalIdColumnMapping = new PluginParameter<String>(
			"internalIdColumnMapping",
			"Internal ID",
			"Mapping that picks the BASE internal ID of the item from the data columns. This " +
			"parameter is required if the 'Internal ID' identification method is selected, " +
			"otherwise it is optional. Example: \\BASE ID\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> nameColumnMapping = new PluginParameter<String>(
			"nameColumnMapping",
			"Name",
			"Mapping that picks the name of the item from the data columns. This " +
			"parameter is required if the 'Name' identification method is selected, " +
			"otherwise it is optional. Example: \\Name\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> descriptionColumnMapping = new PluginParameter<String>(
			"descriptionColumnMapping",
			"Description",
			"Mapping that picks the description of the item from the data columns." +
			"Example: \\Description\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> externalIdColumnMapping = new PluginParameter<String>(
			"externalIdColumnMapping",
			"External ID",
			"Mapping that picks the external ID of the item from the data columns. This " +
			"parameter is required if the 'External ID' identification method is selected, " +
			"otherwise it is optional. Example: \\External ID\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> protocolColumnMapping = new PluginParameter<String>(
			"protocolColumnMapping",
			"Protocol",
			"Mapping that picks the name or ID of a protocol from the data columns. The plug-in will first " +
			"try to find a protocol with the given name. If none is found and the value is numeric it will " +
			"try to load by internal ID. " +
			"Example: \\Protocol\\",
			optionalColumnMapping
			);

	private static final PluginParameter<String> itemNotFoundErrorParameter = new PluginParameter<String>(
			"itemNotFoundError",
			"Item not found",
			"What to do if the plug-in can't find an item with the name or ID from the file.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);
	
	private static final PluginParameter<String> itemExistsErrorParameter = new PluginParameter<String>(
			"itemExistsError",
			"Item already exists",
			"What to do if the plug-in finds an existing item with the name or ID from the file.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);

	
	private static final PluginParameter<String> multipleItemsFoundErrorParameter = new PluginParameter<String>(
			"multipleItemsFoundError",
			"Multiple items found",
			"What to do if the plug-in finds more than one item with the same name or ID.\n\n"+
			"skip = Skip the current data line and continue\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);

	
	private static final PluginParameter<String> referenceNotFoundErrorParameter = new PluginParameter<String>(
			"referenceNotFoundError",
			"Referenced item not found",
			"What to do if the plug-in can't find a referenced item, eg. a protocol, etc.\n\n"+
			"ignore = Do not set/update the reference\n" +
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);

	private static final PluginParameter<String> multipleReferencesFoundErrorParameter = new PluginParameter<String>(
			"multipleReferencesFoundError",
			"Multiple referenced items found",
			"What to do if the plug-in finds more than one referenced item (eg. protocol) with the same name or ID.\n\n"+
			"ignore = Do not set/update the reference\n" +
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);
	
	private static final PluginParameter<String> noWritePermissionToItemErrorParameter = new PluginParameter<String>(
			"noWritePermissionToItemError",
			"Not granted write permission to item",
			"What to do if the plug-in finds an existing item but has no write permission to it.\n\n"+
			"skip = Skip the current data line and continue.\n"+
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "fail"} ))
		);
	
	private static final PluginParameter<String> noUsePermissionToReferenceErrorParameter = new PluginParameter<String>(
			"noUsePermissionToReferenceError",
			"Not granted use permission to reference",
			"What to do if the plug-in finds a reference but has no permission to use it.\n\n"+
			"ignore = Do not set/update the reference\n" +
			"fail = Stop with an error message",
			new StringParameterType(255, null, false, 1, 0, 0, 
				Arrays.asList( new String[] { "ignore", "fail"} ))
		);
	public AbstractItemImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	// --------------------------------------------
	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = getGuiContexts().contains(context) ? 
			null : "This plug-in can't be used in context: " + context;
		return message;
	}
	
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN) || command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				boolean forJob = command.equals(Request.COMMAND_CONFIGURE_JOB);
				RequestInformation ri = getConfigureParserParameters(context, true);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				ParameterValues wrapper = new ParameterValuesWrapper(request, job, configuration, forJob);
				if (forJob)
				{
					// Check that file can be parsed with given regular expressions
					FlatFileParser ffp = getInitializedFlatFileParser(wrapper);
					File f = (File)request.getParameterValue("file");
					InputStream in = f.getDownloadStream(0);
					ffp.setInputStream(in, getCharset(request));
					ffp.setDefaultNumberFormat(getNumberFormat());
					FlatFileParser.LineType result = ffp.parseHeaders();
					in.close();
					if (result == FlatFileParser.LineType.UNKNOWN)
					{
						response.setError("The file could not be parsed with the given settings, " +
							"no data found after " + ffp.getLineCount() + " lines.", null);
						return;
					}
				}
				
				// Check that a mapping has been provided for the selected idMethod
				IdMethod idMethod = getIdMethod((String)request.getParameterValue("idMethod"));
				if (idMethod == null)
				{
					response.setError("Unknown method for item identification: " + 
							request.getParameterValue("idMethod"), null);
					return;
				}
				PluginParameter<?> idMapping = ri.getParameter(idMethod.getColumnMappingParameterName());
				if (request.getParameterValue(idMapping.getName()) == null)
				{
					response.setError("No colum mapping provided for item identification: " + 
							idMapping.getLabel(), null);
					return;
				}

				// Check the mapping expressions
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter pp : ri.getParameters())
				{
					if (pp.getName().contains("ColumnMapping"))
					{
						checkColumnMapping((String)request.getParameterValue(pp.getName()), allowComplex, pp.getLabel());
					}
				}
				
				// Everything is ok, save values
				/*
				if (forJob) 
				{
					storeValue(wrapper, request, fileParameter);
					storeValue(wrapper, request, ri.getParameter("mode"));
				}
				storeValue(wrapper, request, dataHeaderRegexpParameter);
				storeValue(wrapper, request, dataSplitterRegexpParameter);
				storeValue(wrapper, request, trimQuotesParameter);
				storeValue(wrapper, request, ignoreRegexpParameter);
				storeValue(wrapper, request, dataFooterRegexpParameter);
				storeValue(wrapper, request, minDataColumnsParameter);
				storeValue(wrapper, request, maxDataColumnsParameter);
				storeValue(wrapper, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(wrapper, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				storeValue(wrapper, request, ri.getParameter(Parameters.DATE_FORMAT_PARAMETER));
				
				// Identification method
				storeValue(wrapper, request, ri.getParameter("idMethod"));
				if (forJob)
				{
					storeValue(wrapper, request, ri.getParameter("includeMine"));
					storeValue(wrapper, request, ri.getParameter("includeSharedTo"));
					storeValue(wrapper, request, ri.getParameter("includeInProject"));
					storeValue(wrapper, request, ri.getParameter("includeOthers"));
				}
				
				// Column mappings
				storeValue(wrapper, request, complexMappings);
				*/
				for (PluginParameter pp : ri.getParameters())
				{
//					if (pp.getName().contains("ColumnMapping"))
					if (pp.getParameterType() != null)
					{
						storeValue(wrapper, request, ri.getParameter(pp.getName()));
					}
				}
//				storeValue(wrapper, request, ri.getParameter("dataDirectory"));
				if (forJob)
				{
					response.setContinue(CONFIGURE_OPTIONS);
				}
				else
				{
					response.setDone("Plug-in configuration completed");
				}
			}
			else if (command.equals(CONFIGURE_OPTIONS))
			{
				RequestInformation ri = getConfigureOptionsParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
			
				// Error handling
				storeValue(job, request, ri.getParameter(Parameters.LOGFILE_PARAMETER));
				response.setDryRun(Boolean.TRUE.equals(request.getParameterValue(Parameters.DRY_RUN_PARAMETER)));
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, itemNotFoundErrorParameter);
				storeValue(job, request, itemExistsErrorParameter);
				storeValue(job, request, multipleItemsFoundErrorParameter);
				storeValue(job, request, referenceNotFoundErrorParameter);
				storeValue(job, request, multipleReferencesFoundErrorParameter);
				storeValue(job, request, ri.getParameter(Parameters.INVALID_USE_OF_NULL_ERROR));
				storeValue(job, request, stringTooLongErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				storeValue(job, request, numberOutOfRangeErrorParameter);
				storeValue(job, request, noWritePermissionToItemErrorParameter);
				storeValue(job, request, noUsePermissionToReferenceErrorParameter);
				
				// Save context info
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);		
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}	
	
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigureParserParameters(context, false);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureParserParameters(context, true);
		}
		else if (command.equals(CONFIGURE_OPTIONS))
		{
			requestInformation = getConfigureOptionsParameters(context);
		}
		return requestInformation;
	}
	// ------------------------------------
	/*
		From the AbstractFlatFileImporter class
		---------------------------------------
	*/
	private DbControl dc;
	private FlatFileParser ffp;
	private NumberFormat numberFormat;
	private DateFormatter dateFormatter;
	
	private IdMethod idMethod;
	private String idMapping;
	private Mapper idMapper;
	private ItemQuery<I> itemQuery;

	private String lastIdentifier;
	private I lastItem;
	private int numMultiLines;
	private boolean skippedLast;
	
	// Error handling
	private boolean dryRun;
	private ClassMapErrorHandler errorHandler;
	private boolean cropStrings;
	private boolean createNotFoundItems;
	private boolean failIfNotFoundItems;
	private boolean updateExistingItems;
	private boolean failIfItemExists;
	private boolean failIfMultipleFoundItems;
	private boolean failIfNotFoundReference;
	private boolean failIfMultipleFoundReferences;
	private boolean failIfNoWritePermissionToItem;
	private boolean failIfNoUsePermissionToReference;
	
	// Status report
	private int numUpdated;
	private int numCreated;
	private int numError;
	private int numNotFound;
	private int numExisted;
	
	@Override
	/**
		The flat file parser is initialised with settings from either the job or configuration.
	*/
	protected FlatFileParser getInitializedFlatFileParser()
		throws BaseException
	{
		ParameterValuesWrapper wrapper = new ParameterValuesWrapper(null, job, configuration, true);
		return getInitializedFlatFileParser(wrapper);
	}
	
	/**
		Don't use AbstractFlatFileImporter to handle errors.
	*/
	@Override
	protected void setUpErrorHandling()
	{}
	
	/**
		Setup error handling and pre-load some of the configuration options.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
	{
		this.ffp = ffp;
		this.numberFormat = ffp.getDefaultNumberFormat();
		String dateFormat = Values.getString((String)job.getValue(Parameters.DATE_FORMAT_PARAMETER), "yyyy-MM-dd");
		this.dateFormatter = new DateFormatter(dateFormat);
		this.dateFormatter.getDateFormat().setLenient(false);
		this.idMethod = getIdMethod((String)job.getValue("idMethod"));
		this.idMapping = (String)job.getValue(idMethod.getColumnMappingParameterName());
		this.itemCache = new HashMap<String, BasicItem>();
		String mode = (String)job.getValue("mode");
		boolean updateMode = mode == null || mode.contains("update");
		boolean createMode = mode == null || mode.contains("create");
		
		// Error handling, we don't use the error handling in AbstractFlatFileImporter
		this.dryRun = job.getJob().isDryRun();
		this.createNotFoundItems = createMode || "create".equals(getErrorOption("itemNotFoundError"));
		this.failIfNotFoundItems = !createMode && "fail".equals(getErrorOption("itemNotFoundError"));
		this.updateExistingItems = updateMode || "update".equals(getErrorOption("itemExistsError"));
		this.failIfItemExists = !updateMode && "fail".equals(getErrorOption("itemExistsError"));
		this.failIfMultipleFoundItems = "fail".equals(getErrorOption("multipleItemsFoundError"));
		this.failIfNotFoundReference = "fail".equals(getErrorOption("referenceNotFoundError"));
		this.failIfMultipleFoundReferences = "fail".equals(getErrorOption("multipleReferencesFoundError"));
		this.failIfNoWritePermissionToItem = "fail".equals(getErrorOption("noWritePermissionToItemError"));
		this.failIfNoUsePermissionToReference = "fail".equals(getErrorOption("noUsePermissionToReferenceError"));
		
		this.cropStrings = "crop".equals(getErrorOption("stringTooLongError"));

		ErrorHandler defaultErrorHandler = new SimpleErrorHandler("skip".equals(job.getValue("defaultError")));
		errorHandler = new ClassMapErrorHandler(defaultErrorHandler);
		errorHandler.addErrorHandler(StringTooLongException.class, 
				getErrorHandler((String)job.getValue("stringTooLongError"), defaultErrorHandler));
		errorHandler.addErrorHandler(NumberOutOfRangeException.class, 
				getErrorHandler((String)job.getValue("numberOutOfRangeError"), defaultErrorHandler));
		errorHandler.addErrorHandler(NumberFormatException.class,
				getErrorHandler((String)job.getValue("numberFormatError"), defaultErrorHandler));
		errorHandler.addErrorHandler(InvalidUseOfNullException.class, 
				getErrorHandler((String)job.getValue(Parameters.INVALID_USE_OF_NULL_ERROR), defaultErrorHandler));

	}
	
	/**
		Setup column mapping. Creates DbControl and query to find items.
	*/
	@Override
	protected void beginData()
	{
		this.dc = sc.newDbControl();
		
		// Mapper to get name or external ID
		this.idMapper = getMapper(ffp, idMapping, null, null);
		createColumnMappers(ffp, cropStrings);
		
		// Include options
		Set<Include> includes = EnumSet.noneOf(Include.class);
		if (Boolean.TRUE.equals((Boolean)job.getValue("includeMine"))) includes.add(Include.MINE);
		if (!Boolean.FALSE.equals((Boolean)job.getValue("includeSharedTo"))) includes.add(Include.MINE);
		if (!Boolean.FALSE.equals((Boolean)job.getValue("includeInProject"))) includes.add(Include.IN_PROJECT);
		if (Boolean.TRUE.equals((Boolean)job.getValue("includeOthers"))) includes.add(Include.OTHERS);
		itemQuery = idMethod.prepareQuery(dc, createItemQuery());
		itemQuery.include(includes);
	}
	
	@SuppressWarnings("unchecked")
	@Override
	protected void handleData(Data data)
		throws BaseException
	{
		// Find the item(s) to update; mapper gives us name, internal ID, or some other identifer
		String identifier = idMapper.getValue(data);
		String cacheKey = itemQuery.getItemType().name() + ":" + identifier;
		boolean sameAsLast = StringUtil.isEqualOrNull(identifier, lastIdentifier);
		if (!sameAsLast)
		{
			if (lastItem != null) doneWithItem(dc, lastItem, numMultiLines);
			numMultiLines = 0;
			lastIdentifier = identifier;
		}
		else
		{
			numMultiLines++;
			if (skippedLast)
			{
				// If the first line in a multi-line entry was skipped, skip the remaining lines also
				log("Skipped: " + idMethod + "=" + identifier, data);
				return;
			}
		}
		lastItem = null;
		I item = null;
		if (identifier != null)
		{
			// Try to find an existing item
			// We look in the cache first
			if (itemCache.containsKey(cacheKey))
			{
				item = (I)itemCache.get(cacheKey);
			}
			else
			{
				// and only query the database if not found
				List<I> items = idMethod.find(dc, itemQuery, identifier);
				// pick out items on which write permission is granted
				List<I> writableItems = new ArrayList<I>();
				for (I i : items)
				{
					if (i.hasPermission(Permission.WRITE))
					{
						writableItems.add(i);
					}
				}
				if (writableItems.size() == 1)
				{
					item = writableItems.get(0);
					itemCache.put(cacheKey, item);
				}
				else if (writableItems.size() > 1)
				{
					// If we find multiple items, we either throw an error or do nothing
					if (failIfMultipleFoundItems)
					{
						throw new BaseException("Found " + writableItems.size() + " items [" + idMethod + "=" + identifier + "]");
					}
					log("Found " + writableItems.size() + " items: " + idMethod + "=" + identifier, data);
					return;
				}
				else if (items.size() > 0) 
				{
					// Found item(s) but none with write permission to. Either throw an exception or do nothing
					if (failIfNoWritePermissionToItem)
					{
						throw new PermissionDeniedException(Permission.WRITE, 
								itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]");
					}
					log("Not granted write permission to " + itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]", data);
					return;
				}
			}
		}

		if (item == null)
		{
			// No item was found
			if (failIfNotFoundItems)
			{
				throw new ItemNotFoundException(itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]");
			}
			else if (createNotFoundItems)
			{
				// Create a new item
				try
				{
					item = createItem(dc, data);
					dc.saveItem(item);
					itemCache.put(cacheKey, item);
					lastItem = item;
					log("Created: " + idMethod + "=" + identifier, data);
					skippedLast = false;
					numCreated++;
				}
				catch (Throwable t)
				{
					try
					{
						errorHandler.handleError(t);
						log("Skipped: " + idMethod + "=" + identifier, data, t);
						skippedLast = true;
						numError++;
						// The error should be ignored if we get passed the above line
					}
					catch (Throwable t2)
					{
						// handleData doesn't allow us to throw any Throwable
						throw t2 instanceof RuntimeException ?
							(RuntimeException)t2 : new BaseException(t);
					}
				}
			}
			else
			{
				log("Not found: " + idMethod + "=" + identifier, data);
				skippedLast = true;
				numNotFound++;
			}
		}
		else
		{
			// Found a single item
			itemCache.put(cacheKey, item);
			lastItem = item;
			if (failIfItemExists)
			{
				throw new ItemAlreadyExistsException(itemQuery.getItemType()+"[" + idMethod + "=" + identifier + "]");
			}
			else if (updateExistingItems)
			{
				try
				{
					if (sameAsLast)
					{
						updateMultiLineItem(dc, item, data, numMultiLines);
					}
					else
					{
						updateItem(dc, item, data);
						numUpdated++;
					}
					log("Updated: " + idMethod + "=" + identifier, data);
					skippedLast = false;
				}
				catch (Throwable t)
				{
					try
					{
						errorHandler.handleError(t);
						log("Skipped: " + idMethod + "=" + identifier, data);
						skippedLast = true;
						numError++;
						// The error should be ignored if we get passed the above line
					}
					catch (Throwable t2)
					{
						// handleData doesn't allow us to throw any Throwable
						throw t2 instanceof RuntimeException ?
								(RuntimeException)t2 : new BaseException(t);
					}
				}
			}
			else
			{
				// Ignore existing items
				log("Exists: " + idMethod + "=" + identifier, data);
				skippedLast = true;
				numExisted++;
			}
		}
	}
	
	@Override
	protected void end(boolean success)
	{
		try
		{
			if (success)
			{
				if (lastItem != null) doneWithItem(dc, lastItem, numMultiLines);
				if (!dryRun) dc.commit();
			}
		}
		catch (BaseException ex)
		{
			success = false;
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
			super.end(success);
		}
	}
	
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String msg = "Parsed " + ffp.getParsedLines() + " lines";
		if (numCreated > 0) msg += "; " + numCreated + " items created";
		if (numUpdated > 0) msg += "; " + numUpdated + " items updated";
		if (numExisted > 0) msg += "; " + numExisted + " items skipped because the already exists";
		if (numError > 0) msg += "; " + numError + " items skipped because of invalid values";
		if (numNotFound > 0) msg += "; " + numNotFound + " items skipped because they could not be found";
		return msg;
	}
	
	// ------------------------------------

	/*
		Methods to be implemented by subclasses
		---------------------------------------
	*/
	/**
		Get an array with all identification methods that can be used to find
		items. The default implementation returns an array with the {@link
		PropertyIdMethod#DEFAULT_METHODS}. Subclasses that import items that can use
		other methods, should override this method.
	*/
	protected IdMethod[] getIdMethods()
	{
		return PropertyIdMethod.DEFAULT_METHODS;
	}
	
	/**
		Add extra job parameters that the subclass needs. This
		method is called during configuration of the plug-in and
		the default implementation does nothing.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters)
	{}
	
	/**
		Add extra parameters to the "Parser settings" section that the subclass 
		needs. This method is called during configuration of the plug-in and
		the default implementation does nothing.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreParsingParameters(List<PluginParameter<?>> parameters)
	{}

	/**
		Add column mapping parameters for item-specific properties. A subclass must
		add parameters for all properties, including name and description. This class
		has already created some useful parameters for this purpose, for example:
		{@link #nameColumnMapping} and {@link #descriptionColumnMapping}. This
		method is called during configuration of the plug-in.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{}
	
	/**
		Add extra parameters to the "Error handling" section that the subclass 
		needs. This method is called during configuration of the plug-in and
		the default implementation does nothing.
		
		@param parameters The list of parameters to add the column
			mapping parameters to
	*/
	protected void addMoreErrorParameters(List<PluginParameter<?>> parameters)
	{}
	
	/**
		Create a new query that returns items of the importable type. The
		query should be a plain query not including any restrictions or 
		other elements. This will automatically be added as needed by the
		core implementation. This method is called during execution of the 
		plug-in. Typically, the result from <code>getQuery()</code> should
		be returned, eg. <code>Sample.getQuery()</code>.
		@return A query
	*/
	protected abstract ItemQuery<I> createItemQuery();

	/**
		Create mappers that can be used to extract data values from the file
		that is currently being parsed by the <code>ffp</code> {@link FlatFileParser}.
		This method is called during initialisation of the plug-in when the 
		{@link FlatFileParser#parseHeaders()} has successfully found the start of 
		data. Use the {@link #getMapper(FlatFileParser, String, Integer, Mapper)}
		to create column mappers. Example:
		
		<pre class="code">
nameMapper = getMapper(ffp, job.getValue("nameColumnMapping"), 
   cropStrings ? Nameable.MAX_NAME_LENGTH : null, null);
</pre>
		
		
		@param ffp The parser that is parsing the data file
		@param cropStrings If strings that are too long to fit in the database
			should be cropped or not (=generate an error)
	*/
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{}

	/**
		Create a new item and populate it with data from the current line. Do
		not call {@link DbControl#saveItem(BasicItem)}. This is done automatically
		by the core implementation. 
		
		<pre class="code">
BioSource bs = BioSource.getNew(dc);
// Set properties; see example on updateItem method
return bs;
</pre>
		<p>
		If the current data line doesn't have data for all required 
		properties, this method should throw an {@link InvalidUseOfNullException}
		exception.
		
		<p>
		This method is called during import when current item can be found
		and if the user has selected to create missing item. 
		
		@param dc The DbControl to use for database access
		@param data The current data line
		@return A new item
	*/
	protected abstract I createItem(DbControl dc, FlatFileParser.Data data);
	
	/**
		Update an item with data from the current line. Example:
		<pre class="code">
if (nameMapper != null) item.setName(nameMapper.getValue(data));
if (descriptionMapper != null) item.setDescription(descriptionMapper.getValue(data));
...
</pre>

		This method is called during import to update an existing item.

		@param dc The DbControl to use for database access
		@param item The item to update
		@param data The current data line
	*/
	protected abstract void updateItem(DbControl dc, I item, FlatFileParser.Data data);
	
	/**
		Update an item with data from a multi-line entry. Items which can have multiple
		parents should have data for the other parents on lines following
		the first line.
		
		@param dc The DbControl to use for database access
		@param item The item to update
		@param data The current data line
		@param multiLineNum The number of the multi-line entry, starting with 1
			for the first multi-line
	*/
	protected void updateMultiLineItem(DbControl dc, I item, FlatFileParser.Data data, int multiLineNum)
	{}
	
	/**
		Called when the importer is finished with an item. This method is useful when working
		with multi-line entries, when the subclass needs to do some additional work after
		all data has been set. NOTE! This assumes that the lines in the data file are sorted
		in the correct order and the no lines for the same item will follow later on.
		@param dc The DbControl to use for database access
		@param item The item to update
		@param numMultiLines The number of the multi-line entries, starting with 1
			for the first multi-line (0 for single-line entries)
	 */
	protected void doneWithItem(DbControl dc, I item, int numMultiLines)
	{}
	// ---------------------------------------
	
	/**
		Get the id method to use for finding items. This implementation 
		searches the methods returned by {@link #getIdMethods()} for a method
		with the given name. If no method can be found {@link PropertyIdMethod#NAME}
		is returned.
	*/
	protected IdMethod getIdMethod(String method)
	{
		for (IdMethod im : getIdMethods())
		{
			if (im.getMethod().equals(method))
			{
				return im;
			}
		}
		return PropertyIdMethod.NAME;
	}
	
	/**
		Parse a string to a date using the date format specified by the 
		'Date format' parameter. This is a utility method that subclasses can
		use to extract date values, since the {@link Mapper} interface doesn't
		supports it.
	*/
	protected Date parseDate(String date)
	{
		return dateFormatter.parseString(date);
	}
	
	private ItemQuery<Protocol> protocolQuery;
	/**
		Same as {@link #findProtocol(DbControl, IdMethod, String, ProtocolType)} with
		a null protocol type.
	*/
	protected Protocol findProtocol(DbControl dc, IdMethod idMethod, String identifier)
	{
		return findProtocol(dc, idMethod, identifier, null);
	}
	
	/**
		Find a protocol with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		<p>
		NOTE! If this method is called with a non-null protocol type parameter and 
		no protocol is found the query will be retried without any protocol type.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@param type The protocol type of the protocol, or null if the type doesn't matter
		@return A protocol, or null if no item could be found
		@since 2.9
	*/
	protected Protocol findProtocol(DbControl dc, IdMethod idMethod, String identifier, ProtocolType type)
	{
		if (identifier == null) return null;
		if (protocolQuery == null) 
		{
			// where ... (protocolType = :protocolType AND checkType = 1) || checkType = 0
			Expression checkType = Expressions.parameter("checkType");
			protocolQuery = initReferenceQuery(dc, idMethod, Protocol.getQuery());
			protocolQuery.restrictPermanent(
				Restrictions.or(
					Restrictions.and(
						Restrictions.eq(Hql.property("protocolType"), Expressions.parameter("protocolType")),
						Restrictions.eq(checkType, Expressions.integer(1))
					),
					Restrictions.eq(checkType, Expressions.integer(0))
				)
			);
		}
		protocolQuery.setParameter("protocolType", type == null ? 0 : type.getId(), Type.INT);
		protocolQuery.setParameter("checkType", type == null ? 0 : 1, Type.INT);
		Protocol protocol = findReferencedItem(dc, idMethod, protocolQuery, identifier, true);
		if (protocol == null)
		{
			protocolQuery.setParameter("checkType", 0, Type.INT);
			protocol = findReferencedItem(dc, idMethod, protocolQuery, identifier, false);
		}
		return protocol;
	}

	private ItemQuery<Label> labelQuery;
	/**
		Find a label with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A label, or null if no item could be found
	*/
	protected Label findLabel(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (labelQuery == null) labelQuery = initReferenceQuery(dc, idMethod, Label.getQuery());
		return findReferencedItem(dc, idMethod, labelQuery, identifier, false);
	}

	private ItemQuery<Hardware> hybStationQuery;
	/**
		Find a hybridization station with a given identifier. This is a utility method that
		subclasses can use when creating or updating items. 
		
		<p>
		NOTE! A scanner is a {@link Hardware} item with it type set to 
		{@link HardwareType#HYBRIDIZATION_STATION}.
		
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A scanner, or null if no item could be found
	*/
	protected Hardware findHybridizationStation(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (hybStationQuery == null) 
		{
			hybStationQuery = initReferenceQuery(dc, idMethod, Hardware.getQuery());
			hybStationQuery.restrictPermanent(
				Restrictions.eq(Hql.property("hardwareType"), 
				Expressions.integer(SystemItems.getId(HardwareType.HYBRIDIZATION_STATION)))
			);
		}
		return findReferencedItem(dc, idMethod, hybStationQuery, identifier, false);
	}
	
	private ItemQuery<Hardware> scannerQuery;
	/**
		Find a scanner with a given identifier. This is a utility method that
		subclasses can use when creating or updating items. 
		
		<p>
		NOTE! A scanner is a {@link Hardware} item with it type set to 
		{@link HardwareType#SCANNER}.
		
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A scanner, or null if no item could be found
	*/
	protected Hardware findScanner(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (scannerQuery == null) 
		{
			scannerQuery = initReferenceQuery(dc, idMethod, Hardware.getQuery());
			scannerQuery.restrictPermanent(
				Restrictions.eq(Hql.property("hardwareType"), 
				Expressions.integer(SystemItems.getId(HardwareType.SCANNER)))
			);
		}
		return findReferencedItem(dc, idMethod, scannerQuery, identifier, false);
	}

	private ItemQuery<Hardware> robotQuery;
	/**
		Find a print robot with a given identifier. This is a utility method that
		subclasses can use when creating or updating items. 
		
		<p>
		NOTE! A print robot is a {@link Hardware} item with it type set to 
		{@link HardwareType#PRINT_ROBOT}.
		
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A print robot, or null if no item could be found
	*/
	protected Hardware findPrintRobot(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (robotQuery == null) 
		{
			robotQuery = initReferenceQuery(dc, idMethod, Hardware.getQuery());
			robotQuery.restrictPermanent(
				Restrictions.eq(Hql.property("hardwareType"), 
				Expressions.integer(SystemItems.getId(HardwareType.PRINT_ROBOT)))
			);
		}
		return findReferencedItem(dc, idMethod, robotQuery, identifier, false);
	}
	
	private ItemQuery<Software> featureQuery;
	/**
		Find a feature extraction software with a given identifier. This is a utility 
		method that subclasses can use when creating or updating items. 
		
		<p>
		NOTE! A feature extraction software is a {@link Software} item with it's
		type set to {@link SoftwareType#FEATURE_EXTRACTION}.
		
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A software item, or null if no item could be found
	*/
	protected Software findFeatureExtractionSoftware(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (featureQuery == null) 
		{
			featureQuery = initReferenceQuery(dc, idMethod, Software.getQuery());
			featureQuery.restrictPermanent(
				Restrictions.eq(Hql.property("softwareType"), 
				Expressions.integer(SystemItems.getId(SoftwareType.FEATURE_EXTRACTION)))
			);
		}
		return findReferencedItem(dc, idMethod, featureQuery, identifier, false);
	}

	private ItemQuery<Platform> platformQuery;
	/**
		Find a platform with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A platform, or null if no item could be found
	*/
	protected Platform findPlatform(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (platformQuery == null) 
		{
			platformQuery = initReferenceQuery(dc, idMethod, Platform.getQuery());
		}
		return findReferencedItem(dc, idMethod, platformQuery, identifier, false);
	}
	/**
		Get the default platform of the currently active project.
		@return The active platform, or null if no project is active or
			the project doesn't specify a platform
		@since 2.9
	*/
	protected Platform getProjectDefaultPlatform(DbControl dc)
	{
		if (sc.getActiveProjectId() == 0) return null;
		Project activeProject = Project.getById(dc, sc.getActiveProjectId());
		return (Platform)activeProject.getDefaultItem(dc, Project.Default.PLATFORM);
	}
	
	/**
		Get the default platform variant of the currently active project.
		@return The active platform variant, or null if no project is active or
			the project doesn't specify a variant
		@since 2.9
	*/
	protected PlatformVariant getProjectDefaultVariant(DbControl dc)
	{
		if (sc.getActiveProjectId() == 0) return null;
		Project activeProject = Project.getById(dc, sc.getActiveProjectId());
		return (PlatformVariant)activeProject.getDefaultItem(dc, Project.Default.PLATFORM_VARIANT);
	}
	
	private ItemQuery<PlatformVariant> variantQuery;
	/**
		Find a platform variant with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param platform If given, searched only variants of this platform
		@param identifier The identifier protocol
		@return A platform, or null if no item could be found
	*/
	protected PlatformVariant findVariant(DbControl dc, Platform platform, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (variantQuery == null) 
		{
			variantQuery = initReferenceQuery(dc, idMethod, PlatformVariant.getQuery());
			// variant.platform = :platform or 1 = :anyplatform
			variantQuery.restrict(
				Restrictions.or(
					Restrictions.eq(Hql.property("platform.id"), Expressions.parameter("platform")),
					Restrictions.eq(Expressions.integer(1), Expressions.parameter("anyplatform"))
				)
			);
		}
		variantQuery.setParameter("platform", platform == null ? 0 : platform.getId(), Type.INT);
		variantQuery.setParameter("anyplatform", platform == null ? 1 : 0, Type.INT);
		return findReferencedItem(dc, idMethod, variantQuery, identifier, false);
	}
	
	/**
		Get the default raw data type of the currently active project.
		@return The active raw data type, or null if no project is active or
			the project doesn't specify raw data type
		@since 2.9
	*/
	protected RawDataType getProjectDefaultRawDataType(DbControl dc)
	{
		if (sc.getActiveProjectId() == 0) return null;
		Project activeProject = Project.getById(dc, sc.getActiveProjectId());
		return activeProject.getDefaultRawDataType();
	}

	
	private ItemQuery<DataFileType> fileTypeQuery;
	/**
		Find a data file type with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A data file, or null if no item could be found
	*/
	protected DataFileType findDataFileType(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (fileTypeQuery == null) 
		{
			fileTypeQuery = initReferenceQuery(dc, idMethod, DataFileType.getQuery());
		}
		return findReferencedItem(dc, idMethod, fileTypeQuery, identifier, false);
	}
	
	private ItemQuery<BioSource> bioSourceQuery;
	/**
		Find a biosource with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A biosource, or null if no item could be found
	*/
	protected BioSource findBioSource(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (bioSourceQuery == null) bioSourceQuery = initReferenceQuery(dc, idMethod, BioSource.getQuery());
		return findReferencedItem(dc, idMethod, bioSourceQuery, identifier, false);
	}

	private ItemQuery<Sample> sampleQuery;
	/**
		Find a sample with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A sample, or null if no item could be found
	*/
	protected Sample findSample(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (sampleQuery == null) sampleQuery = initReferenceQuery(dc, idMethod, Sample.getQuery());
		return findReferencedItem(dc, idMethod, sampleQuery, identifier, false);
	}

	private ItemQuery<Extract> extractQuery;
	/**
		Find an extract with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An extract, or null if no item could be found
	*/
	protected Extract findExtract(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (extractQuery == null) extractQuery = initReferenceQuery(dc, idMethod, Extract.getQuery());
		return findReferencedItem(dc, idMethod, extractQuery, identifier, false);
	}

	private ItemQuery<LabeledExtract> labeledExtractQuery;
	/**
		Find a labled extract with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A labeled extract, or null if no item could be found
	*/
	protected LabeledExtract findLabeledExtract(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (labeledExtractQuery == null) 
		{
			labeledExtractQuery = initReferenceQuery(dc, idMethod, LabeledExtract.getQuery());
		}
		return findReferencedItem(dc, idMethod, labeledExtractQuery, identifier, false);
	}
	
	private ItemQuery<BioPlate> bioPlateQuery;
	/**
	 	Find a bioplate with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A bioplate, or null if no item could be found
	 */
	protected BioPlate findBioPlate(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (bioPlateQuery == null)
		{
			bioPlateQuery = initReferenceQuery(dc, idMethod, BioPlate.getQuery());
		}
		return findReferencedItem(dc, idMethod, bioPlateQuery, identifier, false);
	}
	
	private ItemQuery<ArraySlide> arraySlideQuery;
	/**
		Find an array slide with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An array slide, or null if no item could be found
	*/
	protected ArraySlide findArraySlide(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (arraySlideQuery == null) 
		{
			arraySlideQuery = initReferenceQuery(dc, idMethod, ArraySlide.getQuery());
		}
		return findReferencedItem(dc, idMethod, arraySlideQuery, identifier, false);
	}
	
	private ItemQuery<Hybridization> hybQuery;
	/**
		Find a hybridization with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A hybridization, or null if no item could be found
	*/
	protected Hybridization findHybridization(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (hybQuery == null) 
		{
			hybQuery = initReferenceQuery(dc, idMethod, Hybridization.getQuery());
		}
		return findReferencedItem(dc, idMethod, hybQuery, identifier, false);
	}

	private ItemQuery<Scan> scanQuery;
	/**
		Find a scan with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return A scan, or null if no item could be found
	*/
	protected Scan findScan(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (scanQuery == null) 
		{
			scanQuery = initReferenceQuery(dc, idMethod, Scan.getQuery());
		}
		return findReferencedItem(dc, idMethod, scanQuery, identifier, false);
	}
	
	private ItemQuery<ArrayDesign> designQuery;
	/**
		Find an array design with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An array design, or null if no item could be found
	*/
	protected ArrayDesign findArrayDesign(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (designQuery == null) 
		{
			designQuery = initReferenceQuery(dc, idMethod, ArrayDesign.getQuery());
		}
		return findReferencedItem(dc, idMethod, designQuery, identifier, false);
	}

	private ItemQuery<ArrayBatch> batchQuery;
	/**
		Find an array batch with a given identifier. This is a utility method that
		subclasses can use when creating or updating items.
		<p>
		NOTE! The first time this method is called a query object is initialised 
		using the {@link IdMethod#prepareQuery(DbControl, ItemQuery)} method.
		Subsequent calls uses the same query. Thus, this method should always be
		called with the same id method object, otherwise the result is undefined.
		
		@param dc The DbControl to use for database access
		@param identifier The identifier protocol
		@return An array batch, or null if no item could be found
	*/
	protected ArrayBatch findArrayBatch(DbControl dc, IdMethod idMethod, String identifier)
	{
		if (identifier == null) return null;
		if (batchQuery == null) 
		{
			batchQuery = initReferenceQuery(dc, idMethod, ArrayBatch.getQuery());
		}
		return findReferencedItem(dc, idMethod, batchQuery, identifier, false);
	}

	/**
		Cache that holds loaded/created items. The cache key is item type + identifier,
		eq. SAMPLE:My sample, EXTRACT:An extract. The same item may be cached under more 
		than one key depending on which identifier has been used to look it up,
		eg name, id, etc.: SAMPLE:My sample, SAMPLE:10.
	*/
	private Map<String, BasicItem> itemCache;
	
	/**
		Initialise a query that is used to find referenced items. This method
		delegates to {@link IdMethod#prepareQuery(DbControl, ItemQuery)} and
		then adds {@link Include} options: MINE, IN_PROJECT, SHARED and OTHERS.
				
		@param dc A DbControl used for database access
		@param idMethod The identification method to use
		@param query The query to initialised
		@return The initialised query
	*/
	protected <T extends BasicItem> ItemQuery<T> initReferenceQuery(DbControl dc, IdMethod idMethod, ItemQuery<T> query)
	{
		query = idMethod.prepareQuery(dc, query);
		query.include(Include.MINE, Include.IN_PROJECT, Include.SHARED, Include.OTHERS);
		return query;
	}
	
	/**
		Find a referenced item. This method will first look in the {@link #itemCache}.
		If no item is found in the cache it uses the query and id method 
		({@link IdMethod#find(DbControl, ItemQuery, String)}) to look for the item.
		The result of the lookup is stored in the cache (even if no item is found).

		@param dc A DbControl used for database access
		@param idMethod The identification method 
		@param query
		@param identifier
		@param ignoreNotFound TRUE to ignore not found errors
		@return The item, or null if none is found
		@throws ItemNotFoundException If no item is found and the error handling is set to
			fail for this condition
		@throws BaseException If multiple items are found and the error handling is set to
			fail for this condition
	*/
	@SuppressWarnings("unchecked")
	protected <T extends BasicItem> T findReferencedItem(DbControl dc, IdMethod idMethod, 
		ItemQuery<T> query, String identifier, boolean ignoreNotFound)
	{
		T item = null;
		Item itemType = query.getItemType();
		String cacheKey = itemType.name() + ":" + identifier;
		if (itemCache.containsKey(cacheKey))
		{
			item = (T)itemCache.get(cacheKey);
		}
		else
		{
			List<T> result = idMethod.find(dc, query, identifier);
			if (result.size() == 0)
			{
				// Not found
				if (failIfNotFoundReference && !ignoreNotFound)
				{
					throw new ItemNotFoundException(itemType+"[" +idMethod + "=" + identifier + "]");
				}
			}
			else
			{
				// Pick out items where use-permission is granted
				List<T> accessibleItems = new ArrayList<T>();
				for (T t : result)
				{
					if (t.hasPermission(Permission.USE)) 
					{
						accessibleItems.add(t);
					}
				}
				if (accessibleItems.size() > 1)
				{
					// Several items found
					if (failIfMultipleFoundReferences)
					{
						throw new BaseException("Found " + accessibleItems.size() + " " + itemType + " [" +idMethod + "=" + identifier + "]");
					}
				}
				else if (accessibleItems.size() == 0)
				{
					// No use permission on item
					if (failIfNoUsePermissionToReference)
					{
						throw new PermissionDeniedException(Permission.USE, itemType+"[" +idMethod + "=" + identifier + "]");
					}					
				}
				else
				{
					item = accessibleItems.get(0);
				}
				itemCache.put(cacheKey, item);
			}
		}
		return item;
	}
	
	private RequestInformation configureParser;
	private RequestInformation configureOptions;

	private Pattern getPattern(ParameterValues parameters, String name)
		throws BaseException
	{
		String regexp = (String)parameters.getValue(name);
		return regexp != null ? Pattern.compile(regexp) : null;
	}
	
	protected FlatFileParser getInitializedFlatFileParser(ParameterValues parameters)
		throws BaseException
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setDataHeaderRegexp(getPattern(parameters, dataHeaderRegexpParameter.getName()));
		ffp.setDataSplitterRegexp(getPattern(parameters, dataSplitterRegexpParameter.getName()));
		Boolean trimQuotes = (Boolean)parameters.getValue(trimQuotesParameter.getName());
		if (trimQuotes != null) ffp.setTrimQuotes(trimQuotes);
		ffp.setIgnoreRegexp(getPattern(parameters, ignoreRegexpParameter.getName()));
		ffp.setDataFooterRegexp(getPattern(parameters, dataFooterRegexpParameter.getName()));
		ffp.setMinDataColumns(IntegerUtil.getInt((Integer)parameters.getValue(minDataColumnsParameter.getName())));
		ffp.setMaxDataColumns(IntegerUtil.getInt((Integer)parameters.getValue(maxDataColumnsParameter.getName())));
		return ffp;
	}
	
	protected RequestInformation getConfigureParserParameters(GuiContext context, boolean forJob)
	{
		if (configureParser == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			if (forJob)
			{
				// The file to import from
				parameters.add(fileParameter);
				
				List<String> allowedModes = new ArrayList<String>();
				String defaultMode = "update";
				if (sc.hasPermission(Permission.CREATE, context.getItem()))
				{
					allowedModes.add("create+update");
					allowedModes.add("create");
					defaultMode = "create+update";
				}
				allowedModes.add("update");
				PluginParameter<String> modeParameter = new PluginParameter<String>(
						"mode", 
						"Mode", 
						"Select an import mode: \n" + 
						"create+update = create missing items, update existing items\n" +
						"create = create missing items only\n" +
						"update = update existing items only\n\n" +
						"NOTE! Not all modes may be available due to lack of permissions.",
						new StringParameterType(255, defaultMode, true, 1, 0, 0, allowedModes)
						);
				parameters.add(modeParameter);
				addMoreJobParameters(parameters);
			}

			// Identification method
			parameters.add(identificationSection);
			Enumeration<String, String> idMethods = new Enumeration<String, String>();
			for (IdMethod idm : getIdMethods())
			{
				idMethods.add(idm.getMethod(), idm.getTitle());
			}

			PluginParameter<String> idMethodParameter = new PluginParameter<String>(
				"idMethod",
				"Identification method",
				"The method to use to find already existing items. You can use either the " +
				"internal BASE ID, the name or, for some items, some another appropriate identifer " +
				"property. You need to specify a column mapping for " +
				"the selected method.",
				new StringParameterType(null, "name", true, 1, 0, 0, idMethods)
			);
			parameters.add(idMethodParameter);
			
			if (forJob)
			{
				BooleanParameterType optionalBooleanType = new BooleanParameterType(null, false);
				ItemContext cc = sc.getCurrentContext(context.getItem());
				Set<Include> include = cc.getInclude();
				// Should we search in MINE, SHARED, IN_PROJECT, etc.
				parameters.add(new PluginParameter<Boolean>(
						"includeMine",
						"Owned by me",
						"Search items that are owned by yourself.",
						include.contains(Include.MINE) ? Boolean.TRUE : null,
						optionalBooleanType)
					);
				parameters.add(new PluginParameter<Boolean>(
						"includeShared",
						"Shared to me",
						"Search items that have been shared to you by other users.",
						include.contains(Include.SHARED) ? Boolean.TRUE : null,
						optionalBooleanType)
					);
					
				if (sc.getActiveProjectId() != 0)
				{
					parameters.add(new PluginParameter<Boolean>(
							"includeInProject",
							"In current project",
							"Search items that are shared to the current project.",
							include.contains(Include.IN_PROJECT) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
				}
				if (sc.hasPermission(Permission.READ, context.getItem()))
				{
					parameters.add(new PluginParameter<Boolean>(
							"includeOthers",
							"Owned by others",
							"Search items that are owned by other users.",
							include.contains(Include.OTHERS) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
				}
			}

			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(dataHeaderRegexpParameter);
			parameters.add(dataSplitterRegexpParameter);
			parameters.add(trimQuotesParameter);
			parameters.add(ignoreRegexpParameter);
			parameters.add(dataFooterRegexpParameter);
			parameters.add(minDataColumnsParameter);
			parameters.add(maxDataColumnsParameter);
			parameters.add(Parameters.charsetParameter(null, null, getCharset()));
			parameters.add(Parameters.decimalSeparatorParameter(null, null, getDecimalSeparator()));
			parameters.add(Parameters.dateFormatParameter(null, null, 
				sc.getUserClientSetting("formatter.date.formatstring")));
			addMoreParsingParameters(parameters);
			
			// Column mappings
			parameters.add(mappingSection);
			parameters.add(complexMappings);
			addMoreColumnMappingParameters(parameters);
			
			if (forJob)
			{
				configureParser = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Select a file from which to import item data",
					"Here you select which file to import data from, and the " +
					"regular expressions and other settings used to parse it. Step 1 of 2.",
					parameters
				);
			}
			else
			{
				configureParser = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_PLUGIN,
					"Specify parser regular expressions and settings",
					"This plug-in can be configured with the regular expressions and " +
					"settings used to parse files.",
					parameters
				);				
			}
		}
		return configureParser;
	}
	
	protected RequestInformation getConfigureOptionsParameters(GuiContext context)
	{
		if (configureOptions == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			String mode = (String)job.getValue("mode");
			boolean updateMode = mode == null || mode.contains("update");
			boolean createMode = mode == null || mode.contains("create");
			
			// Error handling options
			parameters.add(errorSection);
			parameters.add(Parameters.logFileParameter(null, null, null));
			parameters.add(Parameters.dryRunParameter(null, null, null));
			parameters.add(defaultErrorParameter);
			if (!createMode) parameters.add(itemNotFoundErrorParameter);
			if (!updateMode) parameters.add(itemExistsErrorParameter);
			parameters.add(multipleItemsFoundErrorParameter);
			if (updateMode) parameters.add(noWritePermissionToItemErrorParameter);
			parameters.add(referenceNotFoundErrorParameter);
			parameters.add(noUsePermissionToReferenceErrorParameter);
			parameters.add(multipleReferencesFoundErrorParameter);
			parameters.add(Parameters.invalidUseOfNullError(null, null, null));
			parameters.add(stringTooLongErrorParameter);
			parameters.add(numberFormatErrorParameter);
			parameters.add(numberOutOfRangeErrorParameter);
			addMoreErrorParameters(parameters);
			
			configureOptions = new RequestInformation
			(
				CONFIGURE_OPTIONS,
				"Other settings and error handling options",
				"Specify other settings and error handling options. Step 2 of 2.",
				parameters
			);
		}
		return configureOptions;	
	}


}
