/*
 	$Id: TestAnyToAny.java 4514 2008-09-11 20:07:12Z jari $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import java.util.Date;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Sample;

public class TestAnyToAny
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing any-to-any");
		write_header();
		// Standard tests: create, load, list
		int sampleId = TestSample.test_create(0, null, true);
		int fileId1 = TestFile.test_create("Linked file #1", false, true);
		int fileId2 = TestFile.test_create("Linked file #2", false, true);
		int id = test_create(sampleId, fileId1, "first");
		int id2 = test_create(sampleId, fileId2, "second");
		test_load(id);
		test_list_from(sampleId, 2);
		test_list_to(fileId1, 1);
		test_exists(sampleId, "first", true);
		test_exists(sampleId, "third", false);
		

		// Extra tests:
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();

		// Standard test: Delete
		test_delete(id);
		test_delete_all(sampleId, 1);
		
		TestSample.test_delete(sampleId);
		TestFile.test_delete(fileId1);
		TestFile.test_delete(fileId2);
		write("++Testing any-to-any "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int sampleId, int fileId, String name)
	{
		if (sampleId == 0 || fileId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			File f = File.getById(dc, fileId);
			AnyToAny a = AnyToAny.getNew(dc, s, f, name, true);
			a.setDescription("Added at "+new Date());
			dc.saveItem(a);
			dc.commit();
			id = a.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(a);
			write_item(0, a);
			write("--Create any-to-any OK");
		}
		catch (Throwable ex)
		{
			write("--Create any-to-any FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnyToAny a = AnyToAny.getById(dc, id);
			write_item(0, a);
			write("--Load any-to-any OK");
		}
		catch (Throwable ex)
		{
			write("--Load any-to-any FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			AnyToAny a = AnyToAny.getById(dc, id);
			dc.deleteItem(a);
			dc.commit();
			write("--Delete any-to-any OK");
		}
		catch (Throwable ex)
		{
			write("--Delete any-to-any FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tFrom    \tTo");
			write("-- \t-- \t--------- \t-----------\t--------\t----------");
		}
	}
	static void write_item(int i, AnyToAny a)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+a.getId()+"\t"+a.getName()+"\t"+
			a.getDescription()+"\t"+a.getFrom()+"\t"+a.getTo());
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_list_from(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ItemResultList<AnyToAny> l = AnyToAny.getLinksFrom(s).list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List any-to-any from item OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List any-to-any from item FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_to(int fileId, int expectedResults)
	{
		if (fileId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			File f = File.getById(dc, fileId);
			ItemResultList<AnyToAny> l = AnyToAny.getLinksTo(f).list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List any-to-any to item OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List any-to-any to item FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			int deletedLinks = AnyToAny.unlinkAllFrom(dc, s);
			dc.commit();
			if (expectedResults >= 0 && expectedResults != deletedLinks)
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+deletedLinks);
			}
			write("--Delete all any-to-any from item OK ("+deletedLinks+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all any-to-any from item FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_exists(int sampleId, String name, boolean expected)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			if (AnyToAny.exists(dc, s, name) != expected)
			{
				throw new BaseException("Expected exists=" + expected + " for link=" + name);
			}
			write("--Exists any-to-any (" + name + "=" + expected + ") OK");
		}
		catch (Throwable ex)
		{
			write("--Exists any-to-any FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}	

	
}
