/*
	$Id: TestSample.java 4740 2009-02-05 13:39:25Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Hakkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.util.biomaterial.PooledChildrenTransformer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

public class TestSample
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing samples");
		write_header();
		// Standard tests: create, load, list
		int id = test_create(0, null, true);
		int id2 = test_create(0, null, false);
		
		int bioSourceId = TestBioSource.test_create(null, false);
		int id3 = test_create(bioSourceId, null, false);
		int id4 = test_create_pooled(id, id2, id3);

		test_load(id);
		test_list(-1);
		
		write_events_header();
		test_add_event(id, 100.0f);
		test_list_events(id, 3);  // creation event, pooling event for id4, and extra event
		
		write_sources_header();
		test_list_sources(id4, 3); // The three id, id2 and id3 samples
		test_list_pooledChildren(id2, 1);
		test_pooledChildrenTransformer(id2, 1);

		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete(id4);
		test_delete(id3);
		test_delete(id2);
		test_delete(id);
		TestBioSource.test_delete(bioSourceId);
	
		write("++Testing samples "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int bioSourceId, String name, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.SAMPLE)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getNew(dc);
			if (setAll)
			{
				s.setName("Test sample");
				s.setDescription("Added at "+new Date());
				s.setExternalId("XCV-455-SAMPLE");
				s.setOriginalQuantity(500.0f);
			}
			if (bioSourceId != 0)
			{
				BioSource bs = BioSource.getById(dc, bioSourceId);
				s.setBioSource(bs);
				s.setName(bs.getName() + ".s");
			}
			if (name != null) s.setName(name);
			dc.saveItem(s);
			dc.commit();
			id = s.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(s);
			write_item(0, s);
			write("--Create sample OK");
		}
		catch (Throwable ex)
		{
			write("--Create sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_create_pooled(int... sourceIds)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.SAMPLE)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getNew(dc);
			s.setName("Test pooled sample");
			s.setDescription("Added at "+new Date());
			s.setExternalId("POOLED-455-SAMPLE");
			s.setOriginalQuantity(500.0f);
			s.setPooled(true);
			BioMaterialEvent evt = s.getCreationEvent();
			for (int sourceId : sourceIds)
			{
				Sample src = Sample.getById(dc, sourceId);
				Float original = src.getOriginalQuantity();
				evt.addSource(src, original == null ? null : original/2);
			}
			dc.saveItem(s);
			dc.commit();
			id = s.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(s);
			write_item(0, s);
			write("--Create pooled sample OK");
		}
		catch (Throwable ex)
		{
			write("--Create pooled sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, id);
			write_item(0, s);
			write("--Load sample OK");
		}
		catch (Throwable ex)
		{
			write("--Load sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Sample> l = Sample.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List samples OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List samples FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, id);
			dc.deleteItem(s);
			dc.commit();
			write("--Delete sample OK");
		}
		catch (Throwable ex)
		{
			write("--Delete sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Sample> l = Sample.getQuery().list(dc);
			Trashcan.delete(dc.getSessionControl(), l, true, null);
			dc.commit();
			write("--Delete all samples OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all samples FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tBioSource\tQuantity\tRemain");
			write("-- \t-- \t--------- \t-----------\t-----------\t---------\t--------\t------");
		}
	}
	static void write_item(int i, Sample s)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+s.getId()+"\t"+s.getName()+"\t"+s.getDescription()+
			"\t"+s.getExternalId()+"\t"+s.getBioSource()+"\t"+s.getOriginalQuantity()+"\t"+s.getRemainingQuantity());
	}
	static void write_events_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tBiomaterial\tEvent type\tEvent date\tEntry date\tUsed quantity\tComment");
			write("-- \t-- \t-----------\t----------\t----------\t----------\t-------------\t-------");
		}
	}
	static void write_item(int i, BioMaterialEvent evt, Sample s)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+evt.getId()+"\t"+evt.getBioMaterial()+"\t"+evt.getEventType()+
			"\t"+evt.getEventDate()+"\t"+evt.getEntryDate()+"\t"+evt.getUsedQuantity(s)+"\t"+evt.getComment());
	}
	static void write_sources_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tPooled\tQuantity\tRemain\tUsed");
			write("-- \t-- \t--------- \t-----------\t-----------\t------\t--------\t------\t----");
		}
	}
	static void write_item(int i, MeasuredBioMaterial b, BioMaterialEvent evt)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+b.getId()+"\t"+b.getName()+"\t"+b.getDescription()+
			"\t"+b.getExternalId()+"\t"+b.isPooled()+"\t"+b.getOriginalQuantity()+"\t"+b.getRemainingQuantity()+
			"\t"+evt.getUsedQuantity(b));
	}

	static void write(String message)
	{
		System.out.println(message);
	}
	
	static int test_add_event(int sampleId, float usedQuantity)
	{
		if (sampleId == 0) return 0;
		int eventId = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			BioMaterialEvent bme = s.newEvent();
			bme.setUsedQuantity(usedQuantity);
			bme.setComment("Added a new event");
			dc.saveItem(bme);
			dc.commit();
			eventId = bme.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bme);
			write_item(0, bme, s);
			write("--Add event for sample OK");
		}
		catch (Throwable ex)
		{
			write("--Add event for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return eventId;
	}
	
	static void test_list_events(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ItemResultList<BioMaterialEvent> l = s.getEvents().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i), s);
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List events for sample OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List events for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_sources(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			BioMaterialEvent evt = s.getCreationEvent();
			ItemResultList<? extends MeasuredBioMaterial> l = evt.getSources().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i), evt);
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List sources for sample OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List sources for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	static void test_list_pooledChildren(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			ItemResultList<BioMaterialEvent> ev = s.getPoolingEvents().list(dc);
			
			for (int i=0; i<ev.size(); i++)
			{
				write_item(i, (Sample)ev.get(i).getBioMaterial());
			}
			if (expectedResults >= 0 && expectedResults != ev.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+ev.size());
			}
			write("--List pools from sample OK ("+ev.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List pools for sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_pooledChildrenTransformer(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			PooledChildrenTransformer<Sample> pct = new PooledChildrenTransformer<Sample>(dc, false, Sample.getQuery());
			List<Sample> children = new ArrayList<Sample>();
			int numAdded = pct.transform(Collections.singleton(s), children);

			for (int i=0; i<children.size(); i++)
			{
				write_item(i, children.get(i));
			}
			if (expectedResults >= 0 && expectedResults != numAdded && numAdded != children.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+numAdded);
			}
			write("--List pools from sample using transformer OK ("+numAdded+")");
		}
		catch (Throwable ex)
		{
			write("--List pools for sample using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}
