/**
	$Id: LowessTest.java 4514 2008-09-11 20:07:12Z jari $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test.performance;


import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.FloatParameterType;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.test.PluginUtil;
import net.sf.basedb.test.TestUtil;
import net.sf.basedb.util.Values;

/**
	Create a normalized bioassayset with LOWESS.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-11 22:07:12 +0200 (Thu, 11 Sep 2008) $
*/
public class LowessTest
{

	public static boolean ok = true;

	/**
		Run the tests in this class.
	*/
	public static boolean run(String user, String password)
	{
		DbControl dc = null;
		TestUtil.write("++Entering LOWESS test\n");
		try
		{
			TestUtil.login(user, password, false);

			// Create raw bioassays
			dc = TestUtil.getDbControl();
			
			// Load prepared items
			Experiment experiment = PrepareTest.getExperiment(dc);
			Float fitFraction = Values.getFloat(Config.get("lowess.fitFraction"), 0.33f);
			Float delta = Values.getFloat(Config.get("lowess.delta"), 0.1f);
			Integer iterations = Values.getInteger(Config.get("lowess.iterations"), 4);
			Integer blockGroup = Values.getInteger(Config.get("lowess.blockGroup"), 12);
			BioAssaySet filtered = Util.findBioAssaySet(dc, experiment, "Filtered%");
			long numBioAssays = filtered.getBioAssays().count(dc);
			if (filtered == null)
			{
				throw new ItemNotFoundException("Filtered bioassayset");
			}
			
			
			long num = experiment.getBioAssaySets().count(dc) + 1;
			dc.close();

			long time = createNormalizedBioAssaySet(filtered, "LOWESS #" + num, fitFraction, delta, iterations, blockGroup) / 1000;
			
			TestUtil.write("---------------------------------------------\n");
			TestUtil.write("Source: " + filtered.getName() + "\n");
			TestUtil.write("  BioAssasys: " + numBioAssays + "\n");
			TestUtil.write("  Spots: " + filtered.getNumSpots() + "\n");
			TestUtil.write("  Reporters: " + filtered.getNumReporters() + "\n");
			TestUtil.write("Fit fraction=" + fitFraction + "\n");
			TestUtil.write("Delta=" + delta + "\n");
			TestUtil.write("Iterations=" + iterations + "\n");
			TestUtil.write("Block group=" + blockGroup + "\n");
			TestUtil.write("Create LOWESS bioassayset from " + filtered.getName() + ": " + 
					Values.formatTime(time) + "\n");
			TestUtil.write("---------------------------------------------\n");
		
			TestUtil.logout();
		}
		catch (Throwable t)
		{
			ok = false;
			t.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		TestUtil.write("++Leaving LOWESS test (" + (ok ? "OK" : "FAILED") + ")\n");
		return ok;
	}

	/**
		Create a filtered bioassay set.
	*/
	public static long createNormalizedBioAssaySet(BioAssaySet bioAssaySet, String name, 
		Float fitFraction, Float delta, Integer iterations, Integer blockGroup)
		throws Exception
	{
		TestUtil.write("--Create LOWESS bioassay set: " + name + "\n");
		DbControl dc = TestUtil.getDbControl();
		try
		{		
			PluginDefinition plugin = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.LowessNormalization");
			Job job = Job.getNew(dc, plugin, null, bioAssaySet.getExperiment());
			job.setName("PerformanceTest: Create LOWESS bioassay set: " + name);
			job.setSendMessage(false);
			job.setRemoveJobWhenFinished(true);
			job.setParameterValue("source", 
					new ItemParameterType<BioAssaySet>(BioAssaySet.class, null), bioAssaySet);
			job.setParameterValue("childName", 
					new StringParameterType(255, null, true), name);
			job.setParameterValue("fitFraction", new FloatParameterType(), fitFraction);
			job.setParameterValue("delta", new FloatParameterType(), delta);
			job.setParameterValue("iter", new IntegerParameterType(), iterations);
			job.setParameterValue("blockGroup", new IntegerParameterType(), blockGroup);
			dc.saveItem(job);
			dc.commit();
			long time = System.currentTimeMillis();
			PluginUtil.executeJob(job);
			return  System.currentTimeMillis() - time;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	
}
