/**
	$Id: Util.java 4625 2008-11-04 15:45:36Z martin $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test.roles;

import java.util.Collection;
import java.util.List;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Group;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Label;
import net.sf.basedb.core.Plate;
import net.sf.basedb.core.PlateGeometry;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.Quota;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.Role;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.User;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;

/**

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-11-04 16:45:36 +0100 (Tue, 04 Nov 2008) $
*/
public class Util
{

	/**
		Create a user.
	*/
	public static User createUser(DbControl dc, String name, String login, String password,
		Quota quota, Group quotaGroup, Collection<Role> roles, Collection<Group> groups)
	{
		User u = User.getNew(dc, login, password);
		u.setName(name);
		u.setQuota(quota);
		u.setQuotaGroup(quotaGroup);

		if (roles != null)
		{
			for (Role r : roles)
			{
				r.addUser(u);
			}
		}

		if (groups != null)
		{
			for (Group g : groups)
			{
				g.addUser(u);
			}
		}
		
		dc.saveItem(u);
		return u;
	}
	
	/**
		Find the quota with a matching name.
	*/
	public static Quota findQuota(DbControl dc, String name)
	{
		ItemQuery<Quota> query = Quota.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
		Find the group with a matching name.
	*/
	public static Group findGroup(DbControl dc, String name)
	{
		ItemQuery<Group> query = Group.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}

	/**
		Find the project with a matching name.
	*/
	public static Project findProject(DbControl dc, String name)
	{
		ItemQuery<Project> query = Project.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		query.include(Include.SHARED);
		return query.list(dc).get(0);
	}

	/**
		Find the protocol with a matching name.
	*/
	public static Protocol findProtocol(DbControl dc, String name)
	{
		ItemQuery<Protocol> query = Protocol.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
		Find the hardware with a matching name.
	*/
	public static Hardware findHardware(DbControl dc, String name)
	{
		ItemQuery<Hardware> query = Hardware.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}

	/**
		Find the software with a matching name.
	*/
	public static Software findSoftware(DbControl dc, String name)
	{
		ItemQuery<Software> query = Software.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}

	/**
		Find the plate geometry with matching rows and columns
	*/
	public static PlateGeometry findGeometry(DbControl dc, int rows, int columns)
	{
		ItemQuery<PlateGeometry> query = PlateGeometry.getQuery();
		query.restrict(Restrictions.eq(Hql.property("rows"), Expressions.integer(rows)));
		query.restrict(Restrictions.eq(Hql.property("columns"), Expressions.integer(columns)));
		return query.list(dc).get(0);
	}

	/**
		Find the plates in the current project.
	*/
	public static List<Plate> findPlates(DbControl dc)
	{
		ItemQuery<Plate> query = Plate.getQuery();
		query.order(Orders.asc(Hql.property("name")));
		return query.list(dc);
	}

	/**
		Find the annotation type with a matching name and item type.
	*/
	public static AnnotationType findAnnotationType(DbControl dc, String name, Item item)
	{
		ItemQuery<AnnotationType> query = AnnotationType.getQuery(item);
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}

	/**
		Annotatate an item.
	*/
	public static void setAnnotation(Annotatable item, AnnotationType at, Object value, Unit unit)
	{
		AnnotationSet as = item.getAnnotationSet();
		Annotation a = as.getAnnotation(at);
		a.setValue(value, unit);
	}
	
	/**
		Find the label with a matching name.
	*/
	public static Label findLabel(DbControl dc, String name)
	{
		ItemQuery<Label> query = Label.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		query.include(Include.SHARED);
		return query.list(dc).get(0);
	}

	/**
		Find the array slide with a matching name.
	*/
	public static ArraySlide findArraySlide(DbControl dc, String name)
	{
		ItemQuery<ArraySlide> query = ArraySlide.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
		Find the array design with a matching name.
	*/
	public static ArrayDesign findArrayDesign(DbControl dc, String name)
	{
		ItemQuery<ArrayDesign> query = ArrayDesign.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
		Find the plugin configuration with a matching name.
	*/
	public static PluginConfiguration findPluginConfiguration(DbControl dc, String name)
	{
		ItemQuery<PluginConfiguration> query = PluginConfiguration.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}

	/**
		Find the experiment with a matching name.
	*/
	public static Experiment findExperiment(DbControl dc, String name)
	{
		ItemQuery<Experiment> query = Experiment.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}

	/**
		Find a bioassay set with a matching name.
	*/
	public static BioAssaySet findBioAssaySet(DbControl dc, Experiment experiment, String name)
	{
		ItemQuery<BioAssaySet> query = BioAssaySet.getQuery(experiment);
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
		Find the formula with a matching name.
	*/
	public static Formula findFormula(DbControl dc, RawDataType rdt, Formula.Type type, String name)
	{
		ItemQuery<Formula> query = Formula.getQuery(type, rdt);
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		query.include(Include.SHARED);
		return query.list(dc).get(0);
	}

	/**
	 	Returns the first raw bioassay that was found with a certain name
	 */
	public static RawBioAssay findRawBioAssay(DbControl dc, String name)
	{
		ItemQuery<RawBioAssay> query = RawBioAssay.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
	 	Returns the first sample that was found with a certain name
	 */
	public static Sample findSample(DbControl dc, String name)
	{
		ItemQuery<Sample> query = Sample.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
	
	/**
	 	Returns the first biosource that was found with a certain name
	 */
	public static BioSource findBioSource(DbControl dc, String name)
	{
		ItemQuery<BioSource> query = BioSource.getQuery();
		query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
		return query.list(dc).get(0);
	}
}
